@@ -26,56 +26,35 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <mysql.h>
 
-#ifndef _MSC_VER
+#ifdef _MSC_VER
+# include <windows.h>
+# define inline __inline
+#else /* !_MSC_VER */
 #include <unistd.h>
-#ifndef BSD
-#include <crypt.h>
-#endif /* BSD */
 #endif
 
+#include <mysql.h>
+
 #include <wzd_backend.h>
+#include <libwzd-auth/wzd_md5.h>
+#include <libwzd-auth/wzd_md5crypt.h>
 #include <wzd_debug.h>
 
-enum {
-  UCOL_REF=0,
-  UCOL_USERNAME,
-  UCOL_USERPASS,
-  UCOL_ROOTPATH,
-  UCOL_UID,
-  UCOL_FLAGS,
-  UCOL_MAX_UL_SPEED,
-  UCOL_MAX_DL_SPEED,
-  UCOL_NUM_LOGINS,
-  UCOL_RATIO,
-  UCOL_USER_SLOTS,
-  UCOL_LEECH_SLOTS,
-  UCOL_LAST_LOGIN,
-};
+#include "libmysql.h"
 
-enum {
-  GCOL_REF=0,
-  GCOL_GROUPNAME,
-  GCOL_GID,
-};
+#define MYSQL_BACKEND_VERSION   121
 
-enum {
-  UIPCOL_REF=0,
-  UIPCOL_IP,
-};
+/* IMPORTANT needed to check version */
+BACKEND_NAME(mysql);
+BACKEND_VERSION(MYSQL_BACKEND_VERSION);
 
 
-
-static MYSQL mysql;
+MYSQL mysql;
 static char *db_user, *db_passwd, *db_hostname, *db;
 
-static void wzd_mysql_error(const char *filename, const char  *func_name, int line); /*, const char *error); */
 /*static int wzd_parse_arg(const char *arg);*/ /* parse arg (login:password@hostname:table) */
-static int wzd_parse_arg(char *arg);
-
-/* basic syntax checking to avoid injections */
-static int wzd_mysql_check_name(const char *name);
+static int wzd_parse_arg(const char *arg);
 
 /* get mysql value, in a more robust way than just a copy
  * return 0 if ok, non-zero otherwise (ex: value is NULL)
@@ -84,85 +63,81 @@ static inline int wzd_row_get_string(char *dst, unsigned int dst_len, MYSQL_ROW
 static inline int wzd_row_get_long(long *dst, MYSQL_ROW row, unsigned int index);
 static inline int wzd_row_get_uint(unsigned int *dst, MYSQL_ROW row, unsigned int index);
 static inline int wzd_row_get_ulong(unsigned long *dst, MYSQL_ROW row, unsigned int index);
+static inline int wzd_row_get_ullong(u64_t *dst, MYSQL_ROW row, unsigned int index);
 
-static int * wzd_mysql_get_user_list(void);
-static int * wzd_mysql_get_group_list(void);
+static uid_t * wzd_mysql_get_user_list(void);
+static gid_t * wzd_mysql_get_group_list(void);
 
 
 
 
-static void wzd_mysql_error(const char *filename, const char  *func_name, int line)/*, const char *error)*/
+
+void _wzd_mysql_error(const char *filename, const char  *func_name, int line)/*, const char *error)*/
 {
   fprintf(stderr, "%s(%s):%d %s\n", filename, func_name, line, mysql_error(&mysql));
 }
 
-static int wzd_parse_arg(char *arg)
+static int wzd_parse_arg(const char *arg)
 {
   char *ptr;
+  char * buffer;
 
-  ptr = arg;
+  if (!arg) return -1;
 
-  db_user = (char*)strtok_r(arg, ":", &ptr);
-  if (!db_user) return -1;
+  ptr = buffer = strdup(arg); /** \todo free buffer at backend exit ! (small memory leak) */
 
-  db_passwd = (char *)strtok_r(NULL,"@", &ptr);
-  if (!db_passwd) return -1;
+  db_user = strtok_r(buffer, ":", &ptr);
+  if (!db_user) { free(buffer); return -1; }
 
-  db_hostname = (char *)strtok_r(NULL, ":\n", &ptr);
-  if (!db_hostname) return -1;
+  db_passwd = strtok_r(NULL,"@", &ptr);
+  if (!db_passwd) { free(buffer); return -1; }
 
-  db = (char *)strtok_r(NULL, "\n", &ptr);
-  if (!db) return -1;
+  db_hostname = strtok_r(NULL, ":\n", &ptr);
+  if (!db_hostname) { free(buffer); return -1; }
+
+  db = strtok_r(NULL, "\n", &ptr);
+  if (!db) { free(buffer); return -1; }
 
   return 0;
 }
 
 
-int FCN_INIT(int *backend_storage, wzd_user_t * user_list, unsigned int user_max, wzd_group_t * group_list, unsigned int group_max, void *arg)
+int FCN_INIT(const char *arg)
 {
-  if ((wzd_parse_arg((char *)arg)) != 0) {
+  if ((wzd_parse_arg(arg)) != 0) {
     return -1;
   }
 
-#ifdef DEBUG
-  fprintf(stderr, "User: %s\nHostname: %s\nDatabase name: %s\n", db_user, db_hostname, db);
-#endif
-
-  *backend_storage = 1;
-
   mysql_init(&mysql);
 
+  /** \todo XXX FIXME try using CLIENT_SSL for the last arg */
   if (!mysql_real_connect(&mysql, db_hostname, db_user, db_passwd, db, 0, NULL, 0)) {
-    wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
+    _wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
     mysql_close(&mysql);
     return -1;
   }
-#ifdef DEBUG
-  else
-    fprintf(stderr, "Connected to database");
-#endif
 
   return 0;
 }
 
-int FCN_VALIDATE_LOGIN(const char *login, wzd_user_t * user)
+uid_t FCN_VALIDATE_LOGIN(const char *login, wzd_user_t * user)
 {
   char *query;
-  int uid;
+  uid_t uid;
 
-  if (!wzd_mysql_check_name(login)) return -1;
+  if (!wzd_mysql_check_name(login)) return (uid_t)-1;
 
   query = malloc(512);
   snprintf(query, 512, "SELECT * FROM users WHERE username='%s'", login);
 
-  if (mysql_query(&mysql, query) != 0) { 
+  if (mysql_query(&mysql, query) != 0) {
     free(query);
-    wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
-    return -1;
+    _wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
+    return (uid_t)-1;
   }
   free(query);
 
-  uid = -1;
+  uid = (uid_t)-1;
 
 
   /** no !! this returns the number of COLUMNS (here, 14) */
@@ -173,14 +148,14 @@ int FCN_VALIDATE_LOGIN(const char *login, wzd_user_t * user)
     int num_fields;
 
     if (!(res = mysql_store_result(&mysql))) {
-      wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
-      return -1;
+      _wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
+      return (uid_t)-1;
     }
 
     if ( (int)mysql_num_rows(res) != 1 ) {
       /* 0 or more than 1 result  */
       mysql_free_result(res);
-      return -1;
+      return (uid_t)-1;
     }
 
     num_fields = mysql_num_fields(res);
@@ -200,25 +175,26 @@ int FCN_VALIDATE_LOGIN(const char *login, wzd_user_t * user)
   return uid;
 }
 
-int FCN_VALIDATE_PASS(const char *login, const char *pass, wzd_user_t * user)
+uid_t FCN_VALIDATE_PASS(const char *login, const char *pass, wzd_user_t * user)
 {
   char *query;
   char * cipher;
-  int uid;
+  uid_t uid;
+  char buffer[128];
 
-  if (!wzd_mysql_check_name(login)) return -1;
+  if (!wzd_mysql_check_name(login)) return (uid_t)-1;
 
   query = malloc(512);
   snprintf(query, 512, "SELECT * FROM users WHERE username='%s'", login);
 
   if (mysql_query(&mysql, query) != 0) {
     free(query);
-    wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
-    return -1;
+    _wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
+    return (uid_t)-1;
   }
 
   free(query);
-  uid = -1;
+  uid = (uid_t)-1;
 
 
   /** no !! this returns the number of COLUMNS (here, 14) */
@@ -230,14 +206,14 @@ int FCN_VALIDATE_PASS(const char *login, const char *pass, wzd_user_t * user)
     char stored_pass[MAX_PASS_LENGTH];
 
     if (!(res = mysql_store_result(&mysql))) {
-      wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
-      return -1;
+      _wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
+      return (uid_t)-1;
     }
 
     if ( (int)mysql_num_rows(res) != 1 ) {
       /* 0 or more than 1 result */
       mysql_free_result(res);
-      return -1;
+      return (uid_t)-1;
     }
 
     num_fields = mysql_num_fields(res);
@@ -259,13 +235,20 @@ int FCN_VALIDATE_PASS(const char *login, const char *pass, wzd_user_t * user)
     mysql_free_result(res);
 
     if (strlen(stored_pass) == 0)
-      return uid;	// passworldless login
-    /** NO ! The 'anything' pass is '%' ! */
+    {
+      fprintf(stderr,"WARNING: empty password field whould not be allowed !\n");
+      fprintf(stderr,"WARNING: you should run: UPDATE users SET userpass='%%' WHERE userpass is NULL\n");
+      return uid; /* passworldless login */
+    }
 
-    cipher = (char*)crypt(pass, stored_pass);
+    if (strcmp(stored_pass,"%")==0)
+      return uid; /* passworldless login */
 
-    if (!strcasecmp(cipher,stored_pass))
-      return -1;
+    cipher = (char*)md5_hash_r(pass, buffer, sizeof(buffer));
+    if (!cipher) return (uid_t)-1;
+
+    if (strncasecmp(cipher,stored_pass,32))
+      return (uid_t)-1;
 
   } /* else // user does not exist in table
     return -1;*/
@@ -274,24 +257,24 @@ int FCN_VALIDATE_PASS(const char *login, const char *pass, wzd_user_t * user)
   return uid;
 }
 
-int FCN_FIND_USER(const char *name, wzd_user_t * user)
+uid_t FCN_FIND_USER(const char *name, wzd_user_t * user)
 {
   char *query;
-  int uid;
+  uid_t uid;
 
-  if (!wzd_mysql_check_name(name)) return -1;
+  if (!wzd_mysql_check_name(name)) return (uid_t)-1;
 
   query = malloc(512);
   snprintf(query, 512, "SELECT * FROM users WHERE username='%s'", name);
 
   if (mysql_query(&mysql, query) != 0) {
     free(query);
-    wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
-    return -1;
+    _wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
+    return (uid_t)-1;
   }
 
   free(query);
-  uid = -1;
+  uid = (uid_t)-1;
 
   /** no !! this returns the number of COLUMNS (here, 14) */
 /*  if (mysql_field_count(&mysql) == 1)*/
@@ -301,14 +284,14 @@ int FCN_FIND_USER(const char *name, wzd_user_t * user)
     int num_fields;
 
     if (!(res = mysql_store_result(&mysql))) {
-      wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
-      return -1;
+      _wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
+      return (uid_t)-1;
     }
 
     if ( (int)mysql_num_rows(res) != 1 ) {
       /* 0 or more than 1 result */
       mysql_free_result(res);
-      return -1;
+      return (uid_t)-1;
     }
 
     num_fields = mysql_num_fields(res);
@@ -329,46 +312,19 @@ int FCN_FIND_USER(const char *name, wzd_user_t * user)
   return uid;
 }
 
-int FCN_FIND_GROUP(const char *name, wzd_group_t * group)
-{
-  // XXX: forgot about it while  wzd_group_t->gid is not implemented
-
-  return 0;
-}
-
-int FCN_CHPASS(const char *username, const char *new_pass)
-{
-  return 1;
-}
-
-/* if user does not exist, add it */
-int FCN_MOD_USER(const char *name, wzd_user_t * user, unsigned long mod_type)
-{
-  return 1;
-}
-
-int FCN_MOD_GROUP(const char *name, wzd_group_t * group, unsigned long mod_type)
-{
-  return 1;
-}
-
 int  FCN_COMMIT_CHANGES(void)
 {
   return 0;
 }
 
-int FCN_FINI()
+int FCN_FINI(void)
 {
-#ifdef DEBUG
-  fprintf(stderr, "Closing connection");
-#endif
-
   mysql_close(&mysql);
 
   return 0;
 }
 
-wzd_user_t * FCN_GET_USER(int uid)
+wzd_user_t * FCN_GET_USER(uid_t uid)
 {
   char *query;
   MYSQL_RES   *res;
@@ -377,20 +333,20 @@ wzd_user_t * FCN_GET_USER(int uid)
   wzd_user_t * user;
   unsigned int i,j;
 
-  if (uid == -2) return (wzd_user_t*)wzd_mysql_get_user_list();
+  if (uid == (uid_t)-2) return (wzd_user_t*)wzd_mysql_get_user_list();
 
   query = malloc(512);
   snprintf(query, 512, "SELECT * FROM users WHERE uid='%d'", uid);
 
-  if (mysql_query(&mysql, query) != 0) { 
+  if (mysql_query(&mysql, query) != 0) {
     free(query);
-    wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
+    _wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
     return NULL;
   }
 
   if (!(res = mysql_store_result(&mysql))) {
     free(query);
-    wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
+    _wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
     return NULL;
   }
 
@@ -418,12 +374,16 @@ wzd_user_t * FCN_GET_USER(int uid)
   wzd_row_get_string(user->userpass, MAX_PASS_LENGTH, row, UCOL_USERPASS);
   wzd_row_get_string(user->rootpath, WZD_MAX_PATH, row, UCOL_ROOTPATH);
   wzd_row_get_string(user->flags, MAX_FLAGS_NUM, row, UCOL_FLAGS);
-  wzd_row_get_ulong(&user->max_ul_speed, row, UCOL_MAX_UL_SPEED);
-  wzd_row_get_ulong(&user->max_dl_speed, row, UCOL_MAX_DL_SPEED);
+  wzd_row_get_uint((unsigned int*)&user->max_idle_time, row, UCOL_MAX_IDLE_TIME);
+  wzd_row_get_uint(&user->max_ul_speed, row, UCOL_MAX_UL_SPEED);
+  wzd_row_get_uint(&user->max_dl_speed, row, UCOL_MAX_DL_SPEED);
   if (wzd_row_get_uint(&i, row, UCOL_NUM_LOGINS)==0) user->num_logins = i;
   wzd_row_get_uint(&user->ratio, row, UCOL_RATIO);
   if (wzd_row_get_uint(&i, row, UCOL_USER_SLOTS)==0) user->user_slots = i;
   if (wzd_row_get_uint(&i, row, UCOL_LEECH_SLOTS)==0) user->leech_slots = i;
+  wzd_row_get_ulong(&user->userperms, row, UCOL_PERMS);
+  wzd_row_get_ullong(&user->credits, row, UCOL_CREDITS);
+  /* XXX FIXME last login */
 
   mysql_free_result(res);
 
@@ -432,14 +392,14 @@ wzd_user_t * FCN_GET_USER(int uid)
 
   snprintf(query, 512, "select UserIP.ip from UserIP,users where users.uid='%d' AND users.ref=UserIP.ref", uid);
 
-  if (mysql_query(&mysql, query) != 0) { 
+  if (mysql_query(&mysql, query) != 0) {
     free(query);
-    wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
+    _wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
     return user;
   }
   if (!(res = mysql_store_result(&mysql))) {
     free(query);
-    wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
+    _wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
     return user;
   }
 
@@ -458,16 +418,16 @@ wzd_user_t * FCN_GET_USER(int uid)
 
   /* Now get Groups */
 
-  snprintf(query, 512, "select groups.gid from groups,users where users.uid='%d' AND users.ref=groups.ref", uid);
+  snprintf(query, 512, "select groups.gid from groups,users,UGR where users.uid='%d' AND users.ref=UGR.uref AND groups.ref=UGR.gref", uid);
 
-  if (mysql_query(&mysql, query) != 0) { 
+  if (mysql_query(&mysql, query) != 0) {
     free(query);
-    wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
+    _wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
     return user;
   }
   if (!(res = mysql_store_result(&mysql))) {
     free(query);
-    wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
+    _wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
     return user;
   }
 
@@ -485,37 +445,35 @@ wzd_user_t * FCN_GET_USER(int uid)
 
   mysql_free_result(res);
 
-  /* FIXME */
-/*  strncpy(user->ip_allowed[0],"*",MAX_IP_LENGTH);*/
-  user->userperms = 0xffffffff;
-
   free(query);
 
   return user;
 }
 
-wzd_group_t * FCN_GET_GROUP(int gid)
+
+wzd_group_t * FCN_GET_GROUP(gid_t gid)
 {
   char *query;
   MYSQL_RES   *res;
   MYSQL_ROW    row;
   int num_fields;
   wzd_group_t * group;
+  unsigned int i;
 
-  if (gid == -2) return (wzd_group_t*)wzd_mysql_get_group_list();
+  if (gid == (gid_t)-2) return (wzd_group_t*)wzd_mysql_get_group_list();
 
   query = malloc(512);
   snprintf(query, 512, "SELECT * FROM groups WHERE gid='%d'", gid);
 
-  if (mysql_query(&mysql, query) != 0) { 
+  if (mysql_query(&mysql, query) != 0) {
     free(query);
-    wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
+    _wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
     return NULL;
   }
   free(query);
 
   if (!(res = mysql_store_result(&mysql))) {
-    wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
+    _wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
     return NULL;
   }
 
@@ -539,6 +497,14 @@ wzd_group_t * FCN_GET_GROUP(int gid)
     return NULL;
   }
   wzd_row_get_string(group->groupname, HARD_GROUPNAME_LENGTH, row, GCOL_GROUPNAME);
+  wzd_row_get_string(group->defaultpath, WZD_MAX_PATH, row, GCOL_DEFAULTPATH);
+  wzd_row_get_string(group->tagline, MAX_TAGLINE_LENGTH, row, GCOL_TAGLINE);
+  wzd_row_get_ulong(&group->groupperms, row, GCOL_GROUPPERMS);
+  wzd_row_get_uint((unsigned int*)&group->max_idle_time, row, GCOL_MAX_IDLE_TIME);
+  if (wzd_row_get_uint(&i, row, GCOL_NUM_LOGINS)==0) group->num_logins = i;
+  wzd_row_get_uint(&group->max_ul_speed, row, GCOL_MAX_UL_SPEED);
+  wzd_row_get_uint(&group->max_dl_speed, row, GCOL_MAX_DL_SPEED);
+  wzd_row_get_uint(&group->ratio, row, GCOL_RATIO);
 
   mysql_free_result(res);
 
@@ -552,7 +518,8 @@ wzd_group_t * FCN_GET_GROUP(int gid)
 
 
 /* basic syntax checking to avoid injections */
-static int wzd_mysql_check_name(const char *name)
+/** \todo XXX FIXME use mysql_real_escape_string() */
+int wzd_mysql_check_name(const char *name)
 {
   if (strpbrk(name,"'\";"))
     return 0;
@@ -619,38 +586,59 @@ static inline int wzd_row_get_ulong(unsigned long *dst, MYSQL_ROW row, unsigned
   return 1;
 }
 
+static inline int wzd_row_get_ullong(u64_t *dst, MYSQL_ROW row, unsigned int index)
+{
+  char *ptr;
+  u64_t i;
 
-static int * wzd_mysql_get_user_list(void)
+  if (!dst || !row || row[index]==NULL) return 1;
+
+  i = strtoull(row[index], &ptr, 0);
+  if (ptr && *ptr == '\0') {
+    *dst = i;
+    return 0;
+  }
+
+  return 1;
+}
+
+
+static uid_t * wzd_mysql_get_user_list(void)
 {
   char *query;
   MYSQL_RES   *res;
   MYSQL_ROW    row;
-  int * uid_list;
+  uid_t * uid_list;
   unsigned int index, i;
+  my_ulonglong num_rows;
 
   query = malloc(512);
   snprintf(query, 512, "SELECT uid FROM users");
 
   if (mysql_query(&mysql, query) != 0) {
     free(query);
-    wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
+    _wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
     return NULL;
   }
 
   if (!(res = mysql_store_result(&mysql))) {
     free(query);
-    wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
+    _wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
     return NULL;
   }
 
-  uid_list = (int*)wzd_malloc(HARD_DEF_USER_MAX*sizeof(int));
+  /* number of rows */
+  num_rows = mysql_num_rows(res);
+
+  uid_list = wzd_malloc(((u32_t)num_rows+1)*sizeof(uid_t));
 
   index = 0;
   while ( (row = mysql_fetch_row(res)) ) {
     wzd_row_get_uint(&i, row, 0 /* query asks only one column */);
-    uid_list[index++] = (int)i;
+    uid_list[index++] = (uid_t)i;
   }
-
+  uid_list[index] = (uid_t)-1;
+  uid_list[num_rows] = (uid_t)-1;
 
   mysql_free_result(res);
   free(query);
@@ -658,36 +646,42 @@ static int * wzd_mysql_get_user_list(void)
   return uid_list;
 }
 
-static int * wzd_mysql_get_group_list(void)
+static gid_t * wzd_mysql_get_group_list(void)
 {
   char *query;
   MYSQL_RES   *res;
   MYSQL_ROW    row;
-  int * gid_list;
+  gid_t * gid_list;
   unsigned int index, i;
+  my_ulonglong num_rows;
 
   query = malloc(512);
   snprintf(query, 512, "SELECT gid FROM groups");
 
   if (mysql_query(&mysql, query) != 0) {
     free(query);
-    wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
+    _wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
     return NULL;
   }
 
   if (!(res = mysql_store_result(&mysql))) {
     free(query);
-    wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
+    _wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
     return NULL;
   }
 
-  gid_list = (int*)wzd_malloc(HARD_DEF_USER_MAX*sizeof(int));
+  /* number of rows */
+  num_rows = mysql_num_rows(res);
+
+  gid_list = wzd_malloc(((u32_t)num_rows+1)*sizeof(gid_t));
 
   index = 0;
   while ( (row = mysql_fetch_row(res)) ) {
     wzd_row_get_uint(&i, row, 0 /* query asks only one column */);
-    gid_list[index++] = (int)i;
+    gid_list[index++] = (gid_t)i;
   }
+  gid_list[index] = (gid_t)-1;
+  gid_list[num_rows] = (gid_t)-1;
 
 
   mysql_free_result(res);
@@ -696,3 +690,54 @@ static int * wzd_mysql_get_group_list(void)
   return gid_list;
 }
 
+int _wzd_run_update_query(char * query, size_t length, const char * query_format, ...)
+{
+  MYSQL_RES   *res;
+  va_list argptr;
+
+  va_start(argptr, query_format);
+  vsnprintf(query, length, query_format, argptr);
+  va_end(argptr);
+
+  if (mysql_query(&mysql, query) != 0) {
+    free(query);
+    _wzd_mysql_error(__FILE__, __FUNCTION__, __LINE__);
+    return -1;
+  }
+
+  res = mysql_store_result(&mysql);
+
+  if (res) mysql_free_result(res);
+
+
+  return 0;
+}
+
+int wzd_backend_init(wzd_backend_t * backend)
+{
+  if (!backend) return -1;
+
+  backend->name = wzd_strdup("mysql");
+  backend->version = MYSQL_BACKEND_VERSION;
+
+  backend->backend_init = FCN_INIT;
+  backend->backend_exit = FCN_FINI;
+
+  backend->backend_validate_login = FCN_VALIDATE_LOGIN;
+  backend->backend_validate_pass = FCN_VALIDATE_PASS;
+
+  backend->backend_get_user = FCN_GET_USER;
+  backend->backend_get_group = FCN_GET_GROUP;
+
+  backend->backend_find_user = FCN_FIND_USER;
+  backend->backend_find_group = FCN_FIND_GROUP;
+
+  backend->backend_mod_user = FCN_MOD_USER;
+  backend->backend_mod_group = FCN_MOD_GROUP;
+
+  backend->backend_chpass = NULL;
+  backend->backend_commit_changes = FCN_COMMIT_CHANGES;
+
+  return 0;
+}
+
