@@ -62,7 +62,6 @@
 #include <perl.h>
 #include <XSUB.h>
 
-
 /*#include <wzd.h>*/
 #include "wzd_structs.h"
 #include "wzd_log.h"
@@ -72,11 +71,14 @@
 #include "wzd_file.h" /* file_mkdir, file_stat */
 #include "wzd_vfs.h" /* checkpath_new */
 #include "wzd_mod.h" /* essential to define WZD_MODULE_INIT */
+#include "wzd_string.h"
 #include "wzd_vars.h" /* needed to access variables */
 
 #include "wzd_debug.h"
 
 
+#define PERL_ERRORLOGNAME "perlerr.log"
+
 extern void boot_DynaLoader (pTHX_ CV* cv);
 
 
@@ -84,6 +86,8 @@ extern void boot_DynaLoader (pTHX_ CV* cv);
 static PerlInterpreter * my_perl=NULL;
 static wzd_context_t * current_context=NULL;
 
+static int perl_fd_errlog=-1;
+
 #define PERL_ARGS        "wzd::args"
 #define TCL_CURRENT_USER "wzd_current_user"
 #define TCL_REPLY_CODE  "wzd_reply_code"
@@ -91,7 +95,7 @@ static wzd_context_t * current_context=NULL;
 
 /***** Private fcts ****/
 static void do_perl_help(wzd_context_t * context);
-static int perl_init(void);
+static PerlInterpreter * perl_init(void);
 static int execute_perl( SV *function, const char *args);
 static void xs_init(pTHX);
 
@@ -104,7 +108,12 @@ static int perl_hook_protocol(const char *file, const char *args);
 
 /***** PERL commands ***/
 static XS(XS_wzd_test);
+static XS(XS_wzd_chgrp);
+static XS(XS_wzd_chmod);
+static XS(XS_wzd_chown);
 static XS(XS_wzd_ftp2sys);
+static XS(XS_wzd_killpath);
+static XS(XS_wzd_logperl);
 static XS(XS_wzd_putlog);
 static XS(XS_wzd_send_message_raw);
 static XS(XS_wzd_send_message);
@@ -129,15 +138,54 @@ static struct _slave_t _slaves[MAX_SLAVES];
 
 /***********************/
 MODULE_NAME(perl);
-MODULE_VERSION(101);
+MODULE_VERSION(102);
 
 /***********************/
 /* WZD_MODULE_INIT     */
 
 int WZD_MODULE_INIT(void)
 {
-  if (perl_init()) {
+  if (my_perl) /* init already done */
+    return -1;
+
+  {
+    char * logdir;
+    int ret;
+
+    ret = -1;
+    if (chtbl_lookup((CHTBL*)mainConfig->htab, "logdir", (void**)&logdir)== 0)
+    {
+      int fd;
+	  const char * filename;
+
+      wzd_string_t *str = str_allocate();
+      str_sprintf(str,"%s/%s", logdir, PERL_ERRORLOGNAME);
+	  filename = str_tochar(str);
+#ifndef WIN32
+      fd = open(filename,O_CREAT|O_WRONLY,S_IRUSR | S_IWUSR);
+#else
+	  /* activeperl redefines open(), and this causes a segfault here ! */
+      fd = _open(filename,O_CREAT|O_WRONLY,S_IRUSR | S_IWUSR);
+#endif
+      if (fd >= 0) {
+        perl_fd_errlog = fd;
+        ret = 0;
+      }
+      str_deallocate(str);
+    }
+    if (ret) {
+      out_log(LEVEL_HIGH,"perl: i found no 'logdir' in your config file\n");
+      out_log(LEVEL_HIGH,"perl: this means I will be unable to log PERL errors\n");
+      out_log(LEVEL_HIGH,"perl: please refer to the 'logdir' config directive in help\n");
+    }
+  }
+
+  if ( !(my_perl = perl_init()) ) {
     out_log(LEVEL_HIGH,"PERL could not create interpreter\n");
+    if (perl_fd_errlog >= 0) {
+      close(perl_fd_errlog);
+      perl_fd_errlog = -1;
+    }
     return -1;
   }
   memset(_slaves, 0, MAX_SLAVES*sizeof(struct _slave_t));
@@ -154,6 +202,10 @@ void WZD_MODULE_CLOSE(void)
   perl_destruct(my_perl);
   perl_free(my_perl);
   my_perl = NULL;
+  if (perl_fd_errlog >= 0) {
+    close(perl_fd_errlog);
+    perl_fd_errlog = -1;
+  }
   out_log(LEVEL_INFO,"PERL module unloaded\n");
 }
 
@@ -163,7 +215,7 @@ static wzd_hook_reply_t perl_hook_site(unsigned long event_id, wzd_context_t * c
 {
   SV *val;
 
-  if (strcasecmp(token,"perl")==0) {
+  if (strcasecmp(token,"site_perl")==0) {
     if (!my_perl) return 0;
     if (!args || strlen(args)==0) { do_perl_help(context); return EVENT_HANDLED; }
 
@@ -177,8 +229,15 @@ static wzd_hook_reply_t perl_hook_site(unsigned long event_id, wzd_context_t * c
 
     if (SvTRUE(val))
       send_message_with_args(200,context,"PERL command ok");
-    else
+    else {
+      /* log error */
+      if (perl_fd_errlog >= 0) {
+        wzd_string_t * str = str_allocate();
+        str_sprintf(str,"Error in %s: %s\n",args,SvPV_nolen(ERRSV));
+        write(perl_fd_errlog,str_tochar(str),strlen(str_tochar(str)));
+      }
       send_message_with_args(200,context,"PERL command reported errors");
+    }
     return EVENT_HANDLED;
   }
   return EVENT_IGNORED;
@@ -254,15 +313,15 @@ static void do_perl_help(wzd_context_t * context)
 
 
 
-static int perl_init(void)
+static PerlInterpreter * perl_init(void)
 {
   const char perl_definitions[] = {
 "\n"
 "$SIG{__WARN__} = sub {\n"
 "  local $, = \"\\n\";\n"
 "  my ($package, $line, $sub) = caller(1);\n"
-"  wzd::send_message( \"warning from ${package}::${sub} at line $line.\" );\n"
-"  wzd::send_message( @_ );\n"
+"  wzd::logperl( \"warning from ${package}::${sub} at line $line.\" );\n"
+"  wzd::logperl( @_ );\n"
 "};\n"
 "\n"
 "sub Embed::load {\n"
@@ -276,13 +335,13 @@ static int perl_init(void)
 "\n"
 "	 if( $@ ) {\n"
 "		# something went wrong\n"
-"		print( \"Error loading '$file':\\n$@\n\" );\n"
+"		wzd::logperl( \"Error loading '$file':\\n$@\n\" );\n"
 "		return 1;\n"
 "	 }\n"
 "\n"
 "  } else {\n"
 "\n"
-"	 print( \"Error opening '$file': $!\\n\" );\n"
+"	 wzd::logperl( \"Error opening '$file': $!\\n\" );\n"
 "	 return 2;\n"
 "  }\n"
 "\n"
@@ -291,23 +350,21 @@ static int perl_init(void)
   };
   
   char * perl_args[] = { "", "-e", "0", "-w" };
+  PerlInterpreter * interp = NULL;
 
-  if (my_perl) /* init already done ! */
-    return -1;
-
-  my_perl = perl_alloc();
-  if (!my_perl) return -1;
-  perl_construct(my_perl);
+  interp = perl_alloc();
+  if (!interp) return NULL;
+  perl_construct(interp);
 
   /* set to 4 to have warnings */
-  perl_parse(my_perl, xs_init, 3, perl_args, NULL);
+  perl_parse(interp, xs_init, 3, perl_args, NULL);
 
   /* now initialize the perl interpreter by loading the
    * perl_definitions array.
    */
   eval_pv(perl_definitions, TRUE);
 
-  return 0;
+  return interp;
 }
 
 static void xs_init(pTHX)
@@ -320,7 +377,12 @@ static void xs_init(pTHX)
   newXS("DynaLoader::boot_DynaLoader", boot_DynaLoader, file);
 
   newXS("wzd::test", XS_wzd_test, "wzd");
+  newXS("wzd::chgrp", XS_wzd_chgrp, "wzd");
+  newXS("wzd::chmod", XS_wzd_chmod, "wzd");
+  newXS("wzd::chown", XS_wzd_chown, "wzd");
   newXS("wzd::ftp2sys", XS_wzd_ftp2sys, "wzd");
+  newXS("wzd::killpath", XS_wzd_killpath, "wzd");
+  newXS("wzd::logperl", XS_wzd_logperl, "wzd");
   newXS("wzd::putlog", XS_wzd_putlog, "wzd");
   newXS("wzd::send_message_raw", XS_wzd_send_message_raw, "wzd");
   newXS("wzd::send_message", XS_wzd_send_message, "wzd");
@@ -350,7 +412,12 @@ static int execute_perl( SV *function, const char *args)
 
   sv = GvSV(gv_fetchpv("@", TRUE, SVt_PV));
   if (SvTRUE(sv)) {
-    /* perl error */
+    /* perl error, log it */
+    if (perl_fd_errlog >= 0) {
+      wzd_string_t * str = str_allocate();
+      str_sprintf(str,"Error in %s: %s\n",args,SvPV_nolen(ERRSV));
+      write(perl_fd_errlog,str_tochar(str),strlen(str_tochar(str)));
+    }
     POPs; /* remove undef from the top of the stack */
   }
   else if (count != 1) {
@@ -391,7 +458,9 @@ static int _perl_set_slave(void *context)
 
   current_context = context;
   if (found) {
+#ifdef USE_ITHREADS
     PERL_SET_CONTEXT(_slaves[i].interp);
+#endif
     return 0;
   }
 
@@ -400,6 +469,7 @@ static int _perl_set_slave(void *context)
   {
     if ( ! _slaves[i].is_allocated )
     {
+#ifdef USE_ITHREADS
       PERL_SET_CONTEXT(my_perl);
 
       _slaves[i].is_allocated = 1;
@@ -412,6 +482,11 @@ static int _perl_set_slave(void *context)
       /* see perlapi (1) for more info, this flag is needed for win32 */
       
       PERL_SET_CONTEXT(_slaves[i].interp);
+#else /* USE_ITHREADS */
+      _slaves[i].is_allocated = 1;
+      _slaves[i].context = context;
+      _slaves[i].interp = perl_init();
+#endif /* USE_ITHREADS */
 
       return 0;
     }
@@ -438,6 +513,126 @@ static XS(XS_wzd_test)
   XSRETURN_EMPTY;
 }
 
+static XS(XS_wzd_chgrp)
+{
+  char path[WZD_MAX_PATH+1];
+  char * groupname=NULL;
+  char * text;
+  int index;
+  dXSARGS;
+
+  if (!current_context) XSRETURN_UNDEF;
+  if (items < 2) XSRETURN_UNDEF;
+
+  /** \todo print error message */
+  if ( ! SvPOK(ST(0)) )
+    XSRETURN_UNDEF;
+  if ( ! SvPOK(ST(1)) )
+    XSRETURN_UNDEF;
+
+  groupname = SvPV_nolen(ST(0));
+
+  index = 1;
+  text = SvPV_nolen(ST(index));
+
+  if ( checkpath_new(text, path, current_context) ) {
+    out_log(LEVEL_INFO,"perl wzd::chgrp could not resolv path %s\n",text);
+    XSRETURN_UNDEF;
+  }
+  if (file_chown(path,NULL,groupname,current_context)) {
+    XSRETURN_NO;
+  }
+
+  XSRETURN_YES;
+}
+
+static XS(XS_wzd_chmod)
+{
+  char path[WZD_MAX_PATH+1];
+  char * text;
+  char * endptr;
+  int index;
+  char * mode;
+  unsigned long perms;
+  dXSARGS;
+
+  if (!current_context) XSRETURN_UNDEF;
+  if (items < 2) XSRETURN_UNDEF;
+
+  /** \todo print error message */
+  if ( ! (SvPOK(ST(0)) || SvIOK(ST(0))) )
+    XSRETURN_UNDEF;
+  if ( ! SvPOK(ST(1)) )
+    XSRETURN_UNDEF;
+
+  mode = SvPV_nolen(ST(0));
+  perms = strtoul(mode,&endptr,8);
+  if (endptr == mode) {
+    /** TODO try to convert from string rwxr-xr-x */
+    out_log(LEVEL_INFO,"perl wzd::chmod could not convert mode %s to octal number\n",mode);
+    XSRETURN_UNDEF;
+  }
+
+  index = 1;
+  text = SvPV_nolen(ST(index));
+
+  if ( checkpath_new(text, path, current_context) ) {
+    out_log(LEVEL_INFO,"perl wzd::chmod could not resolv path %s\n",text);
+    XSRETURN_UNDEF;
+  }
+  if (_setPerm(path,NULL,NULL,NULL,NULL,perms,current_context)) {
+    XSRETURN_NO;
+  }
+
+  XSRETURN_YES;
+}
+
+static XS(XS_wzd_chown)
+{
+  char path[WZD_MAX_PATH+1];
+  char * username=NULL, * groupname=NULL;
+  char * text;
+  char * ptr;
+  int index;
+  dXSARGS;
+
+  if (!current_context) XSRETURN_UNDEF;
+  if (items < 2) XSRETURN_UNDEF;
+
+  /** \todo print error message */
+  if ( ! SvPOK(ST(0)) )
+    XSRETURN_UNDEF;
+  if ( ! SvPOK(ST(1)) )
+    XSRETURN_UNDEF;
+
+  username = SvPV_nolen(ST(0));
+  if ((ptr = strchr(username,':'))) {
+    groupname = ptr+1;
+    if (ptr == username) { /* chown :group file */
+      username = NULL;
+    }
+    else {
+      *ptr = '\0';
+      /* from man chmod: If a colon or dot but no group name follows the user name, that user is made
+       * the owner of the files and the group of the files is  changed  to  that  user’s login group.
+       */
+    }
+  }
+
+  index = 1;
+  text = SvPV_nolen(ST(index));
+
+  if ( checkpath_new(text, path, current_context) ) {
+    out_log(LEVEL_INFO,"perl wzd::chown could not resolv path %s\n",text);
+    XSRETURN_UNDEF;
+  }
+  if (file_chown(path,username,groupname,current_context)) {
+    XSRETURN_NO;
+  }
+
+  XSRETURN_YES;
+}
+
 static XS(XS_wzd_ftp2sys)
 {
   char path[WZD_MAX_PATH+1];
@@ -460,6 +655,84 @@ static XS(XS_wzd_ftp2sys)
   XSRETURN_PV(path);
 }
 
+static XS(XS_wzd_killpath)
+{
+  char * text;
+  int ret;
+  dXSARGS;
+
+  if (!current_context) XSRETURN_UNDEF;
+  if (items < 1) XSRETURN_UNDEF;
+
+  /** \todo print error message */
+  if ( ! SvPOK(ST(0)) )
+    XSRETURN_UNDEF;
+
+  text = SvPV_nolen(ST(0));
+
+  if (!strcmp(text,"-r") || !strcmp(text,"--real")) {
+    /* ex: killpath -r c:\real */
+    if (items < 2) XSRETURN_UNDEF;
+
+    /** \todo print error message */
+    if ( ! SvPOK(ST(1)) )
+      XSRETURN_UNDEF;
+
+    text = SvPV_nolen(ST(1));
+
+    ret = killpath (text,current_context);
+  } else {
+    char * realpath;
+    realpath = malloc(WZD_MAX_PATH+1);
+    if ( checkpath(text, realpath, current_context) ) {
+      XSRETURN_UNDEF;
+    }
+    ret = killpath (realpath,current_context);
+    free(realpath);
+  }
+  if ( ret != E_OK && ret != E_USER_NOBODY ) {
+    XSRETURN_NO;
+  }
+
+  XSRETURN_YES;
+}
+
+/**
+ * example: wzd::logperl("message\n");
+ *  or    : wzd::logperl "message\n";
+ */
+static XS(XS_wzd_logperl)
+{
+  char * text;
+  int index=0;
+  STRLEN length;
+
+  dXSARGS;
+
+  if (!current_context) XSRETURN_NO;
+  if (items < 1) XSRETURN_NO;
+
+  /** \todo print error message */
+  if ( ! SvPOK(ST(0)) )
+    XSRETURN_NO;
+
+  if (perl_fd_errlog >= 0) {
+    while (SvPOK(ST(index))) {
+      text = SvPV(ST(index),length);
+
+      /** \todo XXX we could format the string using argv[2,] */
+
+      /* replace cookies ? */
+
+      write(perl_fd_errlog,text,length);
+
+      index++;
+    }
+  }
+
+  XSRETURN_YES;
+}
+
 /**
  * example: wzd::putlog(5,"message\n");
  *  or    : wzd::putlog 5,"message\n";
@@ -692,6 +965,13 @@ static XS(XS_wzd_vars_group)
       XSRETURN_PV(buffer);
     else
       XSRETURN_UNDEF;
+  } else if (!strcmp(command,"new")) { /* new user creation */
+    ret = vars_group_new(groupname,getlib_mainConfig());
+    /** \todo handle return */
+    if (!ret)
+      XSRETURN_PV("command ok");
+    else
+      XSRETURN_UNDEF;
   }
 
   XSRETURN_UNDEF;
