@@ -39,6 +39,7 @@
 
 #include <stdio.h>
 #include <string.h>
+#include <stdlib.h>
 
 #ifdef _MSC_VER
 #include <winsock2.h>
@@ -68,11 +69,11 @@
 #include "wzd_structs.h"
 #include "wzd_log.h"
 #include "wzd_misc.h"
+#include "wzd_file.h" /* file_mkdir, file_stat */
 #include "wzd_libmain.h"
 #include "wzd_messages.h"
-#include "wzd_file.h" /* file_mkdir, file_stat */
-#include "wzd_vfs.h" /* checkpath_new */
 #include "wzd_mod.h" /* essential to define WZD_MODULE_INIT */
+#include "wzd_vfs.h" /* checkpath_new */
 #include "wzd_vars.h" /* needed to access variables */
 
 #include "wzd_debug.h"
@@ -81,15 +82,49 @@
 static Tcl_Interp * interp=NULL;
 static wzd_context_t * current_context=NULL;
 
+static int tcl_fd_errlog=-1;
+
 #define TCL_ARGS        "wzd_args"
 #define TCL_CURRENT_USER "wzd_current_user"
 #define TCL_REPLY_CODE  "wzd_reply_code"
 #define TCL_HAS_REPLIED "wzd_replied"
+#define TCL_ERRORLOGNAME "tclerr.log"
+
+#define WZDOUT  ((ClientData)1)
+#define WZDERR  ((ClientData)2)
 
 /***** Private fcts ****/
 static void do_tcl_help(wzd_context_t * context);
 static int tcl_diagnose(void);
 
+
+static int channel_close(ClientData instance, Tcl_Interp *interp);
+static int channel_input(ClientData instance, char *buf, int bufsiz, int *errptr);
+static int channel_output(ClientData instance, const char *buf, int bufsiz, int *errptr);
+static void channel_watch(ClientData instance, int mask);
+static int channel_gethandle(ClientData instance, int direction, ClientData *handleptr);
+
+/***** Private structs ****/
+
+static Tcl_ChannelType channel_type =
+{
+  "wzdmessage",
+  TCL_CHANNEL_VERSION_2,
+  channel_close,
+  channel_input,
+  channel_output,
+  NULL,   /* seek */
+  NULL,   /* set option */
+  NULL,   /* get option */
+  channel_watch,
+  channel_gethandle,
+  NULL,   /* close2 */
+  NULL,   /* block */
+  NULL,   /* flush */
+  NULL,   /* handler */
+  NULL,   /* wideseek */
+};
+
 /***** EVENT HOOKS *****/
 static wzd_hook_reply_t tcl_hook_site(unsigned long event_id, wzd_context_t * context, const char *token, const char *args);
 static int tcl_hook_logout(unsigned long event_id, wzd_context_t *context, const char *username);
@@ -98,7 +133,11 @@ static int tcl_hook_logout(unsigned long event_id, wzd_context_t *context, const
 static int tcl_hook_protocol(const char *file, const char *args);
 
 /***** TCL commands ****/
+static int tcl_chgrp(ClientData data, Tcl_Interp *interp, int argc, const char *argv[]);
+static int tcl_chmod(ClientData data, Tcl_Interp *interp, int argc, const char *argv[]);
+static int tcl_chown(ClientData data, Tcl_Interp *interp, int argc, const char *argv[]);
 static int tcl_ftp2sys(ClientData data, Tcl_Interp *interp, int argc, const char *argv[]);
+static int tcl_killpath(ClientData data, Tcl_Interp *interp, int argc, const char *argv[]);
 static int tcl_putlog(ClientData data, Tcl_Interp *interp, int argc, const char *argv[]);
 static int tcl_send_message(ClientData data, Tcl_Interp *interp, int argc, const char *argv[]);
 static int tcl_send_message_raw(ClientData data, Tcl_Interp *interp, int argc, const char *argv[]);
@@ -115,13 +154,14 @@ static Tcl_Interp * _tcl_getslave(Tcl_Interp *interp, void *context);
 
 /***********************/
 MODULE_NAME(tcl);
-MODULE_VERSION(105);
+MODULE_VERSION(106);
 
 /***********************/
 /* WZD_MODULE_INIT     */
 
 int WZD_MODULE_INIT(void)
 {
+  static Tcl_Channel ch1, ch2;
 #ifdef _MSC_VER
   {
     char buffer[MAX_PATH+1];
@@ -149,7 +189,55 @@ int WZD_MODULE_INIT(void)
     out_log(LEVEL_HIGH,"TCL could not create interpreter\n");
     return -1;
   }
+
+  {
+    char * logdir;
+    int ret;
+
+    ret = -1;
+    if (chtbl_lookup((CHTBL*)mainConfig->htab, "logdir", (void**)&logdir)== 0)
+    {
+      int fd;
+
+      wzd_string_t *str = str_allocate();
+      str_sprintf(str,"%s/%s", logdir, TCL_ERRORLOGNAME);
+      fd = open(str_tochar(str),O_CREAT|O_WRONLY,S_IRUSR | S_IWUSR);
+      if (fd >= 0) {
+        tcl_fd_errlog = fd;
+        ret = 0;
+      }
+      str_deallocate(str);
+    }
+    if (ret) {
+      out_log(LEVEL_HIGH,"tcl: i found no 'logdir' in your config file\n");
+      out_log(LEVEL_HIGH,"tcl: this means I will be unable to log TCL errors\n");
+      out_log(LEVEL_HIGH,"tcl: please refer to the 'logdir' config directive in help\n");
+    }
+  }
+
+  /* replace stdout and stderr */
+  ch1 = Tcl_CreateChannel(&channel_type, "wzdout", WZDOUT, TCL_WRITABLE);
+  ch2 = Tcl_CreateChannel(&channel_type, "wzderr", WZDERR, TCL_WRITABLE);
+  Tcl_SetStdChannel(ch1, TCL_STDOUT);
+  Tcl_SetStdChannel(ch2, TCL_STDERR);
+
+  Tcl_SetChannelOption(interp, ch1, "-buffering", "line");
+  Tcl_SetChannelOption(interp, ch2, "-buffering", "line");
+
+  Tcl_RegisterChannel(interp, ch1);
+  Tcl_RegisterChannel(interp, ch2);
+
+  /** It's a bit stupid to modify things here, because modifications (like changing
+   * standard channels) are NOT inherited by slaves.
+   */
+
+
+
+  Tcl_CreateCommand(interp,"chgrp",tcl_chgrp,(ClientData)NULL,(Tcl_CmdDeleteProc*)NULL);
+  Tcl_CreateCommand(interp,"chmod",tcl_chmod,(ClientData)NULL,(Tcl_CmdDeleteProc*)NULL);
+  Tcl_CreateCommand(interp,"chown",tcl_chown,(ClientData)NULL,(Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand(interp,"ftp2sys",tcl_ftp2sys,(ClientData)NULL,(Tcl_CmdDeleteProc*)NULL);
+  Tcl_CreateCommand(interp,"killpath",tcl_killpath,(ClientData)NULL,(Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand(interp,"putlog",tcl_putlog,(ClientData)NULL,(Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand(interp,"send_message",tcl_send_message,(ClientData)NULL,(Tcl_CmdDeleteProc*)NULL);
   Tcl_CreateCommand(interp,"send_message_raw",tcl_send_message_raw,(ClientData)NULL,(Tcl_CmdDeleteProc*)NULL);
@@ -171,6 +259,11 @@ void WZD_MODULE_CLOSE(void)
   Tcl_DeleteInterp(interp);
   interp = NULL;
 /*  Tcl_Exit(0);*/
+  Tcl_Finalize();
+  if (tcl_fd_errlog >= 0) {
+    close(tcl_fd_errlog);
+    tcl_fd_errlog = -1;
+  }
   out_log(LEVEL_INFO,"TCL module unloaded\n");
 }
 
@@ -178,12 +271,13 @@ void WZD_MODULE_CLOSE(void)
 
 static wzd_hook_reply_t tcl_hook_site(unsigned long event_id, wzd_context_t * context, const char *token, const char *args)
 {
-  if (strcasecmp(token,"tcl")==0) {
+  if (strcasecmp(token,"site_tcl")==0) {
     if (!args || strlen(args)==0) { do_tcl_help(context); return EVENT_HANDLED; }
     {
       Tcl_Obj * TempObj;
       Tcl_Interp * slave = NULL;
       const char *s;
+      char * errorinfo;
       wzd_user_t * user;
       int ret;
 
@@ -204,9 +298,11 @@ static wzd_hook_reply_t tcl_hook_site(unsigned long event_id, wzd_context_t * co
       current_context = NULL;
       s = Tcl_GetVar(slave,TCL_HAS_REPLIED,TCL_GLOBAL_ONLY);
       if (!s || *s!='1') {
-        if (ret != TCL_OK)
+        if (ret != TCL_OK) {
+          errorinfo = (char *) Tcl_GetVar(interp, "errorInfo", 0);
+          out_err(LEVEL_HIGH,"TCL error: %s\n",errorinfo);
           send_message_with_args(200,context,"Error in TCL command");
-        else
+        } else
           send_message_with_args(200,context,"TCL command ok");
       }
     }
@@ -367,7 +463,25 @@ static Tcl_Interp * _tcl_getslave(Tcl_Interp *interp, void *context)
 
   if ( (slave = Tcl_CreateSlave(interp, buffer, 0)) ) {
     int ret;
+
+    static Tcl_Channel ch1, ch2; /** \bug why static ?! */
+    /* replace stdout and stderr */
+    ch1 = Tcl_CreateChannel(&channel_type, "wzdout", WZDOUT, TCL_WRITABLE);
+    ch2 = Tcl_CreateChannel(&channel_type, "wzderr", WZDERR, TCL_WRITABLE);
+    Tcl_SetStdChannel(ch1, TCL_STDOUT);
+    Tcl_SetStdChannel(ch2, TCL_STDERR);
+
+    Tcl_SetChannelOption(slave, ch1, "-buffering", "line");
+    Tcl_SetChannelOption(slave, ch2, "-buffering", "line");
+
+    Tcl_RegisterChannel(slave, ch1);
+    Tcl_RegisterChannel(slave, ch2);
+
+    ret = Tcl_CreateAlias(slave, "chgrp", interp, "chgrp", 0, NULL);
+    ret = Tcl_CreateAlias(slave, "chmod", interp, "chmod", 0, NULL);
+    ret = Tcl_CreateAlias(slave, "chown", interp, "chown", 0, NULL);
     ret = Tcl_CreateAlias(slave, "ftp2sys", interp, "ftp2sys", 0, NULL);
+    ret = Tcl_CreateAlias(slave, "killpath", interp, "killpath", 0, NULL);
     ret = Tcl_CreateAlias(slave, "putlog", interp, "putlog", 0, NULL);
     ret = Tcl_CreateAlias(slave, "send_message", interp, "send_message", 0, NULL);
     ret = Tcl_CreateAlias(slave, "send_message_raw", interp, "send_message_raw", 0, NULL);
@@ -387,6 +501,89 @@ static Tcl_Interp * _tcl_getslave(Tcl_Interp *interp, void *context)
 
 /******* TCL functions ********/
 
+static int tcl_chgrp(ClientData data, Tcl_Interp *interp, int argc, const char *argv[])
+{
+  char path[WZD_MAX_PATH+1];
+  const char * groupname;
+
+  if (argc < 3) return TCL_ERROR;
+  if (!current_context) return TCL_ERROR;
+
+  groupname = argv[1];
+
+  if ( checkpath_new(argv[2], path, current_context) ) {
+    out_log(LEVEL_INFO,"tcl chgrp could not resolv path %s\n",argv[1]);
+    return TCL_ERROR;
+  }
+  if (file_chown(path,NULL,groupname,current_context)) {
+    return TCL_ERROR;
+  }
+
+  return TCL_OK;
+}
+
+static int tcl_chmod(ClientData data, Tcl_Interp *interp, int argc, const char *argv[])
+{
+  char path[WZD_MAX_PATH+1];
+  char * endptr;
+  unsigned long perms;
+
+  if (argc < 3) return TCL_ERROR;
+  if (!current_context) return TCL_ERROR;
+
+  perms = strtoul(argv[1],&endptr,8);
+  if (endptr == argv[1]) {
+    /** TODO try to convert from string rwxr-xr-x */
+    out_log(LEVEL_INFO,"tcl chmod could not convert mode %s to octal number\n",argv[1]);
+    return TCL_ERROR;
+  }
+
+  if ( checkpath_new(argv[2], path, current_context) ) {
+    out_log(LEVEL_INFO,"tcl chmod could not resolv path %s\n",argv[1]);
+    return TCL_ERROR;
+  }
+  if (_setPerm(path,NULL,NULL,NULL,NULL,perms,current_context)) {
+    return TCL_ERROR;
+  }
+
+  return TCL_OK;
+}
+
+static int tcl_chown(ClientData data, Tcl_Interp *interp, int argc, const char *argv[])
+{
+  char path[WZD_MAX_PATH+1];
+  const char * username=NULL, * groupname=NULL;
+  const char * ptr;
+
+  if (argc < 3) return TCL_ERROR;
+  if (!current_context) return TCL_ERROR;
+
+  username = argv[1];
+
+  if ((ptr = strchr(username,':'))) {
+    groupname = ptr+1;
+    if (ptr == username) { /* chown :group file */
+      username = NULL;
+    }
+    else {
+      *(char*)ptr = '\0'; /* yes, we're changing read-only arguments .. */
+      /* from man chmod: If a colon or dot but no group name follows the user name, that user is made
+       * the owner of the files and the group of the files is  changed  to  that  user’s login group.
+       */
+    }
+  }
+
+  if ( checkpath_new(argv[2], path, current_context) ) {
+    out_log(LEVEL_INFO,"tcl chown could not resolv path %s\n",argv[1]);
+    return TCL_ERROR;
+  }
+  if (file_chown(path,username,groupname,current_context)) {
+    return TCL_ERROR;
+  }
+
+  return TCL_OK;
+}
+
 static int tcl_ftp2sys(ClientData data, Tcl_Interp *interp, int argc, const char *argv[])
 {
   char *path;
@@ -404,6 +601,34 @@ static int tcl_ftp2sys(ClientData data, Tcl_Interp *interp, int argc, const char
   return TCL_OK;
 }
 
+static int tcl_killpath(ClientData data, Tcl_Interp *interp, int argc, const char *argv[])
+{
+  int ret;
+
+  if (argc < 2) return TCL_ERROR;
+  if (!current_context) return TCL_ERROR;
+
+  if (!strcmp(argv[1],"-r") || !strcmp(argv[1],"--real")) {
+    ret = killpath(argv[2], current_context);
+  } else {
+    char * realpath;
+    realpath = malloc(WZD_MAX_PATH+1);
+
+    if (checkpath_new(argv[2],realpath,current_context)) {
+      free(realpath);
+      return TCL_ERROR;
+    }
+    ret = killpath(realpath, current_context);
+    free(realpath);
+  }
+
+  if ( ret != E_OK && ret != E_USER_NOBODY ) {
+    return TCL_ERROR;
+  }
+
+  return TCL_OK;
+}
+
 static int tcl_putlog(ClientData data, Tcl_Interp *interp, int argc, const char *argv[])
 {
   char *ptr;
@@ -555,11 +780,11 @@ static int tcl_vars_group(ClientData data, Tcl_Interp *interp, int argc, const c
   } else if (!strcmp(argv[1],"set")) {
     ret = vars_group_set(argv[2],argv[3],(void*)argv[4],1024,getlib_mainConfig());
     return (ret)?TCL_ERROR:TCL_OK;
-#if 0
-  } else if (!strcmp(argv[1],"new")) { /* new user creation */
-    ret = vars_group_new(argv[2],argv[3],argv[4],getlib_mainConfig());
+  } else if (!strcmp(argv[1],"new")) { /* new group creation */
+    ret = vars_group_new(argv[2],getlib_mainConfig());
     /** \todo handle return */
     return (ret)?TCL_ERROR:TCL_OK;
+#if 0
   } else if (!strcmp(argv[1],"addip")) { /* add new ip */
     ret = vars_group_addip(argv[2],argv[3],getlib_mainConfig());
     /** \todo handle return */
@@ -571,7 +796,7 @@ static int tcl_vars_group(ClientData data, Tcl_Interp *interp, int argc, const c
 #endif
   }
 
-  return TCL_OK;
+  return TCL_ERROR;
 }
 
 static int tcl_vars_shm(ClientData data, Tcl_Interp *interp, int argc, const char *argv[])
@@ -736,6 +961,76 @@ static int tcl_vfs(ClientData data, Tcl_Interp *interp, int argc, const char *ar
   }
   else
     ret = TCL_ERROR;
-  
+
   return (ret)?TCL_ERROR:TCL_OK;
 }
+
+/******* I/O Channel ********/
+
+static int channel_close(ClientData instance, Tcl_Interp *interp)
+{
+  int err=0;
+
+  /* currently does nothing */
+  if (instance != WZDOUT && instance != WZDERR)
+  {
+    Tcl_SetErrno(EBADF);
+    err = EBADF;
+  }
+  return err;
+}
+
+static int channel_input(ClientData instance, char *buf, int bufsiz, int *errptr)
+{
+  /* input is currently not supported */
+
+  Tcl_SetErrno(EINVAL);
+  if (errptr)
+    *errptr = EINVAL;
+  return -1;
+}
+
+static int channel_output(ClientData instance, const char *buf, int bufsiz, int *errptr)
+{
+  char * str;
+  int result;
+
+  /* buf is not guaranteed to be 0-terminated */
+  str = malloc (bufsiz+1);
+  if (!str) {
+    Tcl_SetErrno(ENOMEM);
+    if (errptr) *errptr = ENOMEM;
+    return -1;
+  }
+  strncpy(str, buf, bufsiz);
+  str[bufsiz] = '\0';
+
+  result = bufsiz;
+  if (instance == WZDOUT)
+    out_err(LEVEL_INFO,"tcl OUT: [%s]\n", str);
+  else if (instance == WZDERR) {
+    out_err(LEVEL_HIGH,"tcl ERR: [%s]\n", str);
+    if (tcl_fd_errlog >= 0)
+      write(tcl_fd_errlog, str, bufsiz);
+  }
+  else {
+    Tcl_SetErrno(EBADF);
+    if (errptr) *errptr = EBADF;
+    result = -1;
+  }
+
+  free(str);
+  return result;
+}
+
+static void channel_watch(ClientData instance, int mask)
+{
+  Tcl_SetErrno(EINVAL);
+}
+
+static int channel_gethandle(ClientData instance, int direction, ClientData *handleptr)
+{
+  Tcl_SetErrno(EINVAL);
+  return EINVAL;
+}
+
