@@ -29,6 +29,10 @@
    security reasons
  */
 
+#include "wzd_all.h"
+
+#ifndef WZD_USE_PCH
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -59,17 +63,20 @@
 
 #include "wzd_debug.h"
 
+#endif /* WZD_USE_PCH */
+
 static int _format_date(time_t time, char * buffer, size_t length);
 
 int list_match(char *,char *);
 
-int list_call_wrapper(unsigned int sock, wzd_context_t *context, char *line, char *buffer, size_t *buffer_len,
-    int callback(unsigned int,wzd_context_t*,char *))
+int list_call_wrapper(fd_t sock, wzd_context_t *context, char *line, char *buffer, size_t *buffer_len,
+    int callback(fd_t,wzd_context_t*,char *))
 {
   size_t length;
   if (!line) { /* request to flush */
 /*out_err(LEVEL_CRITICAL,"Flushing buffer (%ld bytes)\n",*buffer_len);*/
-    if (!callback(sock,context,buffer)) return 1;
+    if (buffer && buffer[0]!='\0')
+      if (!callback(sock,context,buffer)) return 1;
     return 0;
   }
   length = strlen(line);
@@ -89,8 +96,8 @@ int list_call_wrapper(unsigned int sock, wzd_context_t *context, char *line, cha
 
 
 
-int list(unsigned int sock,wzd_context_t * context,list_type_t format,char *directory,char *mask,
-	 int callback(unsigned int,wzd_context_t*,char *))
+int list(fd_t sock,wzd_context_t * context,list_type_t format,char *directory,char *mask,
+	 int callback(fd_t,wzd_context_t*,char *))
 {
   struct wzd_dir_t * dir;
   struct wzd_file_t * file;
@@ -103,8 +110,8 @@ int list(unsigned int sock,wzd_context_t * context,list_type_t format,char *dire
   char datestr[128];
   char * buffer_ptr;
   size_t length;
-  struct stat st;
   unsigned long watchdog=0;
+  struct statbuf st;
 
   if (!directory || strlen(directory)<1) return 0;
 
@@ -161,10 +168,10 @@ int list(unsigned int sock,wzd_context_t * context,list_type_t format,char *dire
         break;
     }
 
-    if (lstat(ptr_to_buffer,&st)) {
+    if (fs_lstat(ptr_to_buffer,&st)) {
       /* destination does not exist */
       out_log(LEVEL_FLOOD, "list: broken file %s -> %s\n", file->filename, ptr_to_buffer);
-      memset(&st, 0, sizeof(struct stat));
+      memset(&st, 0, sizeof(st));
       st.st_mode = S_IFREG;
     };
 
@@ -177,7 +184,7 @@ int list(unsigned int sock,wzd_context_t * context,list_type_t format,char *dire
       !S_ISREG(st.st_mode)) {
       /* destination does not exist */
       out_log(LEVEL_FLOOD, "list: strange file %s\n", file->filename);
-      memset(&st, 0, sizeof(struct stat));
+      memset(&st, 0, sizeof(st));
     };
 
     if (S_ISLNK(st.st_mode)) {
@@ -254,8 +261,8 @@ int list(unsigned int sock,wzd_context_t * context,list_type_t format,char *dire
 /** build list in LIST format and send the result to callback
  * \deprecated use \ref list
  */
-int old_list(unsigned int sock,wzd_context_t * context,list_type_t format,char *directory,char *mask,
-	 int callback(unsigned int,wzd_context_t*,char *)) {
+int old_list(fd_t sock,wzd_context_t * context,list_type_t format,char *directory,char *mask,
+	 int callback(fd_t,wzd_context_t*,char *)) {
 
 #ifndef _MSC_VER
   DIR *dir;
@@ -278,7 +285,7 @@ int old_list(unsigned int sock,wzd_context_t * context,list_type_t format,char *
   size_t dirlen;
   int i;
   time_t timeval;
-  struct stat st;
+  struct statbuf st;
   struct tm *ntime;
   wzd_user_t * user, * owner;
   short vfs_pad=0;
@@ -330,7 +337,7 @@ int old_list(unsigned int sock,wzd_context_t * context,list_type_t format,char *
       {
         char * ptr = buffer_vfs + strlen(directory) + vfs_pad;
         if (strchr(ptr,'/')==NULL) {
-          if (stat(vfs->physical_dir,&st)<0) {
+          if (fs_stat(vfs->physical_dir,&st)<0) {
             vfs = vfs->next_vfs;
             continue;
           }
@@ -468,7 +475,7 @@ int old_list(unsigned int sock,wzd_context_t * context,list_type_t format,char *
           continue;
         }
         strcpy(filename+dirlen,dir_filename);
-        if (lstat(filename,&st)<0)
+        if (fs_lstat(filename,&st)<0)
         {
 #ifdef _MSC_VER
           if (!FindNextFile(dir,&fileData))
