@@ -24,6 +24,10 @@
  * the source code for OpenSSL in the source distribution.
  */
 
+#include "wzd_all.h"
+
+#ifndef WZD_USE_PCH
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -58,13 +62,13 @@
 
 #include "wzd_debug.h"
 
-/* BSD exports symbols in .so files prefixed with a _ !! */
-#ifdef BSD
+#ifdef NEED_UNDERSCORE
 #define	DL_PREFIX "_"
 #else
 #define	DL_PREFIX
 #endif
 
+#endif /* WZD_USE_PCH */
 
 static int _backend_check_function(void * handle, const char *name, const char *backend_name);
 
@@ -72,37 +76,38 @@ static int _trigger_user_max_dl(wzd_user_t * user);
 static int _trigger_user_max_ul(wzd_user_t * user);
 
 
-char *backend_get_version(wzd_backend_t *backend)
+char *backend_get_version(wzd_backend_def_t *backend)
 {
   char ** version_found;
 
   if (backend->handle)
-    version_found = (char**)dlsym(backend->handle,DL_PREFIX "backend_version");
+    version_found = (char**)dlsym(backend->handle,DL_PREFIX "wzd_backend_version");
   else
     return NULL;
 
   return strdup(*version_found);
 }
 
-char *backend_get_name(wzd_backend_t *backend)
+char *backend_get_name(wzd_backend_def_t *backend)
 {
   char ** backend_name;
 
   if (backend->handle)
-    backend_name = (char**)dlsym(backend->handle,DL_PREFIX "backend_name");
+    backend_name = (char**)dlsym(backend->handle,DL_PREFIX "wzd_backend_name");
   else
     return NULL;
 
   return strdup(*backend_name);
 }
 
-static void backend_clear_struct(wzd_backend_t *backend)
+static void backend_clear_struct(wzd_backend_def_t *backend)
 {
   if (backend->param) {
     wzd_free(backend->param);
     backend->param = NULL;
   }
-  backend->name[0] = '\0';
+  wzd_free(backend->name);
+  backend->name = NULL;
   backend->handle = NULL;
   backend->back_validate_login = NULL;
   backend->back_validate_pass = NULL;
@@ -114,11 +119,13 @@ static void backend_clear_struct(wzd_backend_t *backend)
   backend->back_mod_user = NULL;
   backend->back_mod_group = NULL;
   backend->back_commit_changes = NULL;
+  wzd_free(backend->b);
+  backend->b = NULL;
 }
 
 int backend_validate(const char *backend, const char *pred, const char *version)
 {
-  struct stat statbuf;
+  struct statbuf st;
   int ret;
   void * handle;
   char filename[1024];
@@ -134,7 +141,7 @@ int backend_validate(const char *backend, const char *pred, const char *version)
     path[length]='\0';
   }
 
-  DIRNORM(backend,strlen(backend),0);
+  DIRNORM((char*)backend,strlen(backend),0);
   /* TODO if backend name already contains .so, do not add .o */
   /* if backend name contains /, do not add path */
   if (strchr(backend,'/')==NULL)
@@ -150,21 +157,12 @@ int backend_validate(const char *backend, const char *pred, const char *version)
     out_err(LEVEL_HIGH,"Backend name too long (%s:%d)\n",__FILE__,__LINE__);
     return 1;
   }
-  ret = lstat(filename,&statbuf);
+  ret = fs_lstat(filename,&st);
   if (ret) {
     out_err(LEVEL_HIGH,"Could not stat backend '%s'\n",filename);
     out_err(LEVEL_HIGH,"errno: %d error: %s\n",errno, strerror(errno));
     return 1;
   }
-  /* basic type check */
-#if 0
-#ifdef DEBUG
-  if (S_ISLNK(statbuf.st_mode))
-    out_err(LEVEL_INFO,"%s is a symlink, ok\n",filename);
-  if (S_ISREG(statbuf.st_mode))
-      out_err(LEVEL_INFO,"%s is a regular file, ok\n",filename);
-#endif
-#endif
 
   /* test dlopen */
   handle = dlopen(filename,DL_ARG);
@@ -175,6 +173,19 @@ int backend_validate(const char *backend, const char *pred, const char *version)
     return 1;
   }
 
+  /* search wzd_backend_init. If found, use the new interface */
+  {
+    typedef int (*backend_init_function)(wzd_backend_t*);
+
+    backend_init_function fcn;
+
+    fcn = (backend_init_function)dlsym(handle, DL_PREFIX "wzd_backend_init");
+    if (fcn) {
+      dlclose(handle);
+      return 0;
+    }
+  }
+
   /* check functions */
   ret = 1;
   ret = (_backend_check_function(handle, STR_VALIDATE_LOGIN, backend)) && ret;
@@ -201,14 +212,14 @@ int backend_validate(const char *backend, const char *pred, const char *version)
         dlclose(handle);
         return 1;
       }
-      version_found = (char**)dlsym(handle,DL_PREFIX "backend_version");
+      version_found = (char**)dlsym(handle,DL_PREFIX "wzd_backend_version");
 #ifndef _MSC_VER
       if ( (dlerror()) != NULL )
 #else
       if ( !version_found )
 #endif
       {
-        out_err(LEVEL_CRITICAL,"Backend does not contain any \"backend_version\" information\n");
+        out_err(LEVEL_CRITICAL,"Backend does not contain any \"wzd_backend_version\" information\n");
         dlclose(handle);
         return 1;
       }
@@ -221,19 +232,18 @@ int backend_validate(const char *backend, const char *pred, const char *version)
   } /* if (pred) */
 
   dlclose(handle);
-/*  strncpy(mainConfig->backend.name,backend,HARD_BACKEND_NAME_LENGTH-1);*/
 
   return 0;
 }
 
-int backend_init(const char *backend, int *backend_storage, wzd_user_t * user_list, unsigned int user_max, wzd_group_t * group_list, unsigned int group_max)
+int backend_init(const char *backend, unsigned int user_max, unsigned int group_max)
 {
   void * handle;
   char filename[1024];
   char path[1024];
   int length;
   void *ptr;
-  int (*init_fcn)(int *, wzd_user_t *, unsigned int, wzd_group_t *, unsigned int, void*);
+  int (*init_fcn)(const char *);
   int ret;
 
   /* default: current path */
@@ -245,7 +255,7 @@ int backend_init(const char *backend, int *backend_storage, wzd_user_t * user_li
     path[length]='\0';
   }
 
-  DIRNORM(backend,strlen(backend),0);
+  DIRNORM((char*)backend,strlen(backend),0);
   /* TODO if backend name already contains .so, do not add .o */
   /* if backend name contains /, do not add path */
   if (strchr(backend,'/')==NULL)
@@ -272,22 +282,82 @@ int backend_init(const char *backend, int *backend_storage, wzd_user_t * user_li
   }
 
   mainConfig->backend.handle = handle;
-  ptr = init_fcn = (int (*)(int *, wzd_user_t *, unsigned int, wzd_group_t *, unsigned int, void *))dlsym(handle,DL_PREFIX STR_INIT);
-  mainConfig->backend.back_validate_login = (int (*)(const char *, wzd_user_t *))dlsym(handle,DL_PREFIX STR_VALIDATE_LOGIN);
-  mainConfig->backend.back_validate_pass  = (int (*)(const char *, const char *, wzd_user_t *))dlsym(handle,DL_PREFIX STR_VALIDATE_PASS);
-  mainConfig->backend.back_get_user  = (wzd_user_t * (*)(int))dlsym(handle,DL_PREFIX STR_GET_USER);
-  mainConfig->backend.back_get_group  = (wzd_group_t * (*)(int))dlsym(handle,DL_PREFIX STR_GET_GROUP);
-  mainConfig->backend.back_find_user  = (int (*)(const char *, wzd_user_t *))dlsym(handle,DL_PREFIX STR_FIND_USER);
-  mainConfig->backend.back_find_group  = (int (*)(const char *, wzd_group_t *))dlsym(handle,DL_PREFIX STR_FIND_GROUP);
+
+  /* search wzd_backend_init. If found, use the new interface */
+  {
+    typedef int (*backend_init_function)(wzd_backend_t*);
+
+    backend_init_function fcn;
+    wzd_backend_t * b;
+
+    fcn = (backend_init_function)dlsym(handle, DL_PREFIX "wzd_backend_init");
+    if (fcn) {
+
+      b = mainConfig->backend.b = wzd_malloc(sizeof(wzd_backend_t));
+      memset(b,0,sizeof(wzd_backend_t));
+      b->struct_version = STRUCT_BACKEND_VERSION;
+
+      if (backend != mainConfig->backend.name) /* strings must not overlap */
+      {
+        wzd_free(mainConfig->backend.name);
+        mainConfig->backend.name = wzd_strdup(backend);
+      }
+
+      ret = (*fcn)(b);
+
+      /* compatibility layer (to be removed)
+       *   copy all fields from backend->b to backend
+       */
+      {
+        mainConfig->backend.name = b->name;
+        mainConfig->backend.back_validate_login= b->backend_validate_login;
+        mainConfig->backend.back_validate_pass= b->backend_validate_pass;
+        mainConfig->backend.back_get_user= b->backend_get_user;
+        mainConfig->backend.back_get_group= b->backend_get_group;
+        mainConfig->backend.back_find_user= b->backend_find_user;
+        mainConfig->backend.back_find_group= b->backend_find_group;
+        mainConfig->backend.back_chpass= b->backend_chpass;
+        mainConfig->backend.back_mod_user= b->backend_mod_user;
+        mainConfig->backend.back_mod_group= b->backend_mod_group;
+        mainConfig->backend.back_commit_changes= b->backend_commit_changes;
+      }
+
+      if (b->backend_init) {
+        ret = (b->backend_init)(mainConfig->backend.param);
+        if (ret) { /* backend says NO */
+          backend_clear_struct(&mainConfig->backend);
+          dlclose(handle);
+          return ret;
+        }
+      } else {
+        /* if no init function is present, we consider the module is ok */
+        ret = 0;
+      }
+
+      out_log(LEVEL_INFO,"Backend %s loaded (new interface)\n",backend);
+
+      return ret;
+    }
+  }
+
+  ptr = init_fcn = (int (*)(const char *))dlsym(handle,DL_PREFIX STR_INIT);
+  mainConfig->backend.back_validate_login = (uid_t (*)(const char *, wzd_user_t *))dlsym(handle,DL_PREFIX STR_VALIDATE_LOGIN);
+  mainConfig->backend.back_validate_pass  = (uid_t (*)(const char *, const char *, wzd_user_t *))dlsym(handle,DL_PREFIX STR_VALIDATE_PASS);
+  mainConfig->backend.back_get_user  = (wzd_user_t * (*)(uid_t))dlsym(handle,DL_PREFIX STR_GET_USER);
+  mainConfig->backend.back_get_group  = (wzd_group_t * (*)(gid_t))dlsym(handle,DL_PREFIX STR_GET_GROUP);
+  mainConfig->backend.back_find_user  = (uid_t (*)(const char *, wzd_user_t *))dlsym(handle,DL_PREFIX STR_FIND_USER);
+  mainConfig->backend.back_find_group  = (gid_t (*)(const char *, wzd_group_t *))dlsym(handle,DL_PREFIX STR_FIND_GROUP);
   mainConfig->backend.back_mod_user  = (int (*)(const char *, wzd_user_t *, unsigned long))dlsym(handle,DL_PREFIX STR_MOD_USER);
   mainConfig->backend.back_mod_group  = (int (*)(const char *, wzd_group_t *, unsigned long))dlsym(handle,DL_PREFIX STR_MOD_GROUP);
   mainConfig->backend.back_commit_changes  = (int (*)(void))dlsym(handle,DL_PREFIX STR_COMMIT_CHANGES);
   if (backend != mainConfig->backend.name) /* strings must not overlap */
-    strncpy(mainConfig->backend.name,backend,HARD_BACKEND_NAME_LENGTH-1);
+  {
+    wzd_free(mainConfig->backend.name);
+    mainConfig->backend.name = wzd_strdup(backend);
+  }
 
   if (ptr) {
-    ret = (*init_fcn)(backend_storage, user_list, user_max, group_list, group_max, mainConfig->backend.param);
-/*    ret = (*init_fcn)(backend_storage, user_list, user_max, group_list, group_max, NULL);*/
+    ret = (*init_fcn)(mainConfig->backend.param);
     if (ret) { /* backend says NO */
       backend_clear_struct(&mainConfig->backend);
       dlclose(handle);
@@ -298,7 +368,6 @@ int backend_init(const char *backend, int *backend_storage, wzd_user_t * user_li
     ret = 0;
   }
 
-  mainConfig->backend.backend_storage = *backend_storage;
   out_log(LEVEL_INFO,"Backend %s loaded\n",backend);
 
   return ret;
@@ -309,11 +378,17 @@ int backend_close(const char *backend)
   int (*fini_fcn)(void);
   int ret;
 
+  if (!backend || !mainConfig->backend.name) return 1;
+
   /* step 1: check that backend == mainConfig->backend.name */
   if (strcmp(backend,mainConfig->backend.name)!=0) return 1;
 
   /* step 2: call end function */
-  fini_fcn = (int (*)(void))dlsym(mainConfig->backend.handle,DL_PREFIX STR_FINI);
+  if (mainConfig->backend.b) {
+    fini_fcn = ((wzd_backend_t*)mainConfig->backend.b)->backend_exit;
+  } else {
+    fini_fcn = (int (*)(void))dlsym(mainConfig->backend.handle,DL_PREFIX STR_FINI);
+  }
   if (fini_fcn) {
     ret = (*fini_fcn)();
     if (ret) {
@@ -326,19 +401,24 @@ int backend_close(const char *backend)
   /* close backend */
   ret = 0;
   if (mainConfig->backend.handle)
+  {
+    char * tempname = strdup(backend);
     ret = dlclose(mainConfig->backend.handle);
-  if (ret) {
+    if (ret) {
 #ifdef WIN32
-    ret = GetLastError();
-    out_log(LEVEL_INFO," Could not close backend %s (handled %lu)\n Error %d %s\n",
-      backend,mainConfig->backend.handle, ret,strerror(ret));
-    backend_clear_struct(&mainConfig->backend);
+      ret = GetLastError();
+      out_log(LEVEL_INFO," Could not close backend %s (handled %lu)\n Error %d %s\n",
+          tempname,mainConfig->backend.handle, ret,strerror(ret));
+      backend_clear_struct(&mainConfig->backend);
 #else
-    out_log(LEVEL_INFO,"Could not close backend %s (handle %lu)\n",
-      backend,mainConfig->backend.handle);
-    out_log(LEVEL_INFO," Error '%s'\n",dlerror());
+      out_log(LEVEL_INFO,"Could not close backend %s (handle %lu)\n",
+          tempname,mainConfig->backend.handle);
+      out_log(LEVEL_INFO," Error '%s'\n",dlerror());
 #endif
-    return 1;
+      free(tempname);
+      return 1;
+    }
+    free(tempname);
   }
 
   backend_clear_struct(&mainConfig->backend);
@@ -349,20 +429,22 @@ int backend_close(const char *backend)
 int backend_reload(const char *backend)
 {
   int ret;
-  int backend_storage;
 
   ret = backend_close(backend);
   if (ret) return 1;
 
-  ret = backend_init(backend,&backend_storage,mainConfig->user_list,HARD_DEF_USER_MAX,mainConfig->group_list,HARD_DEF_GROUP_MAX);
+  ret = backend_init(backend,0 /* max users */,0 /* max groups */);
   if (ret) return 1;
-  mainConfig->backend.backend_storage = backend_storage;
 
   return 0;
 }
 
-wzd_user_t * backend_get_user(int userid)
+wzd_user_t * backend_get_user(uid_t userid)
 {
+  wzd_backend_t * b;
+  if ( (b = mainConfig->backend.b) && b->backend_get_user)
+    return b->backend_get_user(userid);
+
   if (!mainConfig->backend.handle || !mainConfig->backend.back_get_user) {
     out_log(LEVEL_CRITICAL,"Attempt to call a backend function on %s:%d while there is no available backend !\n", __FILE__, __LINE__);
     return NULL;
@@ -374,117 +456,106 @@ wzd_user_t * backend_get_user(int userid)
 int backend_find_user(const char *name, wzd_user_t * user, int * userid)
 {
   int ret;
-  if (!mainConfig->backend.handle || !mainConfig->backend.back_find_user) {
-    out_log(LEVEL_CRITICAL,"Attempt to call a backend function on %s:%d while there is no available backend !\n", __FILE__, __LINE__);
-    return 1;
-  }
-  ret = (*mainConfig->backend.back_find_user)(name,user);
-  if (ret >= 0) {
-    if (mainConfig->backend.backend_storage == 0) {
-      /*user = GetUserByID(ret);*/
-      memcpy(user,GetUserByID(ret),sizeof(wzd_user_t));
-      if (userid) *userid = ret;
-      return 0;
-    } else {
-      memcpy(user,GetUserByID(ret),sizeof(wzd_user_t));
-      if (userid) *userid = ret;
-      return 0;
+  wzd_backend_t * b;
+
+  if ( (b = mainConfig->backend.b) && b->backend_find_user)
+    ret = b->backend_find_user(name,user);
+  else {
+    if (!mainConfig->backend.handle || !mainConfig->backend.back_find_user) {
+      out_log(LEVEL_CRITICAL,"Attempt to call a backend function on %s:%d while there is no available backend !\n", __FILE__, __LINE__);
+      return -1;
     }
+    ret = (*mainConfig->backend.back_find_user)(name,user);
+  }
+
+  if (ret >= 0 && user) {
+    wzd_user_t * _tmp_user;
+    _tmp_user = GetUserByID(ret);
+    if (!_tmp_user) return -1;
+    memcpy(user,_tmp_user,sizeof(wzd_user_t));
+    if (userid) *userid = ret;
+    return 0;
   }
   return ret;
 }
 
 /** wrappers to user list */
-wzd_user_t * GetUserByID(unsigned int id)
+wzd_user_t * GetUserByID(uid_t id)
 {
-  unsigned int i=0;
   wzd_user_t *user, *user_return;
+  wzd_backend_t * b;
 
-  if (!mainConfig || !mainConfig->user_list) return NULL;
+  if (!mainConfig) return NULL;
 
-#ifdef BACKEND_STORAGE
-  if (mainConfig->backend.backend_storage==1) {
-    /* try cache first */
-    if ( (user = usercache_get( predicate_uid, (void*)id )) )
-      return user;
+  /* try cache first */
+  if ( (user = usercache_getbyuid( id )) )
+    return user;
 
+  if ( (b = mainConfig->backend.b) && b->backend_get_user)
+    user = b->backend_get_user(id);
+  else {
     if (!mainConfig->backend.handle || !mainConfig->backend.back_get_user) {
       out_log(LEVEL_CRITICAL,"Attempt to call a backend function on %s:%d while there is no available backend !\n", __FILE__, __LINE__);
       return NULL;
     }
-    user = (*mainConfig->backend.back_get_user)( (int)id );
-    if (!user) return NULL;
-    user_return = usercache_add( user );
-    wzd_free(user);
-    return user_return;
+    user = (*mainConfig->backend.back_get_user)( id );
   }
-#endif
 
-  while (i<HARD_DEF_USER_MAX)
-  {
-    if (mainConfig->user_list[i].username[0] != '\0' && mainConfig->user_list[i].uid == id) {
-        return &mainConfig->user_list[i];
-    }
-    i++;
-  }
-  return NULL;
+  if (!user) return NULL;
+  user_return = usercache_add( user );
+  wzd_free(user);
+  return user_return;
 }
 
 wzd_user_t * GetUserByName(const char *name)
 {
-  int i=0;
-  unsigned int uid;
+  uid_t uid;
   wzd_user_t * user=NULL;
+  wzd_backend_t * b;
 
-  if (!mainConfig || !mainConfig->user_list || !name || strlen(name)<=0) return NULL;
+  if (!mainConfig || !name || strlen(name)<=0) return NULL;
 out_err(LEVEL_CRITICAL,"GetUserByName %s\n",name);
 
-#ifdef BACKEND_STORAGE
-  if (mainConfig->backend.backend_storage==1) {
-    /* try cache first */
-    if ( (user = usercache_get( predicate_name, (void*)name )) )
-      return user;
+  /* try cache first */
+  if ( (user = usercache_getbyname( name )) )
+    return user;
 
+  if ( (b = mainConfig->backend.b) && b->backend_find_user)
+    uid = b->backend_find_user(name,user);
+  else {
     if (!mainConfig->backend.handle || !mainConfig->backend.back_find_user) {
       out_log(LEVEL_CRITICAL,"Attempt to call a backend function on %s:%d while there is no available backend !\n", __FILE__, __LINE__);
       return NULL;
     }
     uid = (*mainConfig->backend.back_find_user)(name,user);
-    if (uid >= 0) {
-      return GetUserByID( uid );
-    }
-    return NULL;
-  }
-#endif
-
-  while (i<HARD_DEF_USER_MAX)
-  {
-    if (mainConfig->user_list[i].username[0] != '\0') {
-      if (strcmp(name,mainConfig->user_list[i].username)==0)
-        return &mainConfig->user_list[i];
-    }
-    i++;
   }
 
+  if (uid != (uid_t)-1) {
+    return GetUserByID( uid );
+  }
   return NULL;
 }
 
-unsigned int GetUserIDByName(const char *name)
+uid_t GetUserIDByName(const char *name)
 {
   wzd_user_t * user;
 
   if ( (user=GetUserByName(name)) )
     return user->uid;
 
-  return -1;
+  return (uid_t)-1;
 }
 
 
 
 
 
-wzd_group_t * backend_get_group(int groupid)
+wzd_group_t * backend_get_group(gid_t groupid)
 {
+  wzd_backend_t * b;
+  if ( (b = mainConfig->backend.b) && b->backend_get_group)
+    return b->backend_get_group(groupid);
+
   if (!mainConfig->backend.handle || !mainConfig->backend.back_get_group) {
     out_log(LEVEL_CRITICAL,"Attempt to call a backend function on %s:%d while there is no available backend !\n", __FILE__, __LINE__);
     return NULL;
@@ -495,13 +566,23 @@ wzd_group_t * backend_get_group(int groupid)
 int backend_find_group(const char *name, wzd_group_t * group, int * groupid)
 {
   int ret;
-  if (!mainConfig->backend.handle || !mainConfig->backend.back_find_group) {
-    out_log(LEVEL_CRITICAL,"Attempt to call a backend function on %s:%d while there is no available backend !\n", __FILE__, __LINE__);
-    return 1;
+  wzd_backend_t * b;
+
+  if ( (b = mainConfig->backend.b) && b->backend_find_group)
+    ret = b->backend_find_group(name,group);
+  else {
+    if (!mainConfig->backend.handle || !mainConfig->backend.back_find_group) {
+      out_log(LEVEL_CRITICAL,"Attempt to call a backend function on %s:%d while there is no available backend !\n", __FILE__, __LINE__);
+      return -1;
+    }
+    ret = (*mainConfig->backend.back_find_group)(name,group);
   }
-  ret = (*mainConfig->backend.back_find_group)(name,group);
-  if (mainConfig->backend.backend_storage == 0 && ret >= 0) {
-    memcpy(group,GetGroupByID(ret),sizeof(wzd_group_t));
+
+  if (ret >= 0 && group) {
+    wzd_group_t * _tmp_group;
+    _tmp_group = GetGroupByID(ret);
+    if (!_tmp_group) return -1;
+    memcpy(group,_tmp_group,sizeof(wzd_group_t));
     if (groupid) *groupid = ret;
     return 0;
   }
@@ -510,137 +591,136 @@ int backend_find_group(const char *name, wzd_group_t * group, int * groupid)
 
 
 /** wrappers to Group list */
-wzd_group_t * GetGroupByID(unsigned int id)
+wzd_group_t * GetGroupByID(gid_t id)
 {
-  unsigned int i=0;
   wzd_group_t * group = NULL, * group_return;
+  wzd_backend_t * b;
 
-  if (!mainConfig || !mainConfig->group_list) return NULL;
+  if (!mainConfig) return NULL;
 
-#ifdef BACKEND_STORAGE
-  if (mainConfig->backend.backend_storage==1) {
-    /* try cache first */
-    if ( (group = groupcache_get( predicate_gid, (void*)id )) )
-      return group;
+  /* try cache first */
+  if ( (group = groupcache_getbygid( id )) )
+    return group;
 
+  if ( (b = mainConfig->backend.b) && b->backend_get_group)
+    group = b->backend_get_group(id);
+  else {
     if (!mainConfig->backend.handle || !mainConfig->backend.back_get_group) {
       out_log(LEVEL_CRITICAL,"Attempt to call a backend function on %s:%d while there is no available backend !\n", __FILE__, __LINE__);
       return NULL;
     }
-    group = (*mainConfig->backend.back_get_group)( (int)id );
-    if (!group) return NULL;
-    group_return = groupcache_add( group );
-    wzd_free(group);
-    return group_return;
+    group = (*mainConfig->backend.back_get_group)( id );
   }
-#endif
-/*  if (id >= HARD_DEF_GROUP_MAX ) return NULL;*/ /* used when gid == index */
-/*  return &mainConfig->group_list[id];*/
 
-
-  while (i<HARD_DEF_GROUP_MAX)
-  {
-    if (mainConfig->group_list[i].groupname[0] != '\0' && mainConfig->group_list[i].gid == id) {
-        return &mainConfig->group_list[i];
-    }
-    i++;
-  }
-  return NULL;
+  if (!group) return NULL;
+  group_return = groupcache_add( group );
+  wzd_free(group);
+  return group_return;
 }
 
 wzd_group_t * GetGroupByName(const char *name)
 {
-  int i=0;
-  unsigned int gid;
+  gid_t gid;
   wzd_group_t * group = NULL;
+  wzd_backend_t * b;
 
-  if (!mainConfig || !mainConfig->group_list || !name || strlen(name)<=0) return NULL;
+  if (!mainConfig || !name || strlen(name)<=0) return NULL;
 
-#ifdef BACKEND_STORAGE
-  if (mainConfig->backend.backend_storage==1) {
-    /* try cache first */
-    if ( (group = groupcache_get( predicate_groupname, (void*)name )) )
-      return group;
+  /* try cache first */
+  if ( (group = groupcache_getbyname( name )) )
+    return group;
 
+  if ( (b = mainConfig->backend.b) && b->backend_find_group)
+    gid = b->backend_find_group(name,group);
+  else {
     if (!mainConfig->backend.handle || !mainConfig->backend.back_find_group) {
       out_log(LEVEL_CRITICAL,"Attempt to call a backend function on %s:%d while there is no available backend !\n", __FILE__, __LINE__);
       return NULL;
     }
     gid = (*mainConfig->backend.back_find_group)(name,group);
-    if (gid >= 0) {
-      return GetGroupByID( gid );
-    }
-    return NULL;
   }
-#endif
 
-
-  while (i<HARD_DEF_GROUP_MAX)
-  {
-    if (mainConfig->group_list[i].groupname[0] != '\0') {
-      if (strcmp(name,mainConfig->group_list[i].groupname)==0)
-        return &mainConfig->group_list[i];
-    }
-    i++;
+  if (gid != (gid_t)-1) {
+    return GetGroupByID( gid );
   }
 
   return NULL;
 }
 
-unsigned int GetGroupIDByName(const char *name)
+gid_t GetGroupIDByName(const char *name)
 {
   wzd_group_t * group;
 
   if ( (group=GetGroupByName(name)) )
     return group->gid;
 
+  return (gid_t)-1;
+}
+
+
+int backend_validate_login(const char *name, wzd_user_t * user, uid_t * userid)
+{
+  uid_t ret;
+  wzd_backend_t * b;
+
+  if ( (b = mainConfig->backend.b) && b->backend_validate_login)
+    ret = b->backend_validate_login(name,user);
+  else {
+    if (!mainConfig->backend.handle || !mainConfig->backend.back_validate_login) {
+      out_log(LEVEL_CRITICAL,"Attempt to call a backend function on %s:%d while there is no available backend !\n", __FILE__, __LINE__);
+      return -1;
+    }
+    ret = (*mainConfig->backend.back_validate_login)(name,user);
+  }
+
+  if (ret != (uid_t)-1) {
+    if (user) {
+      wzd_user_t * _tmp_user;
+      _tmp_user = GetUserByID(ret);
+      if (!_tmp_user) return -1;
+      memcpy(user,_tmp_user,sizeof(wzd_user_t));
+    }
+    *userid = ret;
+    return 0;
+  }
   return -1;
 }
 
-
-int backend_validate_login(const char *name, wzd_user_t * user, unsigned int * userid)
+int backend_validate_pass(const char *name, const char *pass, wzd_user_t *user, uid_t * userid)
 {
-  int ret;
-  if (!mainConfig->backend.handle || !mainConfig->backend.back_validate_login) {
-    out_log(LEVEL_CRITICAL,"Attempt to call a backend function on %s:%d while there is no available backend !\n", __FILE__, __LINE__);
-    return 1;
+  uid_t ret;
+  wzd_backend_t * b;
+
+  if ( (b = mainConfig->backend.b) && b->backend_validate_pass)
+    ret = b->backend_validate_pass(name,pass,user);
+  else {
+    if (!mainConfig->backend.handle || !mainConfig->backend.back_validate_pass) {
+      out_log(LEVEL_CRITICAL,"Attempt to call a backend function on %s:%d while there is no available backend !\n", __FILE__, __LINE__);
+      return 1;
+    }
+    ret = (*mainConfig->backend.back_validate_pass)(name,pass,user);
   }
-  ret = (*mainConfig->backend.back_validate_login)(name,user);
-  if (ret >= 0) {
-    if (mainConfig->backend.backend_storage == 0) {
-      /*user = GetUserByID(ret);*/
-      if (user != NULL)
-        memcpy(user,GetUserByID(ret),sizeof(wzd_user_t));
+
+  if (ret != (uid_t)-1) {
+    if (user) {
+      wzd_user_t * _tmp_user;
+      _tmp_user = GetUserByID(ret);
+      if (!_tmp_user) return -1;
+      memcpy(user,_tmp_user,sizeof(wzd_user_t));
     }
     *userid = ret;
     return 0;
   }
-  return ret;
-}
-
-int backend_validate_pass(const char *name, const char *pass, wzd_user_t *user, unsigned int * userid)
-{
-  int ret;
-  if (!mainConfig->backend.handle || !mainConfig->backend.back_validate_pass) {
-    out_log(LEVEL_CRITICAL,"Attempt to call a backend function on %s:%d while there is no available backend !\n", __FILE__, __LINE__);
-    return 1;
-  }
-  ret = (*mainConfig->backend.back_validate_pass)(name,pass,user);
-  if (ret >= 0) {
-    if (mainConfig->backend.backend_storage == 0 && ret >= 0) {
-      /*user = GetUserByID(ret);*/
-      if (user != NULL)
-        memcpy(user,GetUserByID(ret),sizeof(wzd_user_t));
-    }
-    *userid = ret;
-    return 0;
-  }
-  return ret;
+  return -1;
 }
 
 int backend_commit_changes(const char *backend)
 {
   int ret;
+  wzd_backend_t * b;
+
+  if ( (b = mainConfig->backend.b) && b->backend_commit_changes)
+    return b->backend_commit_changes();
 
   if (!mainConfig->backend.handle || !mainConfig->backend.back_commit_changes) {
     out_log(LEVEL_CRITICAL,"Attempt to call a backend function on %s:%d while there is no available backend !\n", __FILE__, __LINE__);
@@ -655,7 +735,9 @@ int backend_commit_changes(const char *backend)
 
 int backend_inuse(const char *backend)
 {
-  int count, i;
+  int count;
+  ListElmt * elmnt;
+  wzd_context_t * context;
   /* unusually, if backend is not loaded it is not in use ... so no error here */
   if (!mainConfig->backend.handle) {
     return -1;
@@ -666,8 +748,9 @@ int backend_inuse(const char *backend)
 
   /* count user logged */
   count = 0;
-  for (i=0; i<HARD_USERLIMIT; i++) {
-    if (context_list[i].magic == CONTEXT_MAGIC) {
+  for (elmnt=list_head(context_list); elmnt != NULL; elmnt = list_next(elmnt)) {
+    context = list_data(elmnt);
+    if (context->magic == CONTEXT_MAGIC) {
       count++;
     }
   }
@@ -678,19 +761,25 @@ int backend_inuse(const char *backend)
 int backend_mod_user(const char *backend, const char *name, wzd_user_t * user, unsigned long mod_type)
 {
   int ret;
-  if (!mainConfig->backend.handle || !mainConfig->backend.back_mod_user) {
-    out_log(LEVEL_CRITICAL,"Attempt to call a backend function on %s:%d while there is no available backend !\n", __FILE__, __LINE__);
-    return 1;
-  }
-  ret = (*mainConfig->backend.back_mod_user)(name,user,mod_type);
+  wzd_backend_t * b;
 
-  usercache_invalidate( predicate_name, (void *)name );
+  if ( (b = mainConfig->backend.b) && b->backend_mod_user)
+    ret = b->backend_mod_user(name,user,mod_type);
+  else {
+    if (!mainConfig->backend.handle || !mainConfig->backend.back_mod_user) {
+      out_log(LEVEL_CRITICAL,"Attempt to call a backend function on %s:%d while there is no available backend !\n", __FILE__, __LINE__);
+      return 1;
+    }
+    ret = (*mainConfig->backend.back_mod_user)(name,user,mod_type);
+  }
 
   if (ret == 0) {
     if (mod_type & _USER_MAX_ULS) _trigger_user_max_ul(user);
     if (mod_type & _USER_MAX_DLS) _trigger_user_max_dl(user);
   }
 
+  usercache_invalidate( predicate_name, (void *)name );
+
   return ret;
 }
 
@@ -698,11 +787,17 @@ int backend_mod_user(const char *backend, const char *name, wzd_user_t * user, u
 int backend_mod_group(const char *backend, const char *name, wzd_group_t * group, unsigned long mod_type)
 {
   int ret;
-  if (!mainConfig->backend.handle || !mainConfig->backend.back_mod_group) {
-    out_log(LEVEL_CRITICAL,"Attempt to call a backend function on %s:%d while there is no available backend !\n", __FILE__, __LINE__);
-    return 1;
+  wzd_backend_t * b;
+
+  if ( (b = mainConfig->backend.b) && b->backend_mod_group)
+    ret = b->backend_mod_group(name,group,mod_type);
+  else {
+    if (!mainConfig->backend.handle || !mainConfig->backend.back_mod_group) {
+      out_log(LEVEL_CRITICAL,"Attempt to call a backend function on %s:%d while there is no available backend !\n", __FILE__, __LINE__);
+      return 1;
+    }
+    ret = (*mainConfig->backend.back_mod_group)(name,group,mod_type);
   }
-  ret = (*mainConfig->backend.back_mod_group)(name,group,mod_type);
 
   groupcache_invalidate( predicate_groupname, (void *)name );
 
@@ -724,14 +819,17 @@ static int _backend_check_function(void * handle, const char *name, const char *
 
 static int _trigger_user_max_dl(wzd_user_t * user)
 {
-  int i;
+  ListElmt * el;
+  wzd_context_t * context;
 
-  for (i=0; i<HARD_USERLIMIT; i++)
+  if (!user) return 0;
+  for (el = list_head(context_list); el != NULL; el = list_next(el))
   {
-    if (context_list[i].magic == CONTEXT_MAGIC &&
-        context_list[i].userid == user->uid)
+    context = list_data(el);
+    if (context->magic == CONTEXT_MAGIC &&
+        context->userid == user->uid)
     {
-      context_list[i].current_dl_limiter.maxspeed = user->max_dl_speed;
+      context->current_dl_limiter.maxspeed = user->max_dl_speed;
     }
   }
 
@@ -740,14 +838,17 @@ static int _trigger_user_max_dl(wzd_user_t * user)
 
 static int _trigger_user_max_ul(wzd_user_t * user)
 {
-  int i;
+  ListElmt * el;
+  wzd_context_t * context;
 
-  for (i=0; i<HARD_USERLIMIT; i++)
+  if (!user) return 0;
+  for (el = list_head(context_list); el != NULL; el = list_next(el))
   {
-    if (context_list[i].magic == CONTEXT_MAGIC &&
-        context_list[i].userid == user->uid)
+    context = list_data(el);
+    if (context->magic == CONTEXT_MAGIC &&
+        context->userid == user->uid)
     {
-      context_list[i].current_ul_limiter.maxspeed = user->max_ul_speed;
+      context->current_ul_limiter.maxspeed = user->max_ul_speed;
     }
   }
 
