@@ -24,6 +24,10 @@
  * the source code for OpenSSL in the source distribution.
  */
 
+#include "wzd_all.h"
+
+#ifndef WZD_USE_PCH
+
 #if defined(_MSC_VER) || (defined  __CYGWIN__ && defined WINSOCK_SUPPORT)
 #include <winsock2.h>
 #else
@@ -50,11 +54,14 @@
 
 
 #include "wzd_structs.h"
+#include "wzd_libmain.h"
 #include "wzd_log.h"
 #include "wzd_misc.h"
 
 #include "wzd_debug.h"
 
+#endif /* WZD_USE_PCH */
+
 #define MAX_CACHE_FILE_LEN	32768
 
 typedef struct wzd_internal_cache_t wzd_internal_cache_t;
@@ -68,7 +75,7 @@ struct wzd_internal_cache_t  {
   int fd;
 
   unsigned long filename_hash;
-  unsigned long datasize;
+  u64_t datasize;
   time_t mtime;
   unsigned short use;
 
@@ -78,7 +85,7 @@ struct wzd_internal_cache_t  {
 };
 
 struct wzd_cache_t {
-  unsigned long current_location;
+  u64_t current_location;
 
   wzd_internal_cache_t * cache;
 };
@@ -100,7 +107,7 @@ wzd_internal_cache_t * wzd_cache_find(unsigned long hash)
   return NULL;
 }
 
-unsigned long wzd_cache_getsize(wzd_cache_t *c)
+u64_t wzd_cache_getsize(wzd_cache_t *c)
 {
   if (!c) return (unsigned int)-1;
   return c->cache->datasize;
@@ -108,14 +115,18 @@ unsigned long wzd_cache_getsize(wzd_cache_t *c)
 
 wzd_cache_t * wzd_cache_open(const char *file, int flags, unsigned int mode)
 {
+#ifdef ENABLE_CACHE
   wzd_cache_t * cache;
   wzd_internal_cache_t * c;
-  struct stat s;
+  struct statbuf s;
   unsigned long hash;
   unsigned long ret;
-  unsigned long length;
+  unsigned int length;
+  u64_t l64;
   int fd;
 
+  if (!file) return NULL;
+
   hash = compute_hashval(file,strlen(file));
 /*  out_err(LEVEL_FLOOD,"HASH %s: %lu\n",file,hash);*/
 
@@ -123,10 +134,10 @@ wzd_cache_t * wzd_cache_open(const char *file, int flags, unsigned int mode)
   flags |= _O_BINARY;
 #endif
 
-  fd = open(file,flags,mode);
+  fd = fs_open(file,flags,mode);
   if (fd==-1) return NULL;
 
-  if (fstat(fd,&s)) { close(fd); return NULL; }
+  if (fs_fstat(fd,&s)) { close(fd); return NULL; }
   FD_REGISTER(fd,"Cached file");
 
   c = wzd_cache_find(hash);
@@ -148,6 +159,7 @@ wzd_cache_t * wzd_cache_open(const char *file, int flags, unsigned int mode)
     cache = malloc(sizeof(wzd_cache_t));
     cache->current_location = 0;
     cache->cache = c;
+    c->use++;
 #ifdef WZD_DBG_CACHE
     out_err(LEVEL_FLOOD,"Cache HIT %s\n",file);
 #endif
@@ -156,23 +168,24 @@ wzd_cache_t * wzd_cache_open(const char *file, int flags, unsigned int mode)
 
   /* MISS */
 #ifdef WZD_DBG_CACHE
-  out_err(LEVEL_FLOOD,"Cache MISS %s\n",file);
+  out_err(LEVEL_FLOOD,"Cache MISS %s (%d)\n",file,fd);
 #endif
 
   cache = (wzd_cache_t*)malloc(sizeof(wzd_cache_t));
   c = malloc(sizeof(wzd_internal_cache_t));
   c->fd = fd;
   c->filename_hash = hash;
-  c->use = 0;
+  c->use = 1;
   c->mtime = s.st_mtime;
   cache->cache = c;
   cache->current_location = 0;
-  length = (unsigned long)s.st_size;
-  if (length > MAX_CACHE_FILE_LEN) {
+  l64 = s.st_size;
+  if (l64 > MAX_CACHE_FILE_LEN) {
     out_err(LEVEL_FLOOD,"File too big to be stored in cache (%ld bytes)\n",length);
     c->data = NULL;
     c->datasize = 0;
   } else {
+    length = (unsigned int)l64;
     c->data = malloc(length+1);
     if ( (ret=(unsigned long)read(fd,c->data,length)) != length ) {
       out_err(LEVEL_FLOOD,"Read only %ld bytes on %ld required\n",ret,length);
@@ -188,25 +201,59 @@ wzd_cache_t * wzd_cache_open(const char *file, int flags, unsigned int mode)
   global_cache = c;
 
   return cache;
+#else /* ENABLE_CACHE */
+
+  struct statbuf s;
+  wzd_cache_t * cache;
+  wzd_internal_cache_t * c;
+  int fd;
+
+  if (!file) return NULL;
+
+#ifdef _MSC_VER
+  flags |= _O_BINARY;
+#endif
+
+  fd = fs_open(file,flags,mode);
+  if (fd==-1) return NULL;
+
+  if (fs_fstat(fd,&s)) { close(fd); return NULL; }
+  FD_REGISTER(fd,"Cached file");
+
+  cache = (wzd_cache_t*)malloc(sizeof(wzd_cache_t));
+  c = malloc(sizeof(wzd_internal_cache_t));
+  c->fd = fd;
+  c->filename_hash = 0;
+  c->use = 1;
+  c->mtime = s.st_mtime;
+  cache->cache = c;
+  cache->current_location = 0;
+  c->datasize = s.st_size;
+  c->data = NULL;
+
+  return cache;
+  
+#endif /* ENABLE_CACHE */
 }
 
 wzd_cache_t* wzd_cache_refresh(wzd_internal_cache_t *c, const char *file, int flags, unsigned int mode)
 {
+#ifdef ENABLE_CACHE
   wzd_cache_t * cache;
-  struct stat s;
+  struct statbuf s;
   unsigned long hash;
-  unsigned long length, ret;
+  u64_t length, ret;
   int fd;
 
   hash = compute_hashval(file,strlen(file));
 
-  fd = open(file,flags,mode);
+  fd = fs_open(file,flags,mode);
   if (fd==-1) return NULL;
+
+  if (fs_fstat(fd,&s)) { close(fd); return NULL; }
   FD_REGISTER(fd,"Cached file");
 
-  if (fstat(fd,&s)) { close(fd); return NULL; }
-
-  if (c->fd != -1) { close(c->fd); FD_UNREGISTER(fd,"Cached file"); }
+  if (c->fd != -1) { close(c->fd); FD_UNREGISTER(c->fd,"Cached file"); }
   if (c->data) free(c->data);
 
   cache = malloc(sizeof(wzd_cache_t));
@@ -215,14 +262,15 @@ wzd_cache_t* wzd_cache_refresh(wzd_internal_cache_t *c, const char *file, int fl
   c->mtime = s.st_mtime;
   cache->cache = c;
   cache->current_location = 0;
-  length = (unsigned long)s.st_size;
+  length = s.st_size;
+  c->use++;
   if (length > MAX_CACHE_FILE_LEN) {
     out_err(LEVEL_FLOOD,"File too big to be stored in cache (%ld bytes)\n",length);
     c->data = NULL;
     c->datasize = 0;
   } else {
-    c->data = malloc(length);
-    if ( (ret=(unsigned long)read(fd,c->data,length)) != length ) {
+    c->data = malloc((unsigned int)length);
+    if ( (ret=read(fd,c->data,length)) != length ) {
       out_err(LEVEL_FLOOD,"Read only %ld bytes\n",ret);
     }
     c->datasize = length;
@@ -234,11 +282,54 @@ wzd_cache_t* wzd_cache_refresh(wzd_internal_cache_t *c, const char *file, int fl
 
   return cache;
 
+#else /* ENABLE_CACHE */
+
+  struct statbuf s;
+  wzd_cache_t * cache;
+  wzd_internal_cache_t c2, c_old;
+  int fd;
+
+  if (!file) return NULL;
+
+#ifdef _MSC_VER
+  flags |= _O_BINARY;
+#endif
+
+  fd = fs_open(file,flags,mode);
+  if (fd==-1) return NULL;
+
+  if (fs_fstat(fd,&s)) { close(fd); return NULL; }
+  FD_REGISTER(fd,"Cached file");
+
+  cache = (wzd_cache_t*)malloc(sizeof(wzd_cache_t));
+  c2.fd = fd;
+  c2.filename_hash = 0;
+  c2.use = 1;
+  c2.mtime = s.st_mtime;
+  cache->cache = c;
+  cache->current_location = 0;
+  c->datasize = s.st_size;
+  c2.data = NULL;
+
+  /* atomic part */
+  memcpy(&c_old, c, sizeof(wzd_internal_cache_t));
+  memcpy(c, &c2, sizeof(wzd_internal_cache_t));
+
+  if (c_old.fd) {
+    FD_UNREGISTER(c_old.fd, "Cached file");
+    close(c_old.fd);
+  }
+  if (c_old.data) free(c_old.data);
+
+  return cache;
+  
+#endif /* ENABLE_CACHE */
 }
 
 /** force update of specific file, only if present in cache */
 void wzd_cache_update(const char *file)
 {
+#ifdef ENABLE_CACHE
   wzd_internal_cache_t * c;
   unsigned long hash;
 
@@ -252,6 +343,11 @@ void wzd_cache_update(const char *file)
 /*    out_err(LEVEL_FLOOD,"cache refresh forced\n");*/
     (void)wzd_cache_refresh(c,file,O_RDONLY,0600);
   }
+#else /* ENABLE_CACHE */
+
+  out_err(LEVEL_HIGH,"*** warning *** call to %s\n",__FUNCTION__);
+  
+#endif /* ENABLE_CACHE */
 }
 
 int wzd_cache_read(wzd_cache_t * c, void *buf, unsigned int count)
@@ -262,6 +358,7 @@ int wzd_cache_read(wzd_cache_t * c, void *buf, unsigned int count)
 /*  out_err(LEVEL_FLOOD,"cache read\n");*/
   /* if in cache, read data and pay attention to size ! */
   /* is file stored in cache ? */
+#ifdef ENABLE_CACHE
   if (cache->data) {
     if ( (c->current_location+count) <= cache->datasize ) {
       memcpy(buf,cache->data + c->current_location,count);
@@ -272,23 +369,30 @@ int wzd_cache_read(wzd_cache_t * c, void *buf, unsigned int count)
     c->current_location = cache->datasize;
     return (int)(cache->datasize-c->current_location);
   } else { /* not in cache */
+#endif
     /* update current_location */
     if (c) {
       ret = (int)read( cache->fd, buf, count );
       if (ret>0) c->current_location += ret;
       return ret;
     }
+#ifdef ENABLE_CACHE
   } /* file in cache ? */
+#endif
   return -1;
 }
 
 int wzd_cache_write(wzd_cache_t * c, void *buf, unsigned int count)
 {
+#ifdef ENABLE_CACHE
   int ret;
+#endif
+
   wzd_internal_cache_t * cache;
   cache = c->cache;
   out_err(LEVEL_FLOOD,"cache write\n");
   /* update current_location */
+#ifdef ENABLE_CACHE
   if (c) {
     /** \todo if in cache, warn user it is rather stupid to cache a file
       * to be modified, and re-open it in non-cache mode
@@ -301,6 +405,7 @@ int wzd_cache_write(wzd_cache_t * c, void *buf, unsigned int count)
     if (ret>0) c->current_location += ret;
     return ret;
   }
+#endif
   return -1;
 }
 
@@ -319,6 +424,7 @@ char * wzd_cache_gets(wzd_cache_t * c, char *buf, unsigned int size)
   cache = c->cache;
   fd = cache->fd;
   /* is file stored in cache ? */
+#ifdef ENABLE_CACHE
   if (cache->data) {
     /* get start position */
     position = c->current_location;
@@ -359,6 +465,7 @@ char * wzd_cache_gets(wzd_cache_t * c, char *buf, unsigned int size)
     c->current_location = position + (dst-buf);
 
   } else { /* file is not in cache ! */
+#endif
 
     /* get start position */
     position = lseek(fd,0,SEEK_CUR);
@@ -385,7 +492,9 @@ char * wzd_cache_gets(wzd_cache_t * c, char *buf, unsigned int size)
     (void)lseek(fd,position + (dst-buf), SEEK_SET );
     /* update current_location */
     c->current_location += strlen(buf);
+#ifdef ENABLE_CACHE
   } /* file in cache ? */
+#endif
 
   return buf;
 }
@@ -394,11 +503,15 @@ void wzd_cache_close(wzd_cache_t * c)
 {
   if (c) {
     c->cache->use--;
-    free(c);
     /** \bug XXX FIXME possible leak here if big file, fd is not closed */
- /*   close( c->fd );
-    free(c);*/
+    if (c->cache->use == 0) {
+      out_err(LEVEL_FLOOD,"Closing file %d\n",c->cache->fd);
+      FD_UNREGISTER(c->cache->fd,"Cached file"); }
+      close( c->cache->fd );
+      free( c-> cache );
+      c->cache = NULL;
   }
+  free(c);
 }
 
 void wzd_cache_purge(void)
@@ -416,6 +529,7 @@ void wzd_cache_purge(void)
     }
     if (cache_current->fd != -1) {
       close(cache_current->fd);
+      FD_UNREGISTER(cache_current->fd,"Cached file");
       cache_current->fd = -1;
     }
     free(cache_current);
@@ -425,9 +539,20 @@ void wzd_cache_purge(void)
 
 
 
-static wzd_user_t * cache_user_pool=NULL;
-static wzd_group_t * cache_group_pool=NULL;
+static CHTBL index_user_name;
+static CHTBL index_user_uid;
+static CHTBL index_group_name;
+static CHTBL index_group_gid;
 
+int hash_uid(const void *key)
+{
+  return ( (unsigned int) key );
+}
+
+int uidcmp(const void *key1, const void *key2)
+{
+  return ( ((unsigned int)key1) != ((unsigned int)key2) );
+}
 
 int predicate_uid(wzd_user_t * user, void * arg)
 {
@@ -449,174 +574,205 @@ int predicate_groupname(wzd_group_t * group, void * arg)
   return (strcmp(group->groupname,(char*)arg)==0);
 }
 
+int predicate_groupname(wzd_group_t * group, void * arg);
+
+#ifdef WZD_DBG_UGCACHE
+static void _user_free(wzd_user_t *user)
+{
+  out_err(LEVEL_CRITICAL,"Freeing user %s (%p)\n",user->username,user);
+}
+#else
+#define _user_free wzd_free
+#endif
 
 void usercache_init(void)
 {
-  cache_user_pool = (wzd_user_t*)wzd_malloc(sizeof(wzd_user_t)*HARD_DEF_USER_MAX);
-  memset(cache_user_pool,0,sizeof(wzd_user_t)*HARD_DEF_USER_MAX);
-  cache_group_pool = (wzd_group_t*)wzd_malloc(sizeof(wzd_group_t)*HARD_DEF_GROUP_MAX);
-  memset(cache_group_pool,0,sizeof(wzd_group_t)*HARD_DEF_GROUP_MAX);
+  chtbl_init(&index_user_name,128,(hash_function)hash_str,(cmp_function)strcmp,wzd_free);
+  chtbl_init(&index_user_uid,37,(hash_function)hash_uid,(cmp_function)uidcmp,wzd_free);
+  chtbl_init(&index_group_name,128,(hash_function)hash_str,(cmp_function)strcmp,wzd_free);
+  chtbl_init(&index_group_gid,37,(hash_function)hash_uid,(cmp_function)uidcmp,wzd_free);
 }
 
 void usercache_fini(void)
 {
-  wzd_free(cache_user_pool);
-  cache_user_pool = NULL;
-  wzd_free(cache_group_pool);
-  cache_group_pool = NULL;
+  chtbl_destroy(&index_user_uid);
+  chtbl_destroy(&index_user_name);
+  chtbl_destroy(&index_group_gid);
+  chtbl_destroy(&index_group_name);
 }
 
 
 wzd_user_t * usercache_add(wzd_user_t * user)
 {
-  unsigned int i;
+  wzd_user_t * loop_user;
+  wzd_user_t * data;
 
   if (!user) return NULL;
 
-  /* first pass: we check that entry is not already present */
-  for (i=0; i < HARD_DEF_USER_MAX; i++) {
-    if (cache_user_pool[i].username[0] != '\0') {
-      if (strcmp(user->username,cache_user_pool[i].username)==0) {
-#ifdef DEBUG
-        if (user->uid != cache_user_pool[i].uid) {
-          out_log(LEVEL_NORMAL,"User with same name but different uid is already present !!\n");
-          out_log(LEVEL_NORMAL,"%s:%d     user: %s\n",__FILE__,__LINE__,user->username);
-        }
+  if ((chtbl_lookup(&index_user_name, user->username, (void**)&data))==0)
+  {
+#ifdef WZD_DBG_UGCACHE
+    out_err(LEVEL_INFO,"user cache refresh %s\n",user->username);
 #endif
-        /* found, we overwrite it */
-        memcpy(&cache_user_pool[i],user,sizeof(wzd_user_t));
-        return &cache_user_pool[i];
-      }
-    }
-  } /* for */
+    memcpy(data, user, sizeof(wzd_user_t));
+    return data;
+  }
 
-  /* second pass: we search a free index */
-  for (i=0; i < HARD_DEF_USER_MAX; i++) {
-    if (cache_user_pool[i].username[0] == '\0') {
-      /* put it in cache */
-#ifdef DEBUG
-      out_err(LEVEL_INFO,"user cache add %s\n",user->username);
+  /* insert entry */
+  loop_user = wzd_malloc(sizeof(wzd_user_t));
+  memcpy(loop_user,user,sizeof(wzd_user_t));
+  if ((chtbl_insert(&index_user_name, loop_user->username, loop_user, NULL, NULL, _user_free))==0 &&
+       chtbl_insert(&index_user_uid, (void*)loop_user->uid, loop_user, NULL, NULL, NULL)==0)
+  {
+#ifdef WZD_DBG_UGCACHE
+    out_err(LEVEL_INFO,"user cache add %s (%p)\n",user->username,loop_user);
 #endif
-      memcpy(&cache_user_pool[i],user,sizeof(wzd_user_t));
-      return &cache_user_pool[i];
-    }
-  } /* for */
+    return loop_user;
+  }
+  wzd_free(loop_user);
 
-#ifdef DEBUG
+#ifdef WZD_DBG_UGCACHE
   out_log(LEVEL_NORMAL,"No more free space in cache\n");
   out_log(LEVEL_NORMAL,"%s:%d     user: %s\n",__FILE__,__LINE__,user->username);
 #endif
   return NULL;
 }
 
-wzd_user_t * usercache_get( predicate_user_t p, void * arg )
+wzd_user_t * usercache_getbyname( const char * name )
 {
-  unsigned int i;
+  wzd_user_t * user;
 
-  if (!cache_user_pool) return NULL;
-
-  for (i=0; i < HARD_DEF_USER_MAX; i++) {
-    if (cache_user_pool[i].username[0] != '\0') {
-      /* test entry */
-      if ( (*p)(&cache_user_pool[i],arg) ) {
-#ifdef DEBUG
-        out_log(LEVEL_INFO,"user cache hit %s\n",cache_user_pool[i].username);
+  if ((chtbl_lookup(&index_user_name, name, (void**)&user))==0) {
+#ifdef WZD_DBG_UGCACHE
+    out_err(LEVEL_INFO,"user cache hit NAME %s\n",user->username);
 #endif
-        return &cache_user_pool[i];
-      }
-    }
-  } /* for */
+    return user;
+  }
+
+  return NULL;
+}
+
+wzd_user_t * usercache_getbyuid( unsigned int uid )
+{
+  wzd_user_t * user;
+
+  if ((chtbl_lookup(&index_user_uid, (void*)uid, (void**)&user))==0) {
+#ifdef WZD_DBG_UGCACHE
+    out_err(LEVEL_INFO,"user cache hit UID %s\n",user->username);
+#endif
+    return user;
+  }
+
+  return NULL;
+}
+
+wzd_user_t * usercache_search( predicate_user_t p, void * arg )
+{
+  wzd_user_t * user;
+
+  if ((chtbl_search(&index_user_name, (cmp_function)p, arg, (void**)&user))==0)
+  {
+    return user;
+  }
+
   return NULL;
 }
 
 void usercache_invalidate( predicate_user_t p, void * arg )
 {
-  unsigned int i;
+  wzd_user_t * user;
 
-  if (!arg) return;
+  user = usercache_search( p, arg );
 
-  for (i=0; i < HARD_DEF_USER_MAX; i++) {
-    if (cache_user_pool[i].username[0] != '\0') {
-      /* test entry */
-      if ( (*p)(&cache_user_pool[i],arg) )
-        memset(&cache_user_pool[i],0,sizeof(wzd_user_t));
-    }
-  } /* for */
+  if (user) chtbl_remove(&index_user_uid, (void*)user->uid);
+  if (user) chtbl_remove(&index_user_name, user->username);
 }
 
 
 
 wzd_group_t * groupcache_add(wzd_group_t * group)
 {
-  unsigned int i;
+  wzd_group_t * loop_group;
+  wzd_group_t * data;
 
   if (!group) return NULL;
 
-  /* first pass: we check that entry is not already present */
-  for (i=0; i < HARD_DEF_GROUP_MAX; i++) {
-    if (cache_group_pool[i].groupname[0] != '\0') {
-      if (strcmp(group->groupname,cache_group_pool[i].groupname)==0) {
-#ifdef DEBUG
-        if (group->gid != cache_group_pool[i].gid) {
-          out_log(LEVEL_NORMAL,"Group with same name but different gid is already present !!\n");
-          out_log(LEVEL_NORMAL,"%s:%d     group: %s\n",__FILE__,__LINE__,group->groupname);
-        }
+  if ((chtbl_lookup(&index_group_name, group->groupname, (void**)&data))==0)
+  {
+#ifdef WZD_DBG_UGCACHE
+    out_err(LEVEL_INFO,"group cache refresh %s\n",group->groupname);
 #endif
-        /* found, we overwrite it */
-        memcpy(&cache_group_pool[i],group,sizeof(wzd_group_t));
-        return &cache_group_pool[i];
-      }
-    }
-  } /* for */
+    memcpy(data, group, sizeof(wzd_group_t));
+    return data;
+  }
 
-  /* second pass: we search a free index */
-  for (i=0; i < HARD_DEF_GROUP_MAX; i++) {
-    if (cache_group_pool[i].groupname[0] == '\0') {
-      /* put it in cache */
-#ifdef DEBUG
-      out_err(LEVEL_INFO,"group cache add %s\n",group->groupname);
+  /* insert entry */
+  loop_group = wzd_malloc(sizeof(wzd_group_t));
+  memcpy(loop_group,group,sizeof(wzd_group_t));
+  if ((chtbl_insert(&index_group_name, loop_group->groupname, loop_group, NULL, NULL, wzd_free))==0 &&
+       chtbl_insert(&index_group_gid, (void*)loop_group->gid, loop_group, NULL, NULL, NULL)==0)
+  {
+#ifdef WZD_DBG_UGCACHE
+    out_err(LEVEL_INFO,"group cache add %s\n",group->groupname);
 #endif
-      memcpy(&cache_group_pool[i],group,sizeof(wzd_group_t));
-      return &cache_group_pool[i];
-    }
-  } /* for */
+    return loop_group;
+  }
+  wzd_free(loop_group);
 
-#ifdef DEBUG
+#ifdef WZD_DBG_UGCACHE
   out_log(LEVEL_NORMAL,"No more free space in cache\n");
   out_log(LEVEL_NORMAL,"%s:%d     group: %s\n",__FILE__,__LINE__,group->groupname);
 #endif
   return NULL;
 }
 
-wzd_group_t * groupcache_get( predicate_group_t p, void * arg )
+wzd_group_t * groupcache_getbyname( const char * name )
 {
-  unsigned int i;
-  for (i=0; i < HARD_DEF_GROUP_MAX; i++) {
-    if (cache_group_pool[i].groupname[0] != '\0') {
-      /* test entry */
-      if ( (*p)(&cache_group_pool[i],arg) ) {
-#ifdef DEBUG
-        out_log(LEVEL_INFO,"group cache hit %s\n",cache_group_pool[i].groupname);
+  wzd_group_t * group;
+
+  if ((chtbl_lookup(&index_group_name, name, (void**)&group))==0) {
+#ifdef WZD_DBG_UGCACHE
+    out_err(LEVEL_INFO,"group cache hit NAME %s\n",group->groupname);
 #endif
-        return &cache_group_pool[i];
-      }
-    }
-  } /* for */
+    return group;
+  }
+
+  return NULL;
+}
+
+wzd_group_t * groupcache_getbygid( unsigned int gid )
+{
+  wzd_group_t * group;
+
+  if ((chtbl_lookup(&index_group_gid, (void*)gid, (void**)&group))==0) {
+#ifdef WZD_DBG_UGCACHE
+    out_err(LEVEL_INFO,"group cache hit GID %s\n",group->groupname);
+#endif
+    return group;
+  }
+
+  return NULL;
+}
+
+wzd_group_t * groupcache_search( predicate_group_t p, void * arg )
+{
+  wzd_group_t * group;
+
+  if ((chtbl_search(&index_group_name, (cmp_function)p, arg, (void**)&group))==0)
+  {
+    return group;
+  }
+
   return NULL;
 }
 
 void groupcache_invalidate( predicate_group_t p, void * arg )
 {
-  unsigned int i;
+  wzd_group_t * group;
 
-  if (!arg) return;
+  group = groupcache_search( p, arg );
 
-  for (i=0; i < HARD_DEF_GROUP_MAX; i++) {
-    if (cache_group_pool[i].groupname[0] != '\0') {
-      /* test entry */
-      if ( (*p)(&cache_group_pool[i],arg) )
-        memset(&cache_group_pool[i],0,sizeof(wzd_group_t));
-    }
-  } /* for */
+  if (group) chtbl_remove(&index_group_gid, (void*)group->gid);
+  if (group) chtbl_remove(&index_group_name, group->groupname);
 }
 
