@@ -24,6 +24,10 @@
  * the source code for OpenSSL in the source distribution.
  */
 
+#include "wzd_all.h"
+
+#ifndef WZD_USE_PCH
+
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
@@ -31,6 +35,8 @@
 #include <sys/stat.h>
 #include <errno.h>
 
+#endif /* WZD_USE_PCH */
+
 #ifdef _MSC_VER
 #include <winsock2.h>
 #include <io.h>
@@ -41,7 +47,6 @@
 #include <tchar.h>
 #include <winioctl.h>
 
-
 // Since MS apparently removed this struct (and its documentation) from
 // the W2k SDK, but still refer to it in 'winioctl.h' for the specific
 // IOCTLs, I decided to rename it and make it available.
@@ -105,6 +110,7 @@ typedef struct
 
 #include "wzd_structs.h"
 
+#include "wzd_libmain.h"
 #include "wzd_log.h"
 #include "wzd_misc.h"
 #include "wzd_file.h"
@@ -138,6 +144,7 @@ void free_file_recursive(struct wzd_file_t * file)
   wzd_acl_line_t *acl_current,*acl_next;
 
   if (!file) return;
+  WZD_MUTEX_LOCK(SET_MUTEX_FILE_T);
   do {
     next_file = file->next_file;
     acl_current = file->acl;
@@ -152,17 +159,22 @@ void free_file_recursive(struct wzd_file_t * file)
     wzd_free (file);
     file = next_file;
   } while (file);
+  WZD_MUTEX_UNLOCK(SET_MUTEX_FILE_T);
 }
 
 static struct wzd_file_t * find_file(const char *name, struct wzd_file_t *first)
 {
   struct wzd_file_t *current=first;
 
+  WZD_MUTEX_LOCK(SET_MUTEX_FILE_T);
   while (current) {
-    if (strcmp(name,current->filename)==0)
+    if (strcmp(name,current->filename)==0) {
+      WZD_MUTEX_UNLOCK(SET_MUTEX_FILE_T);
       return current;
+    }
     current = current->next_file;
   }
+  WZD_MUTEX_UNLOCK(SET_MUTEX_FILE_T);
   return NULL;
 }
 
@@ -172,11 +184,13 @@ static struct wzd_file_t * remove_file(const char *name, struct wzd_file_t **fir
 
   if (!current) return NULL;
 
+  WZD_MUTEX_LOCK(SET_MUTEX_FILE_T);
   /* first to be removed ? */
   if (strcmp(name,current->filename)==0) {
     removed = current;
     *first = removed->next_file;
     removed->next_file = NULL;
+    WZD_MUTEX_UNLOCK(SET_MUTEX_FILE_T);
     return removed;
   }
 
@@ -188,11 +202,13 @@ static struct wzd_file_t * remove_file(const char *name, struct wzd_file_t **fir
       removed = current;
       prev->next_file = current->next_file;
       current->next_file = NULL;
+      WZD_MUTEX_UNLOCK(SET_MUTEX_FILE_T);
       return removed;
     }
     prev = current;
     current = current->next_file;
   } /* while current */
+  WZD_MUTEX_UNLOCK(SET_MUTEX_FILE_T);
   return NULL;
 }
 
@@ -206,6 +222,7 @@ void file_insert_sorted(struct wzd_file_t *entry, struct wzd_file_t **tab)
     return;
   }
 
+  WZD_MUTEX_LOCK(SET_MUTEX_FILE_T);
   while (it) {
     if (strcmp(entry->filename,it->filename)>0)
     {
@@ -220,6 +237,7 @@ void file_insert_sorted(struct wzd_file_t *entry, struct wzd_file_t **tab)
     if (itp == NULL) {
       entry->next_file = *tab;
       *tab = entry;
+      WZD_MUTEX_UNLOCK(SET_MUTEX_FILE_T);
       return;
     }
 
@@ -227,6 +245,7 @@ void file_insert_sorted(struct wzd_file_t *entry, struct wzd_file_t **tab)
     entry->next_file = it;
     itp->next_file = entry;
 
+    WZD_MUTEX_UNLOCK(SET_MUTEX_FILE_T);
     return;
   }
 
@@ -234,6 +253,7 @@ void file_insert_sorted(struct wzd_file_t *entry, struct wzd_file_t **tab)
   /* itp can't be NULL here, the first case would have trapped it */
   itp->next_file = entry;
 
+  WZD_MUTEX_UNLOCK(SET_MUTEX_FILE_T);
   return;
 }
 
@@ -242,11 +262,15 @@ wzd_acl_line_t * find_acl(const char * username, struct wzd_file_t * file)
 {
   wzd_acl_line_t *current = file->acl;
 
+  WZD_MUTEX_LOCK(SET_MUTEX_ACL_T);
   while (current) {
-    if (strcmp(username,current->user)==0)
+    if (strcmp(username,current->user)==0) {
+      WZD_MUTEX_UNLOCK(SET_MUTEX_ACL_T);
       return current;
+    }
     current = current->next_acl;
   }
+  WZD_MUTEX_UNLOCK(SET_MUTEX_ACL_T);
   return NULL;
 }
 
@@ -255,6 +279,7 @@ static struct wzd_file_t * add_new_file(const char *name, const char *owner, con
 {
   struct wzd_file_t *current, *new_file;
 
+  WZD_MUTEX_LOCK(SET_MUTEX_FILE_T);
   new_file = wzd_malloc(sizeof(struct wzd_file_t));
   strncpy(new_file->filename,name,256);
   memset(new_file->owner,0,256);
@@ -274,6 +299,7 @@ static struct wzd_file_t * add_new_file(const char *name, const char *owner, con
       current = current->next_file;
     current->next_file = new_file;
   }
+  WZD_MUTEX_UNLOCK(SET_MUTEX_FILE_T);
   return new_file;
 }
 
@@ -284,6 +310,7 @@ struct wzd_file_t * file_deep_copy(struct wzd_file_t *file_cur)
 
   if (!file_cur) return NULL;
 
+  WZD_MUTEX_LOCK(SET_MUTEX_FILE_T);
   new_file = wzd_malloc(sizeof(struct wzd_file_t));
   memcpy(new_file, file_cur, sizeof(struct wzd_file_t));
   if (file_cur->data)
@@ -308,6 +335,7 @@ struct wzd_file_t * file_deep_copy(struct wzd_file_t *file_cur)
   /* exception: we set next_file to NULL to avoid side effects */
   new_file->next_file = NULL;
 
+  WZD_MUTEX_UNLOCK(SET_MUTEX_FILE_T);
   return new_file;
 }
 
@@ -316,6 +344,8 @@ void addAcl(const char *filename, const char *user, const char *rights, struct w
 {
   wzd_acl_line_t * acl_current, * acl_new;
 
+  WZD_MUTEX_LOCK(SET_MUTEX_ACL_T);
+
   acl_new = wzd_malloc(sizeof(wzd_acl_line_t));
   strncpy(acl_new->user,user,256);
   strncpy(acl_new->perms,rights,3);
@@ -325,6 +355,7 @@ void addAcl(const char *filename, const char *user, const char *rights, struct w
   if (!acl_current) { /* simple case, first insertion */
     file->acl = acl_new;
     acl_new->next_acl = NULL;
+    WZD_MUTEX_UNLOCK(SET_MUTEX_ACL_T);
     return;
   }
 
@@ -332,6 +363,7 @@ void addAcl(const char *filename, const char *user, const char *rights, struct w
     if (strcmp(acl_current->user,user)==0) { /* found ! */
       strncpy(acl_current->perms,rights,3); /* replace old perms */
       wzd_free (acl_new);
+      WZD_MUTEX_UNLOCK(SET_MUTEX_ACL_T);
       return;
     }
     acl_current = acl_current->next_acl;
@@ -340,6 +372,7 @@ void addAcl(const char *filename, const char *user, const char *rights, struct w
   /* new acl for this file */
   acl_new->next_acl = file->acl;
   file->acl = acl_new;
+  WZD_MUTEX_UNLOCK(SET_MUTEX_ACL_T);
 }
 
 /** \todo should be "atomic" */
@@ -355,8 +388,13 @@ int readPermFile(const char *permfile, struct wzd_file_t **pTabFiles)
 
   current_file = *pTabFiles;
 
+  WZD_MUTEX_LOCK(SET_MUTEX_DIRINFO);
   fp = wzd_cache_open(permfile,O_RDONLY,0644);
-  if (!fp) { wzd_cache_close(fp); return E_FILE_NOEXIST; }
+  if (!fp) {
+    wzd_cache_close(fp);
+    WZD_MUTEX_UNLOCK(SET_MUTEX_DIRINFO);
+    return E_FILE_NOEXIST;
+  }
 
   ptr = (char*)current_file;
   current_file = NULL;
@@ -414,6 +452,8 @@ int readPermFile(const char *permfile, struct wzd_file_t **pTabFiles)
   }
 
   wzd_cache_close(fp);
+  WZD_MUTEX_UNLOCK(SET_MUTEX_DIRINFO);
+
   return E_OK;
 }
 
@@ -433,8 +473,13 @@ int writePermFile(const char *permfile, struct wzd_file_t **pTabFiles)
     return unlink(permfile);
   }
 
+  WZD_MUTEX_LOCK(SET_MUTEX_DIRINFO);
+
   fp = fopen(permfile,"w"); /* overwrite any existing file */
-  if (!fp) return -1;
+  if (!fp) {
+    WZD_MUTEX_UNLOCK(SET_MUTEX_DIRINFO);
+    return -1;
+  }
 
   /* if file_cur->filename contains spaces, we MUST quote it when writing name */
   while (file_cur) {
@@ -479,6 +524,8 @@ int writePermFile(const char *permfile, struct wzd_file_t **pTabFiles)
   /* force cache update */
   wzd_cache_update(permfile);
 
+  WZD_MUTEX_UNLOCK(SET_MUTEX_DIRINFO);
+
   return 0;
 }
 
@@ -653,7 +700,7 @@ int _checkPerm(const char *filename, unsigned long wanted_right, wzd_user_t * us
   char dir[WZD_MAX_PATH+1];
   char stripped_filename[WZD_MAX_PATH+1];
   char *ptr;
-  struct stat s;
+  struct statbuf s;
 
   if (!filename || filename[0] == '\0')
     return -1;
@@ -662,7 +709,7 @@ int _checkPerm(const char *filename, unsigned long wanted_right, wzd_user_t * us
   out_err(LEVEL_HIGH,"_checkPerm(%s,%ld,%s)\n",filename,wanted_right,user->username);
 #endif
 
-  strncpy(dir,filename,WZD_MAX_PATH); 
+  strncpy(dir,filename,WZD_MAX_PATH);
 
   if (user->flags && strchr(user->flags,FLAG_ANONYMOUS))
   {
@@ -675,7 +722,7 @@ int _checkPerm(const char *filename, unsigned long wanted_right, wzd_user_t * us
     }
   }
 
-  if (stat(filename,&s)==-1) {
+  if (fs_stat(filename,&s)==-1) {
     if (wanted_right != RIGHT_STOR && wanted_right != RIGHT_MKDIR)
       return -1; /* inexistant ? */
     ptr = strrchr(dir,'/');
@@ -743,7 +790,7 @@ int _setPerm(const char *filename, const char *granted_user, const char *owner,
   char stripped_filename[WZD_MAX_PATH+1];
   char perm_filename[WZD_MAX_PATH+1];
   char *ptr;
-  struct stat s;
+  struct statbuf s;
   size_t length, neededlength;
   struct wzd_file_t * file_list=NULL, * file_cur;
   int ret;
@@ -753,7 +800,7 @@ int _setPerm(const char *filename, const char *granted_user, const char *owner,
 
   strncpy(dir,filename,WZD_MAX_PATH);
 
-  if (stat(filename,&s)==-1) return -1; /* inexistant ? */
+  if (fs_stat(filename,&s)==-1) return -1; /* inexistant ? */
   if (S_ISDIR(s.st_mode)) { /* isdir */
     strcpy(stripped_filename,".");
   } else { /* ! isdir */
@@ -779,11 +826,12 @@ int _setPerm(const char *filename, const char *granted_user, const char *owner,
   strncpy(perm_filename+length,HARD_PERMFILE,neededlength);
 
 
-/*fprintf(stderr,"%s:%d\n",__FILE__,__LINE__);*/
-#ifdef DEBUG
+#ifdef WZD_DBG_PERMS
   out_err(LEVEL_FLOOD,"_setPerm: dir %s filename %s wanted file %s\n",dir,perm_filename,stripped_filename);
 #endif
 
+  WZD_MUTEX_LOCK(SET_MUTEX_PERMISSION);
+
   ret = readPermFile(perm_filename,&file_list);
   if (ret) { /* no permissions file */
     file_cur = add_new_file(stripped_filename,0,0,&file_list);
@@ -820,6 +868,7 @@ int _setPerm(const char *filename, const char *granted_user, const char *owner,
   free_file_recursive(file_list);
   file_list = NULL;
 
+  WZD_MUTEX_UNLOCK(SET_MUTEX_PERMISSION);
   return 0;
 }
 
@@ -832,7 +881,7 @@ int _movePerm(const char *oldfilename, const char *newfilename, const char *owne
   char dst_stripped_filename[BUFFER_LEN];
   char dst_perm_filename[BUFFER_LEN];
   char *ptr;
-  struct stat s,s2;
+  struct statbuf s,s2;
   size_t length, neededlength;
   struct wzd_file_t * src_file_list=NULL, *dst_file_list=NULL,* file_cur, *file_dst;
   wzd_acl_line_t * acl;
@@ -844,7 +893,7 @@ int _movePerm(const char *oldfilename, const char *newfilename, const char *owne
   /* find src perm file name */
   strncpy(dir,oldfilename,BUFFER_LEN);
 
-  if (stat(dir,&s)==-1) return -1; /* inexistant ? */
+  if (fs_stat(dir,&s)==-1) return -1; /* inexistant ? */
   if (S_ISDIR(s.st_mode)) { /* isdir */
     /* TODO XXX FIXME Check validity of this assertion ! */
     /* permissions of directory are self contained ! */
@@ -876,7 +925,7 @@ int _movePerm(const char *oldfilename, const char *newfilename, const char *owne
   strncpy(dir,newfilename,BUFFER_LEN);
 
   /* if dst file is a dir and exists, we can't make the operation */
-  if (stat(dir,&s2)==0) { /* file exists ? */
+  if (fs_stat(dir,&s2)==0) { /* file exists ? */
     if (S_ISDIR(s2.st_mode)) { /* isdir */
       return -1;
     }
@@ -907,12 +956,14 @@ int _movePerm(const char *oldfilename, const char *newfilename, const char *owne
 
   strncpy(dst_perm_filename+length,HARD_PERMFILE,neededlength);
 
-#ifdef DEBUG
-fprintf(stderr,"%s:%d\n",__FILE__,__LINE__);
-fprintf(stderr,"dir %s filename %s wanted file %s\n",dir,src_perm_filename,src_stripped_filename);
-fprintf(stderr,"dir %s filename %s wanted file %s\n",dir,dst_perm_filename,dst_stripped_filename);
+#ifdef WZD_DBG_PERMS
+out_err(LEVEL_FLOOD,"%s:%d\n",__FILE__,__LINE__);
+out_err(LEVEL_FLOOD,"dir %s filename %s wanted file %s\n",dir,src_perm_filename,src_stripped_filename);
+out_err(LEVEL_FLOOD,"dir %s filename %s wanted file %s\n",dir,dst_perm_filename,dst_stripped_filename);
 #endif
 
+  WZD_MUTEX_LOCK(SET_MUTEX_PERMISSION);
+
   ret = readPermFile(src_perm_filename,&src_file_list);
   if (ret) { /* no permissions file */
     file_dst = NULL;
@@ -959,6 +1010,7 @@ fprintf(stderr,"dir %s filename %s wanted file %s\n",dir,dst_perm_filename,dst_s
   free_file_recursive(dst_file_list);
   dst_file_list = NULL;
 
+  WZD_MUTEX_UNLOCK(SET_MUTEX_PERMISSION);
   return 0;
 }
 
@@ -1148,13 +1200,13 @@ int softlink_create(const char *target, const char *linkname)
   char *ptr;
   struct wzd_file_t * perm_list=NULL, * file_cur;
   int ret;
-  struct stat s;
+  struct statbuf s;
 
-  if (stat(target,&s)) { /* target does not exist ?! */
+  if (fs_stat(target,&s)) { /* target does not exist ?! */
     out_err(LEVEL_FLOOD, "symlink: source does not exist (%s)\n", target);
     return -1;
   }
-  if (stat(linkname,&s) != -1) { /* linkname already exist ?! */
+  if (fs_stat(linkname,&s) != -1) { /* linkname already exist ?! */
     out_err(LEVEL_FLOOD, "symlink: destination already exists (%s)\n", linkname);
     return -1;
   }
@@ -1174,7 +1226,7 @@ int softlink_create(const char *target, const char *linkname)
       )
     {
       *ptr = '\0';
-      if (stat(perm_filename,&s)) {
+      if (fs_stat(perm_filename,&s)) {
         out_err(LEVEL_FLOOD, "symlink: destination directory does not exist (%s)\n", perm_filename);
         return -1;
       }
@@ -1185,6 +1237,7 @@ int softlink_create(const char *target, const char *linkname)
   strncpy(stripped_filename, ptr, WZD_MAX_PATH);
   strncpy(ptr, HARD_PERMFILE, WZD_MAX_PATH - (ptr-perm_filename));
 
+  WZD_MUTEX_LOCK(SET_MUTEX_PERMISSION);
   /* read perm file */
   ret = readPermFile(perm_filename,&perm_list);
 
@@ -1197,6 +1250,7 @@ int softlink_create(const char *target, const char *linkname)
       /* error, an entry already exists with the same name */
       out_err(LEVEL_FLOOD, "symlink: link already exists here (%s)\n", perm_filename);
       free_file_recursive(perm_list);
+      WZD_MUTEX_UNLOCK(SET_MUTEX_PERMISSION);
       return EEXIST;
     }
     file_cur = add_new_file(stripped_filename, 0, 0, &perm_list);
@@ -1215,6 +1269,7 @@ int softlink_create(const char *target, const char *linkname)
 
   free_file_recursive(perm_list);
   perm_list = NULL;
+  WZD_MUTEX_UNLOCK(SET_MUTEX_PERMISSION);
 
   return 0;
 }
@@ -1241,6 +1296,7 @@ int softlink_remove(const char *linkname)
   strncpy(stripped_filename, ptr, WZD_MAX_PATH);
   strncpy(ptr, HARD_PERMFILE, WZD_MAX_PATH - (ptr-perm_filename));
 
+  WZD_MUTEX_LOCK(SET_MUTEX_PERMISSION);
   /* read perm file */
   ret = readPermFile(perm_filename,&perm_list);
 
@@ -1251,7 +1307,8 @@ int softlink_remove(const char *linkname)
     {
       free_file_recursive(perm_list);
       out_err(LEVEL_FLOOD, "symlink: trying to remove something that is not a link (%s)\n", linkname);
-     return -1;
+      WZD_MUTEX_UNLOCK(SET_MUTEX_PERMISSION);
+      return -1;
     }
 
     file_cur = remove_file(stripped_filename, &perm_list);
@@ -1264,6 +1321,7 @@ int softlink_remove(const char *linkname)
   }
 
   perm_list = NULL;
+  WZD_MUTEX_UNLOCK(SET_MUTEX_PERMISSION);
 
   return 0;
 }
@@ -1291,7 +1349,7 @@ int file_open(const char *filename, int mode, unsigned long wanted_right, wzd_co
   mode |= _O_BINARY;
 #endif
 
-  fd = open(filename,mode,0666);
+  fd = fs_open(filename,mode,0666);
   if (fd == -1) {
     fprintf(stderr,"Can't open %s,errno %d : %s\n",filename,errno,strerror(errno));
     return -1;
@@ -1332,9 +1390,9 @@ void file_close(int fd, wzd_context_t * context)
   close(fd);
 }
 
-off_t file_seek(int fd, unsigned long offset, int whence)
+fs_off_t file_seek(fd_t fd, fs_off_t offset, int whence)
 {
-  return lseek(fd,offset,whence);
+  return fs_lseek(fd,offset,whence);
 }
 
 /** NOTE:
@@ -1368,7 +1426,7 @@ int file_rmdir(const char *dirname, wzd_context_t * context)
 {
   int ret;
   wzd_user_t * user;
-  struct stat s;
+  struct statbuf s;
 
   user = GetUserByID(context->userid);
 
@@ -1376,7 +1434,7 @@ int file_rmdir(const char *dirname, wzd_context_t * context)
   if (ret) return -1;
 
   /* is a directory ? */
-  if (stat(dirname,&s)) return -1;
+  if (fs_stat(dirname,&s)) return -1;
   if (!S_ISDIR(s.st_mode)) return -1;
 
   /* is dir empty ? */
@@ -1432,8 +1490,8 @@ fprintf(stderr,"Removing directory '%s'\n",dirname);
 
 #ifndef __CYGWIN__
   {
-    struct stat s;
-    lstat(dirname,&s);
+    struct statbuf s;
+    fs_lstat(dirname,&s);
     if (S_ISLNK(s.st_mode))
       return unlink(dirname);
   }
@@ -1524,14 +1582,25 @@ int file_remove(const char *filename, wzd_context_t * context)
       file_cur = NULL;
     }
 
-    /* check in special permissions from config file */
-    if (perm_check("delete", context, mainConfig) == 0) ret = 0; /* specified in config file */
+    /* The delete permission is special: by default, all users can run the DELE
+     * command, but only owner of the file or siteops can delete files.
+     * If the "delete" permission is set, it will allow non-owners to delete the file.
+     */
+    {
+      wzd_command_t * command;
+      wzd_string_t * str = STR("delete");
+
+      command = commands_find(mainConfig->commands_list, str);
+      str_deallocate(str);
+      if (commands_check_permission(command, context) == 0) ret = 0;
+    }
 
   }
 
   if (ret)
     return 1;
 
+  WZD_MUTEX_LOCK(SET_MUTEX_PERMISSION);
   /* remove name in perm file !! */
   ret = readPermFile(perm_filename,&file_list);
   if (!ret) {
@@ -1545,8 +1614,10 @@ int file_remove(const char *filename, wzd_context_t * context)
 #ifdef DEBUG
 fprintf(stderr,"remove error %d (%s)\n", errno, strerror(errno));
 #endif
+    WZD_MUTEX_UNLOCK(SET_MUTEX_PERMISSION);
     return 1;
   }
+  WZD_MUTEX_UNLOCK(SET_MUTEX_PERMISSION);
 
   return 0;
 }
@@ -1559,9 +1630,9 @@ wzd_user_t * file_getowner(const char *filename, wzd_context_t * context)
   int ret;
   struct wzd_file_t * file_list=NULL, * file_cur;
   size_t neededlength, length;
-  struct stat s;
+  struct statbuf s;
 
-  if (stat(filename,&s))
+  if (fs_stat(filename,&s))
     return NULL;
 
   /* find the dir containing the perms file */
@@ -1631,11 +1702,11 @@ struct wzd_file_t * file_stat(const char *filename, wzd_context_t * context)
   char * ptr;
   struct wzd_file_t * file_list=NULL, * file_cur, *file;
   size_t neededlength, length;
-  struct stat s;
+  struct statbuf s;
   int nx=0;
 
   /** \bug no no no, it can be a symlink or a vfs ! */
-/*  if (stat(filename,&s))
+/*  if (fs_stat(filename,&s))
     return NULL;*/
 
   file = NULL;
@@ -1648,7 +1719,7 @@ struct wzd_file_t * file_stat(const char *filename, wzd_context_t * context)
   ptr = strrchr(perm_filename,'/');
   if (!ptr || *(ptr+1)=='\0') return NULL;
 
-  if (!lstat(filename,&s)) {
+  if (!fs_lstat(filename,&s)) {
     if (S_ISDIR(s.st_mode)) { /* isdir */
       strcpy(stripped_filename,".");
     } else { /* ! isdir */
@@ -1664,7 +1735,7 @@ struct wzd_file_t * file_stat(const char *filename, wzd_context_t * context)
     if (ptr) {
       strcpy(stripped_filename,ptr+1);
       *ptr = 0;
-      if (lstat(perm_filename,&s)) {
+      if (fs_lstat(perm_filename,&s)) {
         out_err(LEVEL_FLOOD, "symlink: destination directory does not exist (%s)\n", perm_filename);
         return NULL;
       }
@@ -1703,7 +1774,7 @@ struct wzd_file_t * file_stat(const char *filename, wzd_context_t * context)
 /* if program crash, locks acquired by fcntl (POSIX) or _locking (VISUAL)
  * are released, and then do are less annoying.
  */
-int file_lock(int fd, short lock_mode)
+int file_lock(fd_t fd, short lock_mode)
 {
 #ifdef WZD_DBG_LOCK
 fprintf(stderr,"Locking file %d\n",fd);
@@ -1724,7 +1795,7 @@ fprintf(stderr,"Locking file %d\n",fd);
   return 0;
 }
 
-int file_unlock(int fd)
+int file_unlock(fd_t fd)
 {
 #ifdef WZD_DBG_LOCK
 fprintf(stderr,"Unlocking file %d\n",fd);
@@ -1745,7 +1816,7 @@ fprintf(stderr,"Unlocking file %d\n",fd);
   return 0;
 }
 
-int file_islocked(int fd, short lock_mode)
+int file_islocked(fd_t fd, short lock_mode)
 {
 #ifdef WZD_DBG_LOCK
 fprintf(stderr,"Testing lock for file %d\n",fd);
@@ -1807,12 +1878,12 @@ fprintf(stderr,"Forcing unlock file %s\n",file);
 }
 
 /* wrappers just to keep things in same memory zones */
-ssize_t file_read(int fd,void *data,size_t length)
+ssize_t file_read(fd_t fd,void *data,size_t length)
 {
   return read(fd,data,length);
 }
 
-ssize_t file_write(int fd,const void *data,size_t length)
+ssize_t file_write(fd_t fd,const void *data,size_t length)
 {
   return write(fd,data,length);
 }
@@ -1837,9 +1908,9 @@ int symlink_create(const char *existing, const char *link)
 int symlink_remove(const char *link)
 {
 #ifndef WIN32
-  struct stat s;
+  struct statbuf s;
 
-  if (lstat(link,&s)) return E_FILE_NOEXIST;
+  if (fs_lstat(link,&s)) return E_FILE_NOEXIST;
   if ( !S_ISLNK(s.st_mode) ) return E_FILE_TYPE;
   return unlink(link);
 #else
