@@ -25,15 +25,53 @@
 #ifndef __WZD_LIBMAIN__
 #define __WZD_LIBMAIN__
 
+#include "wzd_mutex.h"
+
+typedef enum {
+  SET_MUTEX_GLOBAL=0,
+
+  SET_MUTEX_SHVARS,
+
+  SET_MUTEX_LIMITER,
+
+  SET_MUTEX_PERMISSION,
+  SET_MUTEX_DIRINFO,
+  SET_MUTEX_FILE_T,
+  SET_MUTEX_ACL_T,
+
+  SET_MUTEX_NUM /* must be last */
+} wzd_set_mutext_t;
+
+WZDIMPORT extern wzd_mutex_t * limiter_mutex;
+WZDIMPORT extern wzd_mutex_t * server_mutex;
+WZDIMPORT extern time_t server_time;
+WZDIMPORT extern wzd_mutex_t * mutex_set[SET_MUTEX_NUM];
+
+void server_restart(int signum);
+
+#define WZD_MUTEX_LOCK(x) wzd_mutex_lock(mutex_set[x])
+#define WZD_MUTEX_UNLOCK(x) wzd_mutex_unlock(mutex_set[x])
+
+
+int server_mutex_set_init(void);
+int server_mutex_set_fini(void);
+
 wzd_config_t * getlib_mainConfig(void);
 void setlib_mainConfig(wzd_config_t *);
 
-wzd_context_t * getlib_contextList(void);
-void setlib_contextList(wzd_context_t *);
+List * getlib_contextList(void);
+void setlib_contextList(List *);
+
+gid_t getlib_server_gid(void);
+void setlib_server_gid(gid_t);
 
 int getlib_server_uid(void);
 void setlib_server_uid(int);
 
 void libtest(void);
 
+
+/** \brief remove a context from the list */
+int context_remove(List * context_list, wzd_context_t * context);
+
 #endif /* __WZD_LIBMAIN__ */
