@@ -24,6 +24,14 @@
  * the source code for OpenSSL in the source distribution.
  */
 
+#ifdef HAVE_CONFIG_H
+# include "../config.h"
+#endif
+
+#include "wzd_all.h"
+
+#ifndef WZD_USE_PCH
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -63,24 +71,38 @@
 #ifndef _MSC_VER
 #include <sys/param.h>
 
-#ifdef BSD
-/* statfs */
-#include <sys/mount.h>
-#else
+#if HAVE_SYS_STATFS_H
+#include <sys/statfs.h> /* statfs */
+#endif
+
+#if HAVE_SYS_STATVFS_H
+#include <sys/statvfs.h> /* statfs */
+#endif
+
+#if HAVE_SYS_VFS_H
 #include <sys/vfs.h> /* statfs */
 #endif
+
+#if HAVE_SYS_MOUNT_H
+#include <sys/mount.h> /* statfs */
+#endif
+
 #endif /* _MSC_VER */
 
 
 #include "wzd_structs.h"
 
+#include "wzd_libmain.h"
 #include "wzd_log.h"
 #include "wzd_misc.h"
 #include "wzd_messages.h"
 #include "wzd_mutex.h"
 #include "wzd_ServerThread.h"
 
-#if defined(WIN32) || defined(BSD)
+#endif /* WZD_USE_PCH */
+
+
+#if defined(WIN32) || defined(BSD) || defined(__sun__)
 #define LONGBITS  0x20
 #else
 /* needed  for LONGBITS */
@@ -94,7 +116,7 @@
 /** Compute the hash value for the given string.  The algorithm
  * is taken from [Aho,Sethi,Ullman], modified to reduce the number of
  * collisions for short strings with very varied bit patterns.
- * See http://www.clisp.org/haible/hashfunc.html.
+ * See http://www.clisp.org/haible/hashfunc.html
  */
 unsigned long compute_hashval (const void *key, size_t keylen)
 {
@@ -233,7 +255,7 @@ int get_system_ip(const char * itface, struct in_addr * ina)
 
   return 0;
 #endif
-#if BSD || defined(_MSC_VER)
+#if BSD || defined(_MSC_VER) || defined(__sun__)
   return -1;
 #else
 /*  struct in_addr *ina = void_in;*/
@@ -244,7 +266,7 @@ int get_system_ip(const char * itface, struct in_addr * ina)
     out_log(LEVEL_CRITICAL,"Can't create new socket (%s:%d)\n",__FILE__,__LINE__);
     ina->s_addr = 0;
     return -1;
-  } 
+  }
   memset(&ifr,0,sizeof(ifr));
   strncpy(ifr.ifr_name,itface,sizeof(ifr.ifr_name));
 
@@ -265,6 +287,20 @@ int get_system_ip(const char * itface, struct in_addr * ina)
 /** returns info on device containing dir/file */
 int get_device_info(const char *file, long * f_type, long * f_bsize, long * f_blocks, long *f_free)
 {
+#ifdef HAVE_STATVFS
+  struct statvfs fs;
+
+  if (statvfs(file,&fs)==0) {
+    if (f_bsize) *f_bsize = fs.f_bsize;
+    if (f_type) *f_type = -1; /* not filled in statvfs */
+    if (f_blocks) *f_blocks = fs.f_blocks;
+    if (f_free) *f_free = fs.f_bavail; /* f_bavail: free blocks avail to non-superuser */
+
+    return 0;
+  }
+
+#else /* HAVE_STATVFS */
+
 #ifndef WIN32
   struct statfs fs;
 
@@ -300,16 +336,18 @@ int get_device_info(const char *file, long * f_type, long * f_bsize, long * f_bl
     if (f_blocks) *f_blocks = df.total_clusters * df.sectors_per_cluster;
   }
 #endif
+
+#endif /* HAVE_STATVFS */
   return -1;
 }
 
 /** internal fct, rename files by copying data */
 static int _int_rename(const char * src, const char *dst)
 {
-  struct stat s;
+  struct statbuf s;
   int ret;
 
-  if (lstat(src,&s)) return -1;
+  if (fs_lstat(src,&s)) return -1;
 
   if (S_ISDIR(s.st_mode)) {
     char buf_src[2048];
@@ -427,7 +465,7 @@ int safe_rename(const char *src, const char *dst)
     fprintf(stderr,"Cross device move\n");
     ret = _int_rename(src,dst);
   }
-  
+
   return ret;
 }
 
@@ -440,11 +478,9 @@ int is_hidden_file(const char *filename)
   if (ptr) {
     if (strcasecmp(ptr+1,HARD_PERMFILE)==0) return 1;
     if (*(ptr+1)=='.' && CFG_GET_OPTION(mainConfig,CFG_OPT_HIDE_DOTTED_FILES)) return 1;
-    if (mainConfig->dir_message[0]!='\0' && strcasecmp(ptr+1,mainConfig->dir_message)==0) return 1;
   } else {
     if (strcasecmp(filename,HARD_PERMFILE)==0) return 1;
     if (filename[0]=='.' && CFG_GET_OPTION(mainConfig,CFG_OPT_HIDE_DOTTED_FILES)) return 1;
-    if (mainConfig->dir_message[0]!='\0' && strcasecmp(filename,mainConfig->dir_message)==0) return 1;
   }
   return 0;
 }
@@ -469,15 +505,15 @@ int is_perm_file(const char *filename)
 /** get file last change time */
 time_t get_file_ctime(const char *file)
 {
-  struct stat s;
-  if ( stat(file,&s) < 0 ) return (time_t)-1;
+  struct statbuf s;
+  if ( fs_stat(file,&s) < 0 ) return (time_t)-1;
   return s.st_ctime;
 }
 
 time_t lget_file_ctime(int fd)
 {
-  struct stat s;
-  if ( fstat(fd,&s) < 0 ) return (time_t)-1;
+  struct statbuf s;
+  if ( fs_fstat(fd,&s) < 0 ) return (time_t)-1;
   return s.st_ctime;
 }
 
@@ -539,6 +575,10 @@ void server_clear_param(wzd_param_t **plist)
     if (current->name) free(current->name);
     if (current->param) free(current->param);
 
+#ifdef DEBUG
+    memset(current,0,sizeof(wzd_param_t));
+#endif
+
     free(current);
 
     current = next;
@@ -568,13 +608,25 @@ char * safe_vsnprintf(const char *format, va_list ap)
   int size = WORK_BUF_LEN;
   char * buffer = wzd_malloc(size);
   int result;
+  va_list ap2;
 
-  result = vsnprintf(buffer, size, format, ap);
+  /** vsnprintf modifies its last argument on some archs, we have to
+   * work on a copy of the va_list
+   */
+/*  va_copy(ap2,ap);*/
+  /** (autoconf.info)Function Portability: better use memcpy than va_copy
+   */
+  memcpy (&ap2,&ap, sizeof(va_list));
+
+  result = vsnprintf(buffer, size, format, ap2);
   if (result >= size)
   {
     buffer = wzd_realloc(buffer, result+1);
-    result = vsnprintf(buffer, result+1, format, ap);
+    va_end(ap2);
+    memcpy (&ap2,&ap, sizeof(va_list));
+    result = vsnprintf(buffer, result+1, format, ap2);
   }
+  va_end(ap2);
 
   return buffer;
 }
@@ -582,168 +634,128 @@ char * safe_vsnprintf(const char *format, va_list ap)
 /** if code is negative, the last line will NOT be formatted as the end
  * of a normal ftp reply
  */
-void v_format_message(int code, unsigned int *plength, char **pbuffer, va_list argptr)
+wzd_string_t * v_format_message(wzd_context_t * context, int code, va_list argptr)
 {
-  const char * token, *token2;
-  const char * msg;
-  char *ptr;
-  unsigned int size;
-  char work_buf[WORK_BUF_LEN];
-  char cookies_buf[WORK_BUF_LEN];
-  char is_terminated=1;
-  int must_free;
-  int ret;
+  wzd_string_t * str = NULL;
   wzd_user_t * user;
   wzd_group_t * group;
-  wzd_context_t * context;
-  /* XXX 4096 should ALWAYS be >= length */
-  char * old_buffer;
-  unsigned int length;
-  char * buffer;
+  char * cookies_buf, * work_buf;
+  char * token, * ptr;
+  const char * msg;
+  int must_free;
+  u16_t is_terminated=1;
+  int ret;
 
-#if 0
-#ifdef DEBUG
-  if (length > WORK_BUF_LEN) {
-    out_err(LEVEL_HIGH,"*** WARNING *** message too long, will be truncated\n");
-    length = WORK_BUF_LEN;
-  }
-#endif
-#endif
+  if (!context) return NULL;
 
   if (code < 0) {
     is_terminated = 0;
-    code = abs(code);
+    code = (-code);
   }
 
-  msg = getMessage(code,&must_free);
-  ptr = work_buf;
-
-  context = GetMyContext();
   user = GetUserByID(context->userid);
   group = user ? GetGroupByID(user->groups[0]) : NULL;
 
+  msg = getMessage(code,&must_free);
+
   /* first, replace cookies */
-  ret = cookie_parse_buffer(msg, user, group, context, cookies_buf, WORK_BUF_LEN);
-  
+  cookies_buf = wzd_malloc(WORK_BUF_LEN+1);
+  ret = cookie_parse_buffer(msg, user, group, context, cookies_buf, WORK_BUF_LEN); /** \todo use wzd_string_t here */
+
   /* then format message */
-  vsnprintf(work_buf,WORK_BUF_LEN,cookies_buf,argptr);
+  work_buf = safe_vsnprintf(cookies_buf,argptr);
+  wzd_free(cookies_buf);
 
-  if (must_free) {
-    free ( (char*)msg );
-  }
+  /* we don't need msg anymore */
+  if (must_free) wzd_free( (char*) msg );
 
-  length = (strlen(work_buf)*150)/100 + 11; /* empirical ratio: after/before interpreting cookies */
-  buffer = malloc(length);
-  *pbuffer = buffer;
-  *plength = length;
-  old_buffer = buffer;
+  str = str_allocate();
 
-  /* adjust size, we will need more space to put the code and \r\n */
-  length -= 7;
+  ptr = work_buf;
+  token = strtok_r(work_buf, "\r\n", &ptr);
 
-  /* remove trailing garbage */
-  {
-    char * ptr = work_buf;
-    unsigned int length = strlen(ptr);
-    while ( *(ptr+length-1) == '\r' || *(ptr+length-1) == '\n') {
-       *(ptr+length-1) = '\0';
-       length--;
-    }
-  }
-  
-  if (!strpbrk(work_buf,"\r\n")) { /* simple case, msg on one line */
+  if (!token || strcmp(ptr,"\n")==0) { /* easy, one line */
     if (is_terminated)
-      snprintf(buffer,length,"%d %s\r\n",code,work_buf);
+      ret = str_sprintf(str,"%d %s\r\n",code,work_buf);
     else
-      snprintf(buffer,length,"%d-%s\r\n",code,work_buf);
+      ret = str_sprintf(str,"%d-%s\r\n",code,work_buf);
+    if (ret < 0) goto lbl_v_format_message;
   }
   else { /* funnier, multiline */
-    /* find first line break */
-    token = strtok_r(work_buf,"\r\n",&ptr);
 
-    /* first line begins by 123- */
-    snprintf(buffer,length,"%d-%s\r\n",code,token);
-    size = strlen(token);
-    length = length - size - 6;
-    buffer = buffer + size + 6;
+    /* first line begins with 123- */
+    str_sprintf(str,"%d-%s\r\n",code,token);
 
-    /* next line */
-    token = strtok_r(NULL,"\r\n",&ptr);
-
-    while (1) {
-      size = strlen(token);
-      /* find next token */
-      token2 = strtok_r(NULL,"\r\n",&ptr);
-      if (!token2) {
-        /* check size for BOF here ! */
-/*fprintf(stderr,"last: remaining %d (written %d, wants %d)\n",length,strlen(old_buffer),size+6);*/
-        if (size+6 >= length) {
-          out_err(LEVEL_CRITICAL,"Mayday, we're running into a BOF (%s:%d)\n",__FILE__,__LINE__);
-          snprintf(old_buffer,20,"%d Truncated\r\n",code);
-          break;
-        }
-        if (is_terminated) /* no more line, remove the - */
-          snprintf(buffer,length,"%d %s\r\n",code,token);
+    while ( (token=strtok_r(NULL, "\r\n", &ptr)) ) {
+      if (strcmp(ptr,"\n")==0) { /* last line */
+        wzd_string_t * end;
+        end = str_allocate();
+        if (is_terminated)
+          ret = str_sprintf(end,"%d %s\r\n",code,token);
         else
-          snprintf(buffer,length,"%d-%s\r\n",code,token);
+          ret = str_sprintf(end,"%d-%s\r\n",code,token);
+        str_append(str,str_tochar(end));
+        str_deallocate(end);
         break;
       }
-      /* check remaining size */
-      /* check size for BOF here ! */
-/*fprintf(stderr,"remaining %d (written %d, wants %d)\n",length,strlen(old_buffer),size+2);*/
-      if (size+2 >= length) {
-        out_err(LEVEL_CRITICAL,"Mayday, we're running into a BOF (%s:%d)\n",__FILE__,__LINE__);
-        snprintf(old_buffer,20,"%d Truncated\r\n",code);
-        break;
-      }
-      /* copy line into out buffer */
-      snprintf(buffer,length,"%s\r\n",token);
-      /* adjust length */
-      length = length - size - 2;
-      /* adjust buffer position */
-      buffer = buffer + size + 2;
-      /* loop */
-      token = token2;
+      str_append(str,token);
+      str_append(str,"\r\n");
     }
+
   }
+
+  wzd_free(work_buf);
+  return str;
+
+lbl_v_format_message:
+  wzd_free(work_buf);
+  str_deallocate(str);
+  return NULL;
 }
 
-void format_message(int code, unsigned int *plength, char **pbuffer, ...)
+wzd_string_t * format_message(wzd_context_t * context, int code, ...)
 {
   va_list argptr;
+  wzd_string_t * str;
 
-  va_start(argptr,pbuffer); /* note: ansi compatible version of va_start */
+  va_start(argptr,code); /* note: ansi compatible version of va_start */
 
-  v_format_message(code,plength,pbuffer,argptr);
+  str = v_format_message(context, code, argptr);
   va_end (argptr);
+
+  return str;
 }
 
 
+
 /************* BANDWIDTH LIMITATION *********/
 
-unsigned long get_bandwidth(void)
+unsigned long get_bandwidth(unsigned long *dl, unsigned long *ul)
 {
-  unsigned long bandwidth=0;
-  unsigned int i;
+  unsigned long ul_bandwidth=0, dl_bandwidth=0;
   unsigned int id;
+  ListElmt * elmnt;
   wzd_user_t * user;
   wzd_context_t * context;
 
-  for (i=0; i<HARD_USERLIMIT; i++) {
-    if (context_list[i].magic == CONTEXT_MAGIC) {
-      context = &context_list[i];
-      id = context_list[i].userid;
+  for (elmnt=list_head(context_list); elmnt!=NULL; elmnt=list_next(elmnt)) {
+    context = list_data(elmnt);
+    if (context && context->magic == CONTEXT_MAGIC) {
+      id = context->userid;
       user = GetUserByID(id);
       if (strncasecmp(context->last_command,"retr",4)==0) {
-        bandwidth += (unsigned long)context->current_dl_limiter.current_speed;
+        dl_bandwidth += (unsigned long)context->current_dl_limiter.current_speed;
       }
       if (strncasecmp(context->last_command,"stor",4)==0) {
-        bandwidth += (unsigned long)context->current_ul_limiter.current_speed;
+        ul_bandwidth += (unsigned long)context->current_ul_limiter.current_speed;
       }
     } /* if CONTEXT_MAGIC */
   } /* forall contexts */
 
-  return bandwidth;
+  if (dl) *dl = dl_bandwidth;
+  if (ul) *ul = ul_bandwidth;
+
+  return (dl_bandwidth+ul_bandwidth);
 }
 
 
@@ -774,9 +786,9 @@ void limiter_add_bytes(wzd_bw_limiter *l, wzd_mutex_t * mutex, int byte_count, i
 #else
   struct _timeb tb;
 #endif
-  double elapsed;
-  double pause_time;
-  double rate_ratio;
+  unsigned long elapsed;
+  unsigned long pause_time;
+/*  double rate_ratio;*/
   unsigned int bw_rate;
 
   if (!l) return;
@@ -792,28 +804,30 @@ wzd_mutex_unlock(mutex);
   {*/
 #ifndef _MSC_VER
     gettimeofday( &tv, &tz );
-    elapsed = (double) (tv.tv_sec - l->current_time.tv_sec);
-    elapsed += (double) (tv.tv_usec - l->current_time.tv_usec) / (double)1000000;
+    elapsed = (tv.tv_sec - l->current_time.tv_sec) * 1000000;
+    elapsed += (tv.tv_usec - l->current_time.tv_usec);
 #else
     _ftime(&tb);
-    elapsed = (double) (tb.time - l->current_time.time);
-    elapsed += (double) (tb.millitm - l->current_time.millitm) / (double)1000;
+    elapsed = (tb.time - l->current_time.time) * 1000000;
+    elapsed += (tb.millitm - l->current_time.millitm) * 1000;
 #endif
-    if (elapsed==(double)0) elapsed=0.01;
+    if (elapsed==0) elapsed=1;
 /*    bw_rate = (unsigned int)((double)l->bytes_transfered / elapsed);*/
-    l->current_speed = (float)((double)l->bytes_transfered / elapsed);
+    l->current_speed = (float)( (l->bytes_transfered * 1000000.f) / elapsed);
     bw_rate = (unsigned int)l->current_speed;
 /*  }*/
-/*fprintf(stderr,"speed: %d max:%d\n",bw_rate,l->maxspeed);*/
   if (l->maxspeed == 0 || bw_rate <= l->maxspeed) {
     return;
   }
-  rate_ratio = (double)bw_rate / (double)l->maxspeed;
-  pause_time = (rate_ratio - (double)1)*elapsed;
+/*fprintf(stderr,"received: %d\n",l->bytes_transfered);*/
+/*fprintf(stderr,"speed: %d max:%d\n",bw_rate,l->maxspeed);*/
+/*  rate_ratio = (double)bw_rate / (double)l->maxspeed;*/
+/*  pause_time = (rate_ratio - (double)1)*elapsed;*/
+  pause_time = (bw_rate-l->maxspeed) * (elapsed / l->maxspeed);
 #ifndef _MSC_VER
-  usleep ((unsigned long)(pause_time * (double)1000000));
+  usleep (pause_time);
 #else
-  Sleep((unsigned long)(pause_time * (double)1000));
+  Sleep((unsigned long)(pause_time / (double)1000));
 #endif
 /*  gettimeofday( &tv, &tz );
   l->current_time.tv_sec = tv.tv_sec;
@@ -848,7 +862,7 @@ int print_file(const char *filename, int code, void * void_context)
     out_log(LEVEL_HIGH,"Problem opening file %s (code %d)\n",filename,code);
     return 1;
   }
-  
+
   snprintf(complete_buffer,5,"%3d-",code);
   if ( (fgets(buffer,1018,fp))==NULL ) {
     out_log(LEVEL_HIGH,"File %s is empty (code %d)\n",filename,code);
@@ -936,13 +950,13 @@ int my_str_compare(const char * src, const char *dst)
     ptr_dst++;
     ptr_src++;
   }
-  
+
   /* test if checking was complete */
   if (*ptr_dst == '\0' || (*ptr_dst=='*' && *(ptr_dst+1)=='\0') ) return 1;
 
   return 0;
 }
-  
+
 /* lower only characters in A-Z ! */
 void ascii_lower(char * s, unsigned int length)
 {
@@ -1017,6 +1031,67 @@ void win_normalize(char * s, unsigned int length, unsigned int lower)
 
 
 
+/* \return 0 if ok, -1 if error, 1 if trying to kill myself */
+int kill_child(unsigned long pid, wzd_context_t * context)
+{
+  ListElmt * elmnt;
+  wzd_context_t * loop_context;
+  int found=0;
+#ifndef WIN32
+  int ret;
+#endif
+
+  /* preliminary check: i can't kill myself */
+  if (pid==context->pid_child) return 1;
+
+  /* checks that pid is really one of the users */
+  for (elmnt=list_head(context_list); elmnt!=NULL; elmnt=list_next(elmnt))
+  {
+    loop_context = list_data(elmnt);
+    if (loop_context && loop_context->magic == CONTEXT_MAGIC && loop_context->pid_child == pid) { found = 1; break; }
+  }
+  if (!found) return -1;
+
+#ifdef _MSC_VER
+  /* \todo XXX FIXME remove/fix test !! */
+  loop_context->exitclient = 1;
+/*  ret = TerminateThread((HANDLE)pid,0);*/
+#else
+  ret = pthread_cancel(pid);
+#endif
+
+  return 0;
+}
+
+/* \return 0 if ok, -1 if error, 1 if trying to kill myself */
+int kill_child_new(unsigned long pid, wzd_context_t * context)
+{
+  ListElmt * elmnt;
+  int found=0;
+
+  /* preliminary check: i can't kill myself */
+  if (pid==context->pid_child) return 1;
+
+  /* checks that pid is really one of the users */
+  for (elmnt=list_head(context_list); elmnt!=NULL; elmnt=list_next(elmnt))
+  {
+    context = list_data(elmnt);
+    if (context && context->magic == CONTEXT_MAGIC && context->pid_child == pid) { found = 1; break; }
+  }
+  if (!found) return -1;
+
+  /* \todo XXX FIXME remove/fix test !! */
+  context->exitclient = 1;
+/*  ret = TerminateThread((HANDLE)pid,0);*/
+/*  ret = pthread_cancel(pid);*/
+
+  return 0;
+}
+
+
+
+
+
 short is_user_in_group(wzd_user_t * user, unsigned int gid)
 {
   unsigned int i;
@@ -1047,7 +1122,7 @@ int group_remove_user(wzd_user_t * user, unsigned int gid)
     user->groups[i] = user->groups[i+1];
   }
   user->group_num--;
-  
+
   return 0;
 }
 
@@ -1055,61 +1130,36 @@ int group_remove_user(wzd_user_t * user, unsigned int gid)
 /** wrappers to context list */
 void * GetMyContext(void)
 {
-  int i;
-#ifdef WZD_MULTIPROCESS
-  wzd_context_t * context=NULL;
-  pid_t pid;
-
-  pid = getpid();
-
-  context = &context_list[0];
-  /* TODO search context list and cleanup context */
-  for (i=0; i<HARD_USERLIMIT; i++)
-  {
-    if (context_list[i].magic == CONTEXT_MAGIC && context_list[i].pid_child == pid) {
-      return (&context_list[i]);
-    }
-  }
-
-#elif defined(WZD_MULTITHREAD)
+  ListElmt * elmnt;
   wzd_context_t * context=NULL;
 
 #ifdef _MSC_VER
   unsigned long thread_id;
 
   thread_id = (unsigned long)GetCurrentThreadId();
-  context = &context_list[0];
   /* TODO search context list and cleanup context */
-  for (i=0; i<HARD_USERLIMIT; i++)
+  for (elmnt=list_head(context_list); elmnt!=NULL; elmnt=list_next(elmnt))
   {
-    if (context_list[i].magic == CONTEXT_MAGIC && context_list[i].thread_id == thread_id) {
-      return (&context_list[i]);
+    context = list_data(elmnt);
+    if (context && context->magic == CONTEXT_MAGIC && context->thread_id == thread_id) {
+      return context;
     }
   }
 #else /* _MSC_VER */
   pthread_t thread_id;
 
   thread_id = pthread_self();
-  context = &context_list[0];
   /* TODO search context list and cleanup context */
-  for (i=0; i<HARD_USERLIMIT; i++)
+  for (elmnt=list_head(context_list); elmnt!=NULL; elmnt=list_next(elmnt))
   {
-    if (context_list[i].magic == CONTEXT_MAGIC &&
-      pthread_equal((pthread_t)context_list[i].thread_id,thread_id)) {
-        return (&context_list[i]);
+    context = list_data(elmnt);
+    if (context && context->magic == CONTEXT_MAGIC &&
+      pthread_equal((pthread_t)context->thread_id,thread_id)) {
+        return context;
     }
   }
 #endif /* _MSC_VER */
 
-#else
-  /* we have only one process */
-  for (i=0; i<HARD_USERLIMIT; i++)
-  {
-    if (context_list[i].magic == CONTEXT_MAGIC)
-      return (&context_list[i]);
-  }
-#endif /* WZD_MULTITHREAD */
-
   return NULL;
 }
 
