@@ -24,6 +24,10 @@
  * the source code for OpenSSL in the source distribution.
  */
 
+#include "wzd_all.h"
+
+#ifndef WZD_USE_PCH
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -60,13 +64,14 @@
 
 #include "wzd_debug.h"
 
-/* BSD exports symbols in .so files prefixed with a _ !! */
-#ifdef BSD
+#ifdef NEED_UNDERSCORE
 #define DL_PREFIX "_"
 #else
 #define DL_PREFIX
 #endif
 
+#endif /* WZD_USE_PCH */
+
 struct event_entry_t {
   unsigned long mask;
   const char *name;
@@ -81,6 +86,7 @@ struct event_entry_t event_tab[] = {
   { EVENT_POSTDOWNLOAD, "POSTDOWNLOAD" },
   { EVENT_MKDIR, "MKDIR" },
   { EVENT_RMDIR, "RMDIR" },
+  { EVENT_DELE, "DELE" },
   { EVENT_SITE, "SITE" },
   { 0, NULL },
 };
@@ -478,7 +484,7 @@ int module_check(const char *filename)
   void * handle;
   void * ptr;
   char * error;
-  struct stat s;
+  struct statbuf s;
   int ret;
 
   if (!filename || filename[0]=='\0') return -1;
@@ -496,7 +502,7 @@ int module_check(const char *filename)
     strcpy(path+2,filename);
   }
 
-  ret = lstat(path,&s);
+  ret = fs_lstat(path,&s);
   if (ret) {
     out_err(LEVEL_HIGH,"Could not stat module '%s'\n",filename);
     out_err(LEVEL_HIGH,"errno: %d error: %s\n",errno, strerror(errno));
