@@ -117,8 +117,8 @@ static void save_pasvrange (FILE *file)
   fprintf( file, "# specify this if you want to get a specific range\n");
   fprintf( file, "#pasv_low_range = 2500\n");
   fprintf( file, "#pasv_high_range  = 3000\n");
-  fprintf( file, "pasv_low_range = %lu\n", mainConfig->pasv_low_range);
-  fprintf( file, "pasv_high_range  = %lu\n", mainConfig->pasv_high_range);
+  fprintf( file, "pasv_low_range = %u\n", mainConfig->pasv_low_range);
+  fprintf( file, "pasv_high_range  = %u\n", mainConfig->pasv_high_range);
   fprintf( file, "\n" );
 }
 
@@ -172,7 +172,7 @@ static void save_dirmessage (FILE *file)
   WZD_ASSERT_VOID( file != NULL );
   fprintf( file, "# the name of the file in each dir that should be added to every answer\n");
   fprintf( file, "#dir_message = .message\n");
-  if (mainConfig->dir_message[0])
+  if (mainConfig->dir_message)
     fprintf( file, "dir_message = %s\n", mainConfig->dir_message);
   fprintf( file, "\n");
 }
@@ -231,9 +231,9 @@ static void save_backend (FILE *file)
     /* TODO XXX FIXME NO NO NO ! string can be auto-generated with backend name ! */
 /*#warning "FIXME NO NO NO ! string can be auto-generated with backend name !"*/
 /*    if (strcmp("plaintext", name) == 0)
-      fprintf( file, "backend_param_plaintext = %s\n", (char*)mainConfig->backend.param);*/
+      fprintf( file, "backend_param_plaintext = %s\n", mainConfig->backend.param);*/
     if (mainConfig->backend.param)
-      fprintf( file, "backend_param_%s = %s\n", name, (char*)mainConfig->backend.param);
+      fprintf( file, "backend_param_%s = %s\n", name, mainConfig->backend.param);
 
     if (version) free (version);
     if (name) free (name);
@@ -286,22 +286,6 @@ static void save_sfvchecker (FILE *file)
 #endif
 
 
-  
-static void save_shmkey (FILE *file)
-{
-  if (file==NULL) return;
-  fprintf( file, "# shm_key: use for share memory (default: 0x1331c0d3)\n");
-  fprintf( file, "#shm_key = 0x2442c0d3\n");
-  fprintf( file, "shm_key = 0x%08lx\n", mainConfig->shm_key);
-  fprintf( file, "\n");
-  fprintf( file, "# force_shm_cleanup: server will always try to delete shm if existing (default: 0)\n");
-  fprintf( file, "# WARNING: if you enable this and a server really exists, ...\n");
-  if (CFG_GET_OPTION(mainConfig,CFG_OPT_FORCE_SHM_CLEANUP))
-    fprintf( file, "force_shm_cleanup = 1\n");
-  else
-    fprintf( file, "#force_shm_cleanup = 1\n");
-  fprintf( file, "\n");
-}
 
 static void save_loglevel (FILE *file)
 {
@@ -322,7 +306,7 @@ static void save_tlsoptions (FILE *file)
   if (file==NULL) return;
   fprintf( file, "# TLS Options\n");
   fprintf( file, "\n");
-  
+
   fprintf( file, "# cygwin, version winsock only: you must specify where the tls wrapper dll is\n");
   current = mainConfig->param_list;
   while (current) {
@@ -337,10 +321,11 @@ static void save_tlsoptions (FILE *file)
   fprintf( file, "\n");
   
   fprintf( file, "# Certificate (only used in ssl mode, otherwise ignored)\n");
-  if (mainConfig->tls_certificate[0])
+  fprintf( file, "#tls_certificate = /etc/wzdftpd/wzd.pem\n");
+/*  if (mainConfig->tls_certificate[0])
     fprintf( file, "tls_certificate = %s\n", mainConfig->tls_certificate);
   else
-    fprintf( file, "#tls_certificate = /etc/wzd.pem\n");
+    fprintf( file, "#tls_certificate = /etc/wzd.pem\n");*/
   fprintf( file, "\n");
   
   fprintf( file, "# Mode (default: explicit)\n");
@@ -358,9 +343,9 @@ static void save_tlsoptions (FILE *file)
   fprintf( file, "# you should not use this option or let \"ALL\" unless you know\n");
   fprintf( file, "# what you are doing\n");
   fprintf( file, "# see openssl ciphers, man openssl(1)\n");
-  if (mainConfig->tls_cipher_list[0])
+/*  if (mainConfig->tls_cipher_list[0])
     fprintf( file, "tls_cipher_list = %s\n", mainConfig->tls_cipher_list);
-  else
+  else*/
     fprintf( file, "#tls_cipher_list = ALL\n");
   fprintf( file, "\n");
   fprintf( file, "# /TLS\n");
@@ -647,11 +632,13 @@ static void save_inclusions (FILE *file)
 
 static void save_permissions (FILE *file)
 {
+#if 0
   char buffer[256];
   char * ptr;
   unsigned int length;
   wzd_command_perm_t * current;
   wzd_command_perm_entry_t * entry;
+#endif /* 0 */
   
   if (file==NULL) return;
   fprintf( file, "##### PERMISSIONS\n");
@@ -662,6 +649,10 @@ static void save_permissions (FILE *file)
   fprintf( file, "# ex: -site_who = =admin -group1 +F =toto\n");
   fprintf( file, "#-delete = -admin\n");
 
+  /** \bug this code is f*cked ! it must now use wzd_command_t, and
+   * perm2str() !
+   */
+#if 0
   /* TODO XXX FIXME perm list will be printed in the reverse order ! */
   current = mainConfig->perm_list;
   while (current) {
@@ -691,6 +682,7 @@ static void save_permissions (FILE *file)
         buffer);
     current = current->next_perm;
   }
+#endif /* 0 */
 
   fprintf( file, "  \n");
 }
@@ -736,7 +728,6 @@ int wzd_savecfg( void )
 #if 0
   save_sfvchecker (file);
 #endif
-  save_shmkey (file);
   save_loglevel (file);
   save_tlsoptions (file);
   save_iprestrictions (file);
