@@ -27,9 +27,18 @@
   * \brief Helper routines for network access
   */
 
+#include "wzd_all.h"
+
+#ifndef WZD_USE_PCH
+
 #if defined(_MSC_VER)
 #include <winsock2.h>
 #include <ws2tcpip.h>
+
+#define EINPROGRESS WSAEINPROGRESS
+#define EWOULDBLOCK WSAEWOULDBLOCK
+#define EISCONN     WSAEISCONN
+
 #else
 #include <unistd.h>
 #include <sys/socket.h>
@@ -46,11 +55,14 @@
 
 #include "wzd_structs.h"
 
+#include "wzd_libmain.h"
 #include "wzd_log.h"
 #include "wzd_socket.h"
 
 #include "wzd_debug.h"
 
+#endif /* WZD_USE_PCH */
+
 /*************** ul2a ***********************************/
 
 char * ul2a(unsigned long q)
@@ -68,22 +80,37 @@ char * ul2a(unsigned long q)
 
 /*************** socket_make ****************************/
 
+int socket_getipbyname(const char *name, char *buffer, size_t length)
+{
+  struct hostent * host;
+  int ret=-1;
+  int sz = MIN(length,sizeof(host->h_addr_list));
+
+  wzd_mutex_lock(server_mutex);
+  host = gethostbyname(name);
+  if (host) {
+    memcpy(buffer, host->h_addr, sz);
+    ret = 0;
+  }
+  wzd_mutex_unlock(server_mutex);
+
+  return ret;
+}
+
+/*************** socket_make ****************************/
+
 /** bind socket at port, if port = 0 picks first free and set it
  * \return -1 or socket
  */
-int socket_make(const char *ip, unsigned int *port, int nListen)
+fd_t socket_make(const char *ip, unsigned int *port, int nListen)
 {
-  struct sockaddr_in sai;
+  struct sockaddr_in sai = { 0 };
 #if defined(IPV6_SUPPORT)
-  struct sockaddr_in6 sai6;
+  struct sockaddr_in6 sai6 = { 0 };
 #endif
 /*  unsigned int c;*/
   size_t c;
-#if defined(_MSC_VER)
-  SOCKET sock;
-#else
-  int sock;
-#endif
+  fd_t sock;
 
   if (ip==NULL || strcmp(ip,"*")==0)
 #if defined(IPV6_SUPPORT)
@@ -95,7 +122,7 @@ int socket_make(const char *ip, unsigned int *port, int nListen)
   {
     struct hostent* host_info;
     // try to decode dotted quad notation
-#if defined(_MSC_VER)
+#if defined(_MSC_VER) || defined(__sun__)
     if ((sai.sin_addr.s_addr = inet_addr(ip)) == INADDR_NONE)
 #else
     if(!inet_aton(ip, &sai.sin_addr))
@@ -115,9 +142,9 @@ int socket_make(const char *ip, unsigned int *port, int nListen)
   }
 
 #if !defined(IPV6_SUPPORT)
-  if ((sock = socket(PF_INET,SOCK_STREAM,0)) == -1) {
+  if ((sock = socket(PF_INET,SOCK_STREAM,0)) == (fd_t)-1) {
 #else
-  if ((sock = socket(PF_INET6,SOCK_STREAM,0)) == -1) {
+  if ((sock = socket(PF_INET6,SOCK_STREAM,0)) == (fd_t)-1) {
 #endif
     out_err(LEVEL_CRITICAL,"Could not create socket %s:%d\n", __FILE__, __LINE__);
     return -1;
@@ -173,14 +200,16 @@ int socket_make(const char *ip, unsigned int *port, int nListen)
   return sock;
 }
 
- 
+
 /*************** socket_close ***************************/
-int socket_close(int sock)
+int socket_close(fd_t sock)
 {
 #if defined(_MSC_VER)
   char acReadBuffer[256];
   int nNewBytes;
 
+  if (sock == (fd_t)-1) return 0; /* invalid fd */
+
   /* Disallow any further data sends.  This will tell the other side
    * that we want to go away now.  If we skip this step, we don't
    * shut the connection down nicely.
@@ -208,7 +237,7 @@ int socket_close(int sock)
       return 1;
     }
     else if (nNewBytes != 0) {
-      out_err(LEVEL_CRITICAL,"\nFYI, received %d unexpected bytes during shutdown.\n",nNewBytes);
+      out_err(LEVEL_FLOOD,"\nFYI, received %d unexpected bytes during shutdown.\n",nNewBytes);
     }
     else {
       /* Okay, we're done! */
@@ -230,9 +259,9 @@ int socket_close(int sock)
 
 /*************** socket_accept **************************/
 
-int socket_accept(int sock, unsigned char *remote_host, unsigned int *remote_port)
+int socket_accept(fd_t sock, unsigned char *remote_host, unsigned int *remote_port)
 {
-  int new_sock;
+  fd_t new_sock;
 #if !defined(IPV6_SUPPORT)
   struct sockaddr_in from;
   size_t len = sizeof(struct sockaddr_in);
@@ -240,11 +269,13 @@ int socket_accept(int sock, unsigned char *remote_host, unsigned int *remote_por
   struct sockaddr_in6 from;
   size_t len = sizeof(struct sockaddr_in6);
 #endif
-  int i;
+#if 0
+  int i=0;
+#endif
 
   new_sock = accept(sock, (struct sockaddr *)&from, &len);
 
-  if (new_sock < 0) {
+  if (new_sock == (fd_t)-1) {
     out_log(LEVEL_CRITICAL,"Accept failed %s:%d\n", __FILE__, __LINE__);
     return -1;
   }
@@ -258,12 +289,14 @@ int socket_accept(int sock, unsigned char *remote_host, unsigned int *remote_por
   fcntl(sock,F_SETFL,(fcntl(sock,F_GETFL)|O_NONBLOCK));
 #endif
 
+#if 0
 #ifndef LINUX
 #ifndef WINSOCK_SUPPORT
 /* see lundftpd : socket.c for explanation */
-  setsockopt(new_sock, SOL_SOCKET, SO_SNDLOWAT, (char*)&i, sizeof(i));
+  setsockopt(new_sock, SOL_SOCKET, SO_SNDLOWAT, &i, sizeof(i)); /** \fixme set correct value for SND_LOWAT ? */
 #endif
 #endif
+#endif /* 0 */
 
 #if !defined(IPV6_SUPPORT)
 #ifndef _MSC_VER
@@ -286,11 +319,64 @@ int socket_accept(int sock, unsigned char *remote_host, unsigned int *remote_por
   return new_sock;
 }
 
+
+
+#ifdef WIN32
+
+/*
+ * waitconnect() returns:
+ * 0    fine connect
+ * -1   select() error
+ * 1    select() timeout
+ * 2    select() returned with an error condition
+ */
+static int _waitconnect(fd_t sockfd, /* socket */
+                int timeout_msec)
+{
+  fd_set fd;
+  fd_set errfd;
+  struct timeval interval;
+  int rc;
+
+  /* now select() until we get connect or timeout */
+  FD_ZERO(&fd);
+  FD_SET(sockfd, &fd);
+
+  FD_ZERO(&errfd);
+  FD_SET(sockfd, &errfd);
+
+  interval.tv_sec = timeout_msec/1000;
+  timeout_msec -= interval.tv_sec*1000;
+
+  interval.tv_usec = timeout_msec*1000;
+
+  rc = select(sockfd+1, NULL, &fd, &errfd, &interval);
+  if(-1 == rc)
+    /* error, no connect here, try next */
+    return -1;
+
+  else if(0 == rc)
+    /* timeout, no connect today */
+    return 1;
+
+  if(FD_ISSET(sockfd, &errfd))
+    /* error condition caught */
+    return 2;
+
+  /* we have a connect! */
+  return 0;
+}
+
+
+#endif /* WIN32 */
+
+
+
 /*************** socket_connect *************************/
 
-int socket_connect(unsigned char * remote_host, int family, int remote_port, int localport, int fd, unsigned int timeout)
+int socket_connect(unsigned char * remote_host, int family, int remote_port, int localport, fd_t fd, unsigned int timeout)
 {
-  int sock;
+  fd_t sock;
   struct sockaddr *sai;
   struct sockaddr_in sai4;
 #if defined(IPV6_SUPPORT)
@@ -299,12 +385,15 @@ int socket_connect(unsigned char * remote_host, int family, int remote_port, int
   size_t len = sizeof(struct sockaddr_in);
   int ret;
   int on=1;
+#ifdef WIN32
+  int error;
+#endif
 
   if (family == WZD_INET4)
   {
     len = sizeof(sai4);
 
-    if ((sock = socket(PF_INET,SOCK_STREAM,0)) < 0) {
+    if ((sock = socket(PF_INET,SOCK_STREAM,0)) == (fd_t)-1) {
       out_log(LEVEL_CRITICAL,"Could not create socket %s:%d\n", __FILE__, __LINE__);
       return -1;
     }
@@ -408,6 +497,45 @@ int socket_connect(unsigned char * remote_host, int family, int remote_port, int
 #endif
 
 #if defined(_MSC_VER)
+
+
+    ret = connect(sock, sai, len);
+    if (ret == -1) {
+      /*long timeout_ms=300000;*/ /* default to 5 min */
+      long timeout_ms=5000;
+      error = WSAGetLastError();
+
+      switch (error) {
+        case EINPROGRESS:
+        case WSAEWOULDBLOCK:
+        case EAGAIN:
+          ret = _waitconnect(sock,timeout_ms);
+          error = WSAGetLastError();
+          if (ret==1) ret=0; /* if we get a timeout here, we do as if we were connected */
+          break;
+        default:
+          out_log(LEVEL_INFO,"Connect failed %s:%d\n", __FILE__, __LINE__);
+          out_log(LEVEL_INFO," errno: %d\n",error);
+          break;
+      }
+    }
+
+    if (ret == 0 || error==0 || error==WSAEWOULDBLOCK || error==WSAEISCONN) {
+      out_err(LEVEL_FLOOD,"Connection OK\n");
+      return sock;
+    }
+
+
+    out_log(LEVEL_INFO,"Connect failed %s:%d\n", __FILE__, __LINE__);
+    out_log(LEVEL_INFO," errno: %d\n",error);
+    socket_close (sock);
+    return -1;
+
+
+
+
+
+
     ret = connect(sock, sai, len);
     if (ret == SOCKET_ERROR) {
       errno = WSAGetLastError();
@@ -444,7 +572,7 @@ int socket_connect(unsigned char * remote_host, int family, int remote_port, int
       out_log(LEVEL_INFO,"Error during connection %d: %s\n",errno,strerror(errno));
       save_errno = WSAGetLastError();
       socket_close(sock);
-      errno = save_errno;
+      WSASetLastError(save_errno);
       return -1;
     }
   }
@@ -476,6 +604,9 @@ int socket_connect(unsigned char * remote_host, int family, int remote_port, int
     out_log(LEVEL_FLOOD,"Connect failed %d %s:%d\n", errno, __FILE__, __LINE__);
     socket_close (sock);
     errno = ret;
+#ifdef WIN32
+    WSASetLastError(ret);
+#endif
     return -1;
   }
 
@@ -485,13 +616,13 @@ int socket_connect(unsigned char * remote_host, int family, int remote_port, int
 /* Returns the local/remote port for the socket. */
 int get_sock_port(int sock, int local)
 {
-#ifndef _MSC_VER
+#if !defined(_MSC_VER) && !defined(__sun__)
   struct sockaddr_storage from;
   char strport[NI_MAXSERV];
 #else
   struct sockaddr_in from;
 #endif
-  socklen_t fromlen;
+  size_t fromlen;
 
   /* Get IP address of client. */
   fromlen = sizeof(from);
@@ -508,7 +639,7 @@ int get_sock_port(int sock, int local)
     }
   }
 
-#ifndef _MSC_VER
+#if !defined(_MSC_VER) && !defined(__sun__)
   /* Work around Linux IPv6 weirdness */
   if (from.ss_family == AF_INET6)
     fromlen = sizeof(struct sockaddr_in6);
@@ -525,21 +656,21 @@ int get_sock_port(int sock, int local)
 
 /* Returns remote/local port number for the current connection. */
 
-int socket_get_remote_port(int sock)
+int socket_get_remote_port(fd_t sock)
 {
   return get_sock_port(sock, 0);
 }
 
-int socket_get_local_port(int sock)
+int socket_get_local_port(fd_t sock)
 {
   return get_sock_port(sock, 1);
 }
 
-int socket_wait_to_read(int sock, unsigned int timeout)
+int socket_wait_to_read(fd_t sock, unsigned int timeout)
 {
   int ret;
   int save_errno;
-  fd_set fds, efds;
+  fd_set rfds, wfds, efds;
   struct timeval tv;
 
   if (sock<0) return -1;
@@ -548,25 +679,34 @@ int socket_wait_to_read(int sock, unsigned int timeout)
     return 0; /* blocking sockets are always ready */
   else {
     while (1) {
-      FD_ZERO(&fds);
+      FD_ZERO(&rfds);
+      FD_ZERO(&wfds);
       FD_ZERO(&efds);
-      FD_SET(sock,&fds);
+      FD_SET(sock,&rfds);
+      FD_SET(sock,&wfds);
       FD_SET(sock,&efds);
       tv.tv_sec = timeout; tv.tv_usec = 0;
 
 #if defined(_MSC_VER)
-      ret = select(0,&fds,NULL,&efds,&tv);
+      ret = select(0,&rfds,&wfds,&efds,&tv);
 #else
-      ret = select(sock+1,&fds,NULL,&efds,&tv);
+      ret = select(sock+1,&rfds,&wfds,&efds,&tv);
 #endif
       save_errno = errno;
 
       if (FD_ISSET(sock,&efds)) {
         if (save_errno == EINTR) continue;
-        out_log(LEVEL_CRITICAL,"Error during socket_wait_to_read: %s\n",strerror(save_errno));
+        out_log(LEVEL_CRITICAL,"Error during socket_wait_to_read: %d %s\n",save_errno,strerror(save_errno));
         return -1;
       }
-      if (!FD_ISSET(sock,&fds)) /* timeout */
+#if 0
+      if (FD_ISSET(sock,&wfds)) {
+        if (save_errno == EINTR) continue;
+        out_log(LEVEL_CRITICAL,"WTF, socket %d wants to write during socket_wait_to_read: %s\n",sock,strerror(save_errno));
+        return -1;
+      }
+#endif
+      if (!FD_ISSET(sock,&rfds)) /* timeout */
         return 1;
       break;
     }
@@ -576,11 +716,11 @@ int socket_wait_to_read(int sock, unsigned int timeout)
   return -1;
 }
 
-int socket_wait_to_write(int sock, unsigned int timeout)
+int socket_wait_to_write(fd_t sock, unsigned int timeout)
 {
   int ret;
   int save_errno;
-  fd_set fds, efds;
+  fd_set rfds, wfds, efds;
   struct timeval tv;
 
   if (sock<0) return -1;
@@ -589,19 +729,27 @@ int socket_wait_to_write(int sock, unsigned int timeout)
     return 0; /* blocking sockets are always ready */
   else {
     while (1) {
-      FD_ZERO(&fds);
+      FD_ZERO(&rfds);
+      FD_ZERO(&wfds);
       FD_ZERO(&efds);
-      FD_SET(sock,&fds);
+      FD_SET(sock,&rfds);
+      FD_SET(sock,&wfds);
       FD_SET(sock,&efds);
       tv.tv_sec = timeout; tv.tv_usec = 0;
 
 #ifdef WIN32
-      ret = select(0,NULL,&fds,&efds,&tv);
+      ret = select(0,NULL,&wfds,&efds,&tv);
 #else
-      ret = select(sock+1,NULL,&fds,&efds,&tv);
+      ret = select(sock+1,NULL,&wfds,&efds,&tv);
 #endif
       save_errno = errno;
 
+      if (ret == -1) return -1;
+
+      if (ret == 0) return 1; /* timeout */
+
+
+
       if (FD_ISSET(sock,&efds)) {
         if (save_errno == EINTR) continue;
 #ifdef WIN32
@@ -610,9 +758,11 @@ int socket_wait_to_write(int sock, unsigned int timeout)
         out_log(LEVEL_CRITICAL,"Error during socket_wait_to_write: %d %s\n",save_errno,strerror(save_errno));
         return -1;
       }
-      if (!FD_ISSET(sock,&fds)) /* timeout */
+#if 0
+      if (!FD_ISSET(sock,&wfds)) /* timeout */
         return 1;
-      break;
+#endif
+      return 0;
     }
     return 0;
   } /* timeout */
