@@ -24,9 +24,16 @@
  * the source code for OpenSSL in the source distribution.
  */
 
+#include "wzd_all.h"
+
+#ifndef WZD_USE_PCH
+
 #include <stdio.h>
+#include <stdlib.h>
 #include <string.h>
 
+#include <time.h>
+
 #ifdef WIN32
 #include <winsock2.h>
 #else
@@ -41,6 +48,7 @@
 
 #include "wzd_structs.h"
 
+#include "wzd_libmain.h"
 #include "wzd_misc.h"
 
 #include "wzd_vars.h"
@@ -51,11 +59,11 @@
 
 #include "wzd_debug.h"
 
+#endif /* WZD_USE_PCH */
 
 
 
 static struct wzd_shm_vars_t * _shm_vars[32] = { NULL };
-static wzd_mutex_t * _shm_mutex = NULL;
 
 
 
@@ -64,7 +72,7 @@ int vars_get(const char *varname, void *data, unsigned int datalength, wzd_confi
   if (!config) return 1;
 
   if (strcasecmp(varname,"bw")==0) {
-    snprintf(data,datalength,"%lu",get_bandwidth());
+    snprintf(data,datalength,"%lu",get_bandwidth(NULL,NULL));
     return 0;
   }
   if (strcmp(varname,"login_pre_ip_check")==0) {
@@ -72,8 +80,12 @@ int vars_get(const char *varname, void *data, unsigned int datalength, wzd_confi
     return 0;
   }
   if (strcmp(varname,"loglevel")==0) {
-    snprintf(data,datalength,"%s",loglevel2str(config->loglevel));
-    return 0;
+    char * level;
+    if (!(chtbl_lookup((CHTBL*)config->htab, "loglevel", (void**)&level))) {
+      snprintf(data,datalength,"%s",level);
+      return 0;
+    }
+    return 1;
   }
   if (strcasecmp(varname,"max_dl")==0) {
     snprintf(data,datalength,"%u",config->global_dl_limiter.maxspeed);
@@ -88,11 +100,11 @@ int vars_get(const char *varname, void *data, unsigned int datalength, wzd_confi
     return 0;
   }
   if (strcasecmp(varname,"pasv_low")==0) {
-    snprintf(data,datalength,"%lu",config->pasv_low_range);
+    snprintf(data,datalength,"%u",config->pasv_low_range);
     return 0;
   }
   if (strcasecmp(varname,"pasv_high")==0) {
-    snprintf(data,datalength,"%lu",config->pasv_high_range);
+    snprintf(data,datalength,"%u",config->pasv_high_range);
     return 0;
   }
   if (strcasecmp(varname,"port")==0) {
@@ -111,7 +123,7 @@ int vars_get(const char *varname, void *data, unsigned int datalength, wzd_confi
   return 1;
 }
 
-int vars_set(const char *varname, void *data, unsigned int datalength, wzd_config_t * config)
+int vars_set(const char *varname, const void *data, unsigned int datalength, wzd_config_t * config)
 {
   int i;
   unsigned long ul;
@@ -146,6 +158,13 @@ int vars_set(const char *varname, void *data, unsigned int datalength, wzd_confi
       return 0;
     }
   }
+  if (strcasecmp(varname,"max_threads")==0) {
+    ul = strtoul(data,&ptr,0);
+    if (ptr && *ptr == '\0') {
+      config->max_threads = ul;
+      return 0;
+    }
+  }
   if (strcasecmp(varname,"max_ul")==0) {
     ul = strtoul(data,&ptr,0);
     if (ptr && *ptr == '\0') {
@@ -194,11 +213,11 @@ int vars_user_get(const char *username, const char *varname, void *data, unsigne
     return 0;
   }
   if (strcasecmp(varname,"max_dl")==0) {
-    snprintf(data,datalength,"%lu",user->max_dl_speed);
+    snprintf(data,datalength,"%u",user->max_dl_speed);
     return 0;
   }
   if (strcasecmp(varname,"max_ul")==0) {
-    snprintf(data,datalength,"%lu",user->max_ul_speed);
+    snprintf(data,datalength,"%u",user->max_ul_speed);
     return 0;
   }
   if (strcasecmp(varname,"credits")==0) {
@@ -268,7 +287,7 @@ int vars_user_addip(const char *username, const char *ip, wzd_config_t *config)
 
   /* commit to backend */
   /* FIXME backend name hardcoded */
-  return backend_mod_user("plaintext", username, user, _USER_IP);
+  return backend_mod_user(config->backend.name, username, user, _USER_IP);
 }
 
 int vars_user_delip(const char *username, const char *ip, wzd_config_t *config)
@@ -324,10 +343,10 @@ int vars_user_delip(const char *username, const char *ip, wzd_config_t *config)
 
   /* commit to backend */
   /* FIXME backend name hardcoded */
-  return backend_mod_user("plaintext", username, user, _USER_IP);
+  return backend_mod_user(config->backend.name, username, user, _USER_IP);
 }
 
-int vars_user_set(const char *username, const char *varname, void *data, unsigned int datalength, wzd_config_t * config)
+int vars_user_set(const char *username, const char *varname, const void *data, unsigned int datalength, wzd_config_t * config)
 {
   wzd_user_t * user;
   unsigned long mod_type;
@@ -370,8 +389,8 @@ int vars_user_set(const char *username, const char *varname, void *data, unsigne
   else if (strcmp(varname, "home")==0) {
     /* check if homedir exist */
     {
-      struct stat s;
-      if (stat(data,&s) || !S_ISDIR(s.st_mode)) {
+      struct statbuf s;
+      if (fs_stat(data,&s) || !S_ISDIR(s.st_mode)) {
         /* Homedir does not exist */
         return 1;
       }
@@ -448,7 +467,7 @@ int vars_user_set(const char *username, const char *varname, void *data, unsigne
 
   /* commit to backend */
   /* FIXME backend name hardcoded */
-  ret = backend_mod_user("plaintext", username, user, mod_type);
+  return backend_mod_user(config->backend.name, username, user, mod_type);
 
   return ret;
 }
@@ -476,8 +495,8 @@ int vars_user_new(const char *username, const char *pass, const char *groupname,
 
   /* check if homedir exist */
   {
-    struct stat s;
-    if (stat(homedir,&s) || !S_ISDIR(s.st_mode)) {
+    struct statbuf s;
+    if (fs_stat(homedir,&s) || !S_ISDIR(s.st_mode)) {
       return 3;
     }
   }
@@ -511,8 +530,7 @@ int vars_user_new(const char *username, const char *pass, const char *groupname,
   user.leech_slots=0;
 
   /* add it to backend */
-  /* FIXME backend name hardcoded */
-  ret = backend_mod_user("plaintext",username,&user,_USER_ALL);
+  ret = backend_mod_user(config->backend.name,username,&user,_USER_ALL);
 
   return ret;
 }
@@ -531,11 +549,11 @@ int vars_group_get(const char *groupname, const char *varname, void *data, unsig
     return 0;
   }
   if (strcasecmp(varname,"max_dl")==0) {
-    snprintf(data,datalength,"%lu",group->max_dl_speed);
+    snprintf(data,datalength,"%u",group->max_dl_speed);
     return 0;
   }
   if (strcasecmp(varname,"max_ul")==0) {
-    snprintf(data,datalength,"%lu",group->max_ul_speed);
+    snprintf(data,datalength,"%u",group->max_ul_speed);
     return 0;
   }
   if (strcasecmp(varname,"name")==0) {
@@ -553,7 +571,7 @@ int vars_group_get(const char *groupname, const char *varname, void *data, unsig
   return 1;
 }
 
-int vars_group_set(const char *groupname, const char *varname, void *data, unsigned int datalength, wzd_config_t * config)
+int vars_group_set(const char *groupname, const char *varname, const void *data, unsigned int datalength, wzd_config_t * config)
 {
   wzd_group_t * group;
   unsigned long mod_type;
@@ -586,8 +604,8 @@ int vars_group_set(const char *groupname, const char *varname, void *data, unsig
   else if (strcmp(varname,"home")==0) {
     /* check if homedir exist */
     {
-      struct stat s;
-      if (stat(data,&s) || !S_ISDIR(s.st_mode)) {
+      struct statbuf s;
+      if (fs_stat(data,&s) || !S_ISDIR(s.st_mode)) {
         /* Homedir does not exist */
         return 2;
       }
@@ -630,11 +648,62 @@ int vars_group_set(const char *groupname, const char *varname, void *data, unsig
 
   /* commit to backend */
   /* FIXME backend name hardcoded */
-  ret = backend_mod_group("plaintext", groupname, group, mod_type);
+  ret = backend_mod_group(config->backend.name, groupname, group, mod_type);
 
   return ret;
 }
 
+int vars_group_new(const char *groupname, wzd_config_t * config)
+{
+  char *homedir;
+  int ret;
+  wzd_group_t newgroup;
+  int i;
+
+  /* check if group already exists */
+  if ( GetGroupByName(groupname) ) {
+    return 1;
+  }
+
+  /* FIXME rootpath */
+  homedir = "";
+#if 0
+  mygroup = GetGroupByID(me->groups[0]);
+  if (mygroup) {
+    homedir = mygroup->defaultpath;
+  } else {
+    homedir = me->rootpath;
+  }
+  /* check if homedir exist */
+  {
+    struct statbuf s;
+    if (fs_stat(homedir,&s) || !S_ISDIR(s.st_mode)) {
+      ret = send_message_with_args(501,context,"Homedir does not exist");
+      str_deallocate(groupname);
+      return 1;
+    }
+  }
+#endif
+
+  /* create new group */
+  strncpy(newgroup.groupname,groupname,sizeof(newgroup.groupname));
+  strncpy(newgroup.defaultpath,homedir,WZD_MAX_PATH);
+  newgroup.groupperms = 0;
+  newgroup.max_idle_time = 0;
+  newgroup.max_dl_speed = 0;
+  newgroup.max_ul_speed = 0;
+  newgroup.ratio = 0;
+  newgroup.num_logins = 0;
+  newgroup.tagline[0] = '\0';
+  for (i=0; i<HARD_IP_PER_GROUP; i++)
+    newgroup.ip_allowed[i][0]='\0';
+
+  /* add it to backend */
+  ret = backend_mod_group(config->backend.name,groupname,&newgroup,_GROUP_ALL);
+
+  return (ret) ? 1 : 0;
+}
+
 
 static unsigned int _str_hash(const char *key)
 {
@@ -652,10 +721,6 @@ static unsigned int _str_hash(const char *key)
 void vars_shm_init(void)
 {
   memset(_shm_vars, 0, sizeof(_shm_vars));
-  if (_shm_mutex) {
-    wzd_mutex_destroy(_shm_mutex);
-  }
-  _shm_mutex = wzd_mutex_create(0x5566423);
 }
 
 void vars_shm_free(void)
@@ -663,7 +728,7 @@ void vars_shm_free(void)
   unsigned int i;
   struct wzd_shm_vars_t * var, * next_var;
 
-  wzd_mutex_lock(_shm_mutex);
+  WZD_MUTEX_LOCK(SET_MUTEX_SHVARS);
   for (i=0; i<32; i++)
   {
     var = _shm_vars[i];
@@ -680,11 +745,7 @@ void vars_shm_free(void)
       var = next_var;
     }
   }
-  wzd_mutex_unlock(_shm_mutex);
-  if (_shm_mutex) {
-    wzd_mutex_destroy(_shm_mutex);
-    _shm_mutex = NULL;
-  }
+  WZD_MUTEX_UNLOCK(SET_MUTEX_SHVARS);
 }
 
 /* finds shm entry corresponding to 'varname'
@@ -715,13 +776,18 @@ struct wzd_shm_vars_t * vars_shm_find(const char *varname, wzd_config_t * config
 int vars_shm_get(const char *varname, void *data, unsigned int datalength, wzd_config_t * config)
 {
   struct wzd_shm_vars_t * var;
+  int ret = 1;
 
+  WZD_MUTEX_LOCK(SET_MUTEX_SHVARS);
   var = vars_shm_find(varname, config);
-  if (!var) return 1;
 
-  memcpy(data, var->data, MIN(datalength,var->datalength));
+  if (var) {
+    memcpy(data, var->data, MIN(datalength,var->datalength));
+    ret = 0;
+  }
 
-  return 0;
+  WZD_MUTEX_UNLOCK(SET_MUTEX_SHVARS);
+  return ret;
 }
 
 /* change varname with data contents size of data is datalength
@@ -732,6 +798,8 @@ int vars_shm_set(const char *varname, void *data, unsigned int datalength, wzd_c
 {
   struct wzd_shm_vars_t * var;
 
+  WZD_MUTEX_LOCK(SET_MUTEX_SHVARS);
+
   var = vars_shm_find(varname, config);
 
   if (!var) { /* new variable, must create it */
@@ -747,13 +815,10 @@ int vars_shm_set(const char *varname, void *data, unsigned int datalength, wzd_c
     memcpy(var->data, data, datalength);
     var->datalength = datalength;
 
-    wzd_mutex_lock(_shm_mutex);
     /* insertion */
     var->next_var = _shm_vars[index];
     _shm_vars[index] = var;
-    wzd_mutex_unlock(_shm_mutex);
   } else {
-    wzd_mutex_lock(_shm_mutex);
     /* modification */
     if (datalength < var->datalength)
       memcpy(var->data, data, datalength);
@@ -762,8 +827,8 @@ int vars_shm_set(const char *varname, void *data, unsigned int datalength, wzd_c
       memcpy(var->data, data, datalength);
       var->datalength = datalength;
     }
-    wzd_mutex_unlock(_shm_mutex);
   }
+  WZD_MUTEX_UNLOCK(SET_MUTEX_SHVARS);
 
   return 0;
 }
