@@ -24,6 +24,10 @@
  * the source code for OpenSSL in the source distribution.
  */
 
+#include "wzd_all.h"
+
+#ifndef WZD_USE_PCH
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -53,17 +57,19 @@
 
 #include "wzd_debug.h"
 
+#endif /* WZD_USE_PCH */
+
 /** remove a vfs from list */
 int vfs_remove( wzd_vfs_t **vfs_list, const char *vpath )
 {
   wzd_vfs_t * current_vfs, * next_vfs;
   wzd_vfs_t * previous_vfs = NULL;
-  
+
   current_vfs = *vfs_list;
   while(current_vfs)
   {
     next_vfs = current_vfs->next_vfs;
-    
+
     if ( (strcmp( current_vfs->virtual_dir, vpath) == 0) )
     {
       if (current_vfs == *vfs_list)
@@ -76,7 +82,7 @@ int vfs_remove( wzd_vfs_t **vfs_list, const char *vpath )
       }
       return 0;
     }
-    
+
     previous_vfs = current_vfs;
     current_vfs = next_vfs;
   }
@@ -117,7 +123,7 @@ int vfs_free(wzd_vfs_t **vfs_list)
 int vfs_add_restricted(wzd_vfs_t ** vfs_list, const char *vpath, const char *path, const char *target)
 {
   wzd_vfs_t * current_vfs, * new_vfs;
-  struct stat s;
+  struct statbuf s;
 
   current_vfs = *vfs_list;
   while (current_vfs)
@@ -130,7 +136,7 @@ int vfs_add_restricted(wzd_vfs_t ** vfs_list, const char *vpath, const char *pat
     current_vfs = current_vfs->next_vfs;
   }
 
-  if (stat(path,&s)) {
+  if (fs_stat(path,&s)) {
     /* destination does not exist */
     return 1;
   }
@@ -138,8 +144,8 @@ int vfs_add_restricted(wzd_vfs_t ** vfs_list, const char *vpath, const char *pat
   new_vfs = wzd_malloc(sizeof(wzd_vfs_t));
   if (!new_vfs) return 1;
 
-  DIRNORM(vpath,strlen(vpath),0);
-  DIRNORM(path,strlen(path),0);
+  DIRNORM((char*)vpath,strlen(vpath),0);
+  DIRNORM((char*)path,strlen(path),0);
   new_vfs->virtual_dir = strdup(vpath);
   new_vfs->physical_dir = strdup(path);
   if (target)
@@ -185,7 +191,7 @@ int vfs_match_perm(const char *perms,wzd_user_t *user)
   buffer=strdup(perms);
   ptr=buffer;
   token = strtok_r(buffer," \t\r\n",&ptr);
- 
+
   while (token) {
     negate=0;
     /* FIXME split token to find entry type : user, group, flag */
@@ -196,18 +202,19 @@ int vfs_match_perm(const char *perms,wzd_user_t *user)
     }
     switch (c) {
     case '=':
-      if (strcasecmp(token,user->username)==0) return (negate) ? 0 : 1;
+      if (strcasecmp(token,user->username)==0) { free(buffer); return (negate) ? 0 : 1; }
       break;
     case '-':
       for (i=0; i<user->group_num; i++) {
         group = GetGroupByID(user->groups[i]);
-        if (strcasecmp(token,group->groupname)==0) return (negate) ? 0 : 1;
+        if (strcasecmp(token,group->groupname)==0) { free(buffer); return (negate) ? 0 : 1; }
       }
       break;
     case '+':
-      if (user->flags && strchr(user->flags,*token)) return (negate) ? 0 : 1;
+      if (user->flags && strchr(user->flags,*token)) { free(buffer); return (negate) ? 0 : 1; }
       break;
     case '*':
+      free(buffer);
       return !negate;
       break;
     default:
@@ -369,13 +376,13 @@ char * vfs_replace_cookies(const char * path, wzd_context_t * context)
 char *stripdir(char * dir, char *buf, int maxlen)
 {
   char * in, * out;
-  char * last; 
+  char * last;
   int ldots;
-        
+
   in   = dir;
   out  = buf;
   last = buf + maxlen;
-  ldots = 0; 
+  ldots = 0;
   *out  = 0;
 
 #ifndef _MSC_VER
@@ -388,10 +395,10 @@ char *stripdir(char * dir, char *buf, int maxlen)
       out = buf + strlen(buf) - 1;
       if (*out != '/') *(++out) = '/';
       out++;
-    }       
+    }
     else
       return NULL;
-  }               
+  }
 
   while (out < last) {
     *out = *in;
@@ -405,6 +412,10 @@ char *stripdir(char * dir, char *buf, int maxlen)
     if (*in == '/' || !*in)
     {
       if (ldots == 1 || ldots == 2) {
+        if (!*in) {
+          if (out-ldots<=dir || *(out-ldots-1) != '/')
+            ldots = 0;
+        }
         while (ldots > 0 && --out > buf)
         {
           if (*out == '/')
@@ -424,18 +435,18 @@ char *stripdir(char * dir, char *buf, int maxlen)
 
     if (!*in)
       break;
-                        
+
     in++;
-  }       
-        
+  }
+
   if (*in) {
     errno = ENOMEM;
     return NULL;
-  }       
-        
+  }
+
   while (--out != buf && (*out == '/' || !*out)) *out=0;
     return buf;
-}       
+}
 
 /** \brief convert ftp-style path to system path
  * \deprecated use \ref checkpath_new
@@ -447,7 +458,7 @@ int checkpath(const char *wanted_path, char *path, wzd_context_t *context)
 
   allowed = malloc(WZD_MAX_PATH);
   cmd = malloc(WZD_MAX_PATH);
-  
+
   {
     snprintf(allowed,WZD_MAX_PATH,"%s/",GetUserByID(context->userid)->rootpath);
     if (strcmp(allowed,"//")==0) allowed[1]='\0';
@@ -461,8 +472,8 @@ int checkpath(const char *wanted_path, char *path, wzd_context_t *context)
     } else {
       strcpy(cmd,allowed);
       strlcat(cmd,wanted_path+1,WZD_MAX_PATH);
-    } 
-  } 
+    }
+  }
   DIRNORM(cmd,strlen(cmd),0);
 /*#ifdef DEBUG
 printf("Checking path '%s' (cmd)\nallowed = '%s'\n",cmd,allowed);
@@ -492,7 +503,7 @@ int checkabspath(const char *wanted_path, char *path, wzd_context_t *context)
 {
   char allowed[WZD_MAX_PATH];
   char cmd[WZD_MAX_PATH];
-  
+
   {
     snprintf(allowed,strlen(allowed),"%s/",GetUserByID(context->userid)->rootpath);
     if (strcmp(allowed,"//")==0) allowed[1]='\0';
@@ -510,8 +521,8 @@ int checkabspath(const char *wanted_path, char *path, wzd_context_t *context)
       return -1; /* we need absolute path, but it doesn't begin with / */
     } else {
       strcpy(cmd,wanted_path);
-    } 
-  } 
+    }
+  }
 /*#ifdef DEBUG
 printf("Checking path '%s' (cmd)\nallowed = '%s'\n",cmd,allowed);
 #endif*/
@@ -553,7 +564,7 @@ int path_abs2rel(const char *abs, char *rel, int rel_len, wzd_context_t *context
 
     /** \todo XXX FIXME this code is NOT finished ... */
     if (strncmp(buffer,vfs->physical_dir,strlen(vfs->physical_dir)) == 0) {
-      
+
     }
   }
 
@@ -583,7 +594,7 @@ int checkpath_new(const char *wanted_path, char *path, wzd_context_t *context)
   char * ptr_ftppath;
   wzd_user_t * user;
   unsigned int sys_offset;
-  struct stat s;
+  struct statbuf s;
   struct wzd_file_t * perm_list, * entry;
 
   user = GetUserByID(context->userid);
@@ -615,7 +626,7 @@ int checkpath_new(const char *wanted_path, char *path, wzd_context_t *context)
       free(syspath); free(ftppath);
       return E_WRONGPATH;
     }
-    
+
     path_simplify(ftppath);
 
     ret = checkpath_new(ftppath, syspath, context);
@@ -670,10 +681,10 @@ int checkpath_new(const char *wanted_path, char *path, wzd_context_t *context)
     strcpy(syspath+sys_offset, lpart);
 
     /** \todo check permissions here */
-    if (lstat(syspath,&s)) {
+    if (fs_lstat(syspath,&s)) {
       /* file/dir does not exist
-       * 3 cases: error, vfs, symlink */ 
-      
+       * 3 cases: error, vfs, symlink */
+
       /* read permission file for parent */
       strcpy(syspath+sys_offset, HARD_PERMFILE);
       perm_list = NULL;
@@ -737,10 +748,12 @@ int checkpath_new(const char *wanted_path, char *path, wzd_context_t *context)
 
           vfs = vfs->next_vfs;
         } /* while (vfs) */
+
+        wzd_free(buffer_vfs);
       } /* check for vfs entries */
 
       /* even if found, check the new destination exists */
-      if (ret || lstat(syspath,&s)) { /* this time, it is really not found */
+      if (ret || fs_lstat(syspath,&s)) { /* this time, it is really not found */
         if (!rpart || *rpart=='\0') {
           /* we return the 'what it would have been' path anyway, so it can be used */
           strcpy(syspath+sys_offset, lpart);
@@ -788,3 +801,58 @@ int checkpath_new(const char *wanted_path, char *path, wzd_context_t *context)
   return 0;
 }
 
+
+int killpath(const char *path, wzd_context_t * context)
+{
+  char * test_realpath;
+  int found = 0;
+  wzd_user_t * me, * user;
+  size_t length;
+
+  if (!path) return E_FILE_NOEXIST;
+
+  length = strlen(path);
+  test_realpath = malloc(WZD_MAX_PATH+1);
+
+  me = GetUserByID(context->userid);
+  WZD_ASSERT( me != NULL );
+  if (checkpath_new(context->currentpath,test_realpath,context)) {
+    free(test_realpath);
+    return E_USER_IDONTEXIST;
+  }
+#if 0
+  /* preliminary check: i can't kill myself */
+  if (strncmp(path,test_realpath,length)==0) {
+    free(test_realpath);
+    return E_USER_ICANTSUICIDE;
+  }
+#endif
+
+  /* kill'em all ! */
+  {
+    ListElmt * elmnt;
+    wzd_context_t * ctxt;
+    for (elmnt=list_head(context_list); elmnt!=NULL; elmnt=list_next(elmnt)) {
+      ctxt = list_data(elmnt);
+      if (ctxt->magic == CONTEXT_MAGIC) {
+        user = GetUserByID(ctxt->userid);
+        WZD_ASSERT( user != NULL );
+        if (ctxt->userid == context->userid) { continue; } /* no suicide */
+        if (checkpath_new(ctxt->currentpath,test_realpath,ctxt) == 0) {
+          if (strncmp(path,test_realpath,length)==0) {
+            found++;
+            kill_child_new(ctxt->pid_child,context);
+          }
+        }
+      }
+    } /* for all contexts */
+  }
+
+  free(test_realpath);
+
+  if (!found) return E_USER_NOBODY;
+
+  return E_OK;
+}
+
+
